theory AdversaryProofs
  imports Main ExecuteTPOD
begin



section \<open>All traces of a TPOD execution are conformant\<close>

section \<open>A TPOD execution without mispredictions trivially satis \<close>




text \<open>The first state of all traces does not speculate.\<close>
text \<open>The initial state of the speculative traces is low equivalent.\<close>
text \<open>The non speculative traces stutter whenever the speculative traces speculate.\<close>

lemma execute_stutters:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
  shows "conformantStutters \<pi>\<^sub>3 \<pi>\<^sub>1 \<and> conformantStutters \<pi>\<^sub>4 \<pi>\<^sub>2"
using assms(1) proof(induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    by (auto simp add: conformantStutters_def liststep_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "conformantStutter s\<^sub>3 s\<^sub>1 s\<^sub>1' \<and> conformantStutter s\<^sub>4 s\<^sub>2 s\<^sub>2'"
    using Step.hyps(1) step4_conformantStutter by blast
  moreover have "length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>2 = length \<pi>\<^sub>4"
    using Step by fastforce
  ultimately show ?case
    by (simp_all add: conformantStutter_append Step.IH)
qed

text \<open>The non speculative traces never mispredict.\<close>

lemma execute_no_mispred_\<pi>12:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "\<not>mispredicts \<pi>\<^sub>1 \<and> \<not>mispredicts \<pi>\<^sub>2" 
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    apply (simp_all only: mispredicts_def)
    by (metis in_set_insert insert_Nil not_Cons_self2 set_ConsD wfs4_s\<^sub>1_s\<^sub>2_no_mispred)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case
    apply (simp_all only: mispredicts_def) 
    using wfs4_s\<^sub>1_s\<^sub>2_no_mispred by auto
qed


text \<open>All four traces execute the same low state.\<close>

lemma execute4_opL_\<pi>:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "((op\<^sub>\<L> \<pi>\<^sub>1)::low_operation list) = op\<^sub>\<L> \<pi>\<^sub>2 \<and> ((op\<^sub>\<L> \<pi>\<^sub>2)::low_operation list) = op\<^sub>\<L> \<pi>\<^sub>3 \<and> ((op\<^sub>\<L> \<pi>\<^sub>3)::low_operation list) = op\<^sub>\<L> \<pi>\<^sub>4"
using assms(1) proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    by (auto simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "((op\<^sub>\<L> s\<^sub>1')::low_operation) = op\<^sub>\<L> s\<^sub>2' \<and> ((op\<^sub>\<L> s\<^sub>2')::low_operation) = op\<^sub>\<L> s\<^sub>3' \<and> ((op\<^sub>\<L> s\<^sub>3')::low_operation) = op\<^sub>\<L> s\<^sub>4'"
    using Step by (auto simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def)
 
  then show ?case
    using Step op\<^sub>\<L>_\<pi>_eq_append_s'_eq by blast
qed

 
text \<open>Trace pairs (\<pi>1, \<pi>3) and (\<pi>2, \<pi>4) execute the same high state.\<close>

lemma execute4_op\<^sub>\<H>_\<pi>:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "((op\<^sub>\<H> \<pi>\<^sub>1)::high_operation list) = op\<^sub>\<H> \<pi>\<^sub>3 \<and> ((op\<^sub>\<H> \<pi>\<^sub>2)::high_operation list) = op\<^sub>\<H> \<pi>\<^sub>4"
using assms(1) proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    by (auto simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "((op\<^sub>\<H> s\<^sub>1')::high_operation) = op\<^sub>\<H> s\<^sub>3' \<and> ((op\<^sub>\<H> s\<^sub>2')::high_operation) = op\<^sub>\<H> s\<^sub>4'"
    using tpod_step\<^sub>4_def tpod_step\<^sub>2_def Step.hyps(7) by blast
 
  then show ?case
    using Step op\<^sub>\<H>_\<pi>_eq_append_s'_eq by blast 
qed

text \<open>The non speculative traces are low equivalent.\<close>

lemma execute4_low_equiv\<^sub>\<L>\<^sub>\<pi>:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "\<pi>\<^sub>1 \<approx>\<^sub>\<L> \<pi>\<^sub>2"
using assms(1) proof(induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case by (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case
    by (meson low_equiv_append2 tpod_step\<^sub>4_def)
qed

text \<open>All traces have conformant entry points? (this will be harder to do).\<close>

lemma execute4_conformantEntrypoints:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "conformantEntrypoints \<pi>\<^sub>1 \<and> conformantEntrypoints \<pi>\<^sub>2 \<and> conformantEntrypoints \<pi>\<^sub>3 \<and> conformantEntrypoints \<pi>\<^sub>4"
using assms(1) proof(induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    by (auto simp add: conformantEntrypoints_def liststep_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "conformantEntrypoint s\<^sub>1 s\<^sub>1' \<and> conformantEntrypoint s\<^sub>2 s\<^sub>2' \<and> conformantEntrypoint s\<^sub>3 s\<^sub>3' 
      \<and> conformantEntrypoint s\<^sub>4 s\<^sub>4'"
    using Step.hyps(1) step4_conformantEntrypoint by blast
  then show ?case
    using Step by (simp only: conformantEntrypoints_append)
qed

text \<open>All traces have conformant store addresses.\<close>

lemma execute4_conformantStoreAddrs:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
  shows "conformantStoreAddrs \<pi>\<^sub>1 \<and> conformantStoreAddrs \<pi>\<^sub>2 \<and> conformantStoreAddrs \<pi>\<^sub>3 \<and> conformantStoreAddrs \<pi>\<^sub>4"
  apply (simp_all only: conformantStoreAddrs_def)
using assms(1) proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case by (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def tpod_step_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "conformantStoreAddr s\<^sub>1 \<and> conformantStoreAddr s\<^sub>2"
    by (simp add: Step.IH)
  have "conformantStoreAddr s\<^sub>1' \<and> conformantStoreAddr s\<^sub>2' \<and> conformantStoreAddr s\<^sub>3' \<and> conformantStoreAddr s\<^sub>4'"
    using Step.hyps(1) \<open>conformantStoreAddr s\<^sub>1 \<and> conformantStoreAddr s\<^sub>2\<close> step4_conformantStoreAddr by blast
  then show ?case 
    using Step by simp
qed


text \<open>All traces are conformant.\<close>

lemma execute\<^sub>T\<^sub>P\<^sub>O\<^sub>D_conformant:
  assumes "execute\<^sub>T\<^sub>P\<^sub>O\<^sub>D \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "conformant \<pi>\<^sub>1 \<and> conformant \<pi>\<^sub>2 \<and> conformant \<pi>\<^sub>3 \<and> conformant \<pi>\<^sub>4"
  using assms execute4_conformantEntrypoints execute4_conformantStoreAddrs by (auto simp add: conformant_def execute\<^sub>T\<^sub>P\<^sub>O\<^sub>D_def)

end