theory Execute
  imports Step4Semantics
begin

inductive execute :: \<open>MachineState list \<Rightarrow> MachineState list \<Rightarrow> MachineState list \<Rightarrow> MachineState list \<Rightarrow> bool\<close> where
Init: \<open>\<lbrakk>
  wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4;
  violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4;
  tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4
\<rbrakk> \<Longrightarrow> execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]\<close> |
Step: \<open>\<lbrakk>
  valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4');
  violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4';
  tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4';
  execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])
\<rbrakk> \<Longrightarrow> execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])\<close>

inductive_cases execute_indcases: "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"

lemma execute_length_eq: 
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
using assms proof (induction rule: execute.inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case by simp
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case
    using "Step.IH" by auto 
qed

lemma execute_empty_simps [simp]:
  "execute [] \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 = False"
  "execute \<pi>\<^sub>1 [] \<pi>\<^sub>3 \<pi>\<^sub>4 = False"
  "execute \<pi>\<^sub>1 \<pi>\<^sub>2 [] \<pi>\<^sub>4 = False"
  "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 [] = False"
  by(subst execute.simps, fastforce)+

lemma execute_Init_case:
  assumes "execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]"
    shows "violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<and> wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<and> tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  using assms by (auto elim: execute.cases)

lemma execute_Step_case:
  assumes "execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])"
    shows "violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<and> tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  using assms by (auto elim: execute.cases)

lemma execute_single:
  assumes "wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
      and "violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
      and "tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    shows "execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]"
  using assms by (simp add: execute.Init)


lemma execute_pair_from_execute:
  assumes "valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4')"
      and "execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]"
      and "violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
      and "tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    shows "execute [s\<^sub>1, s\<^sub>1'] [s\<^sub>2, s\<^sub>2'] [s\<^sub>3, s\<^sub>3'] [s\<^sub>4, s\<^sub>4']"
  using assms execute.Step by (metis self_append_conv2)


lemma execute_pair:
  assumes "valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4')"
      and "wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
      and "violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
      and "tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
      and "violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
      and "tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    shows "execute [s\<^sub>1, s\<^sub>1'] [s\<^sub>2, s\<^sub>2'] [s\<^sub>3, s\<^sub>3'] [s\<^sub>4, s\<^sub>4']"
proof -
  have "execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]"
    using assms by (simp add: execute_single)
  then show ?thesis
    using assms execute_pair_from_execute by blast
qed

lemma execute_append2:
  assumes "execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])"
      and "valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4')"
      and "violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
      and "tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    shows "execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
  using assms execute.Step by blast

lemma execute_append3:
  assumes "execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
      and "valid_step4 (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') (s\<^sub>1'', s\<^sub>2'', s\<^sub>3'', s\<^sub>4'')"
      and "violation\<^sub>\<L> s\<^sub>1'' s\<^sub>2'' s\<^sub>3'' s\<^sub>4''"
      and "tpod_step\<^sub>4 s\<^sub>1'' s\<^sub>2'' s\<^sub>3'' s\<^sub>4''"
    shows "execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1', s\<^sub>1'']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2', s\<^sub>2'']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3', s\<^sub>3'']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4', s\<^sub>4''])"
proof -
  obtain \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4'
    where concat_\<pi>\<^sub>1: "\<pi>\<^sub>1' = \<pi>\<^sub>1 @ [s\<^sub>1]"
      and concat_\<pi>\<^sub>2: "\<pi>\<^sub>2' = \<pi>\<^sub>2 @ [s\<^sub>2]"
      and concat_\<pi>\<^sub>3: "\<pi>\<^sub>3' = \<pi>\<^sub>3 @ [s\<^sub>3]"
      and concat_\<pi>\<^sub>4: "\<pi>\<^sub>4' = \<pi>\<^sub>4 @ [s\<^sub>4]"
    by simp
  have "execute (\<pi>\<^sub>1' @ [s\<^sub>1']) (\<pi>\<^sub>2' @ [s\<^sub>2']) (\<pi>\<^sub>3' @ [s\<^sub>3']) (\<pi>\<^sub>4' @ [s\<^sub>4'])"
    using concat_\<pi>\<^sub>1 concat_\<pi>\<^sub>2 concat_\<pi>\<^sub>3 concat_\<pi>\<^sub>4 assms(1) by auto
  then have "execute (\<pi>\<^sub>1' @ [s\<^sub>1', s\<^sub>1'']) (\<pi>\<^sub>2' @ [s\<^sub>2', s\<^sub>2'']) (\<pi>\<^sub>3' @ [s\<^sub>3', s\<^sub>3'']) (\<pi>\<^sub>4' @ [s\<^sub>4', s\<^sub>4''])"
    using assms(2-4) execute_append2 by blast
  then show ?thesis
    by (simp add: concat_\<pi>\<^sub>1 concat_\<pi>\<^sub>2 concat_\<pi>\<^sub>3 concat_\<pi>\<^sub>4)
qed

lemma hd_append2: 
       "hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<and>
        hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<and>
        hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<and>
        hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
    using Step by (simp add: hd_append)


lemma execute_single_append:
  assumes "execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4]"
      and "execute \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4'"
      and "valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (hd \<pi>\<^sub>1', hd \<pi>\<^sub>2', hd \<pi>\<^sub>3', hd \<pi>\<^sub>4')"
    shows "execute ([s\<^sub>1] @ \<pi>\<^sub>1') ([s\<^sub>2] @ \<pi>\<^sub>2') ([s\<^sub>3] @ \<pi>\<^sub>3') ([s\<^sub>4] @ \<pi>\<^sub>4')"
using assms(2,3) proof (induct rule: execute.inducts)
  case (Init s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')
  then show ?case 
    apply (subgoal_tac "execute [s\<^sub>1, s\<^sub>1'] [s\<^sub>2, s\<^sub>2'] [s\<^sub>3, s\<^sub>3'] [s\<^sub>4, s\<^sub>4']")
    apply simp
    by (metis assms(1) execute_Init_case execute_pair list.sel(1))
next
  case (Step s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' s\<^sub>1'' s\<^sub>2'' s\<^sub>3'' s\<^sub>4'' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  obtain \<pi>\<^sub>1'' \<pi>\<^sub>2'' \<pi>\<^sub>3'' \<pi>\<^sub>4''
    where concat_\<pi>\<^sub>1: "\<pi>\<^sub>1'' = ([s\<^sub>1] @ \<pi>\<^sub>1)"
      and concat_\<pi>\<^sub>2: "\<pi>\<^sub>2'' = ([s\<^sub>2] @ \<pi>\<^sub>2)"
      and concat_\<pi>\<^sub>3: "\<pi>\<^sub>3'' = ([s\<^sub>3] @ \<pi>\<^sub>3)"
      and concat_\<pi>\<^sub>4: "\<pi>\<^sub>4'' = ([s\<^sub>4] @ \<pi>\<^sub>4)"
    by simp
  then have "execute (\<pi>\<^sub>1'' @ [s\<^sub>1']) (\<pi>\<^sub>2'' @ [s\<^sub>2']) (\<pi>\<^sub>3'' @ [s\<^sub>3']) (\<pi>\<^sub>4'' @ [s\<^sub>4'])"
    using hd_append2 by (metis (no_types, lifting) Step.hyps(5) Step.prems append_assoc)
  then show ?case
    using Step by (metis append_assoc concat_\<pi>\<^sub>1 concat_\<pi>\<^sub>2 concat_\<pi>\<^sub>3 concat_\<pi>\<^sub>4 execute_append2)
qed

lemma execute_trace_append:
  assumes "execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
      and "execute \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4'"
      and "valid_step4 (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') (hd \<pi>\<^sub>1', hd \<pi>\<^sub>2', hd \<pi>\<^sub>3', hd \<pi>\<^sub>4')"
    shows "execute (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1'] @ \<pi>\<^sub>1') (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2'] @ \<pi>\<^sub>2') (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3'] @ \<pi>\<^sub>3') (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'] @ \<pi>\<^sub>4')"
using assms(2,3) proof (induct rule: execute.inducts)
  case (Init s\<^sub>1'' s\<^sub>2'' s\<^sub>3'' s\<^sub>4'')
  then show ?case 
    by (metis append.left_neutral append_Cons assms(1) execute_append3 list.sel(1))
next
  case (Step s\<^sub>1'' s\<^sub>2'' s\<^sub>3'' s\<^sub>4'' s\<^sub>1''' s\<^sub>2''' s\<^sub>3''' s\<^sub>4''' \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4')
  obtain \<pi>\<^sub>1'' \<pi>\<^sub>2'' \<pi>\<^sub>3'' \<pi>\<^sub>4''
    where concat_\<pi>\<^sub>1: "\<pi>\<^sub>1'' = (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1'] @ \<pi>\<^sub>1')"
      and concat_\<pi>\<^sub>2: "\<pi>\<^sub>2'' = (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2'] @ \<pi>\<^sub>2')"
      and concat_\<pi>\<^sub>3: "\<pi>\<^sub>3'' = (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3'] @ \<pi>\<^sub>3')"
      and concat_\<pi>\<^sub>4: "\<pi>\<^sub>4'' = (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'] @ \<pi>\<^sub>4')"
    by simp
  then have "execute (\<pi>\<^sub>1'' @ [s\<^sub>1'']) (\<pi>\<^sub>2'' @ [s\<^sub>2'']) (\<pi>\<^sub>3'' @ [s\<^sub>3'']) (\<pi>\<^sub>4'' @ [s\<^sub>4''])"
    using hd_append2 by (metis (no_types, lifting) Step.hyps(5) Step.prems append_assoc)
  then show ?case
    using Step by (metis append_assoc concat_\<pi>\<^sub>1 concat_\<pi>\<^sub>2 concat_\<pi>\<^sub>3 concat_\<pi>\<^sub>4 execute_append2)
qed

lemma execute_joins:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "execute \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4'"
      and "valid_step4 (last \<pi>\<^sub>1, last \<pi>\<^sub>2, last \<pi>\<^sub>3, last \<pi>\<^sub>4) (hd \<pi>\<^sub>1', hd \<pi>\<^sub>2', hd \<pi>\<^sub>3', hd \<pi>\<^sub>4')"
    shows "execute (\<pi>\<^sub>1 @ \<pi>\<^sub>1') (\<pi>\<^sub>2 @ \<pi>\<^sub>2') (\<pi>\<^sub>3 @ \<pi>\<^sub>3') (\<pi>\<^sub>4 @ \<pi>\<^sub>4')"
proof -
  show ?thesis
    using assms(1,3) apply (induct rule: execute.inducts)
    using assms(2) execute_single_append apply (metis execute_single last_ConsL)
    apply (subgoal_tac "valid_step4 (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') (hd \<pi>\<^sub>1', hd \<pi>\<^sub>2', hd \<pi>\<^sub>3', hd \<pi>\<^sub>4')")
    using assms(2) execute_trace_append apply (metis append_assoc execute_append2)
    by (metis (no_types, lifting) last_ConsL last_ConsR last_appendR list.discI)
qed



definition zip4 :: "'a list \<Rightarrow> 'b list \<Rightarrow> 'c list \<Rightarrow> 'd list \<Rightarrow> ('a \<times> 'b \<times> 'c \<times> 'd) list" where
"zip4 xs ys zs ws = zip xs (zip ys (zip zs ws))"


lemma zip4_single: "zip4 [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] = [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]"
  by (simp add: zip4_def)

lemma zip4_append: "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
  length \<pi>\<^sub>1' = length \<pi>\<^sub>2' \<and> length \<pi>\<^sub>1' = length \<pi>\<^sub>3' \<and> length \<pi>\<^sub>1' = length \<pi>\<^sub>4' \<Longrightarrow>
  zip4 (\<pi>\<^sub>1 @ \<pi>\<^sub>1') (\<pi>\<^sub>2 @ \<pi>\<^sub>2') (\<pi>\<^sub>3 @ \<pi>\<^sub>3') (\<pi>\<^sub>4 @ \<pi>\<^sub>4') = zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ zip4 \<pi>\<^sub>1' \<pi>\<^sub>2' \<pi>\<^sub>3' \<pi>\<^sub>4'"
  by (simp add: zip4_def)

lemma zip4_append1: "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
  zip4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) = zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]"
  by (auto simp add: zip4_append zip4_def)

definition "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<longleftrightarrow> 
  (\<forall>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4. (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4) \<longrightarrow> P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)"

lemma Ball4_single: "Ball4 [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] P = P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  by (simp add: Ball4_def zip4_def)

lemma zip4_append_P: 
  assumes "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
      and "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P"
      and "P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4" 
    shows "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P"  
  using assms by (simp add: zip4_append1 Ball4_def)

lemma zip4_append2_P: 
  assumes "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P"
      and "P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    shows "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P"
proof -
  have "zip4 ((\<pi>\<^sub>1 @ [s\<^sub>1]) @ [s\<^sub>1']) ((\<pi>\<^sub>2 @ [s\<^sub>2]) @ [s\<^sub>2']) ((\<pi>\<^sub>3 @ [s\<^sub>3]) @ [s\<^sub>3']) ((\<pi>\<^sub>4 @ [s\<^sub>4]) @ [s\<^sub>4']) = zip4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
    by auto
  then show ?thesis
    using assms Ball4_def zip4_append_P by metis
qed

lemma zip4_append_individual_all_holds:
  assumes "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
      and "(\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')"
    shows "(\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')"
  apply (subgoal_tac "zip4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) = zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]")
  using assms(2) apply auto[1] 
  using assms(1) by (rule zip4_append1)

lemma zip4_append_last_holds: 
  assumes "(\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')"
    shows "P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  using assms by (induct \<pi>\<^sub>1, simp_all)

lemma zip4_append_\<pi>_holds: 
  assumes "(\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')"
    shows "(\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4')"
  using assms by (induct \<pi>\<^sub>1, simp_all)

lemma zip4_appendi_last_holds: 
  assumes "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P"
    shows "P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  using assms zip4_append1 zip4_append_last_holds Ball4_def by metis

lemma zip4_appendi2_last_holds: 
  assumes "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P"
    shows "P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
  using assms by (rule_tac \<pi>\<^sub>1="(\<pi>\<^sub>1 @ [s\<^sub>1])" and \<pi>\<^sub>2="(\<pi>\<^sub>2 @ [s\<^sub>2])" and \<pi>\<^sub>3="(\<pi>\<^sub>3 @ [s\<^sub>3])" and \<pi>\<^sub>4="(\<pi>\<^sub>4 @ [s\<^sub>4])" in zip4_appendi_last_holds, simp_all)

lemma Ball4_append_\<pi>_and_P_holds:
  assumes "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P"
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<and> P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
  apply (rule conjI)
  using assms apply (simp add: Ball4_def)
  apply (subgoal_tac "\<forall>s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'. (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<in> set (zip4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 @ [(s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4)]) \<longrightarrow> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
  using zip4_append_\<pi>_holds apply metis
  using assms(1) zip4_append_individual_all_holds apply metis
  using assms by (rule zip4_appendi_last_holds)

lemma Ball4_append_\<pi>'_and_P'_holds:
  assumes "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P"
    shows "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P \<and> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
  using assms by (rule_tac \<pi>\<^sub>1="(\<pi>\<^sub>1 @ [s\<^sub>1])" and \<pi>\<^sub>2="(\<pi>\<^sub>2 @ [s\<^sub>2])" and \<pi>\<^sub>3="(\<pi>\<^sub>3 @ [s\<^sub>3])" and \<pi>\<^sub>4="(\<pi>\<^sub>4 @ [s\<^sub>4])" in Ball4_append_\<pi>_and_P_holds, simp_all)

lemma Ball4_append_\<pi>_P_and_P'_holds:
  assumes "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
      and "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P"
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<and> P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<and> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
proof -
  have X: "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P \<and> P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    using assms Ball4_append_\<pi>'_and_P'_holds by metis
  have Y: "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
    using assms(1) by auto
  have Z: "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<and> P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    using X Y Ball4_append_\<pi>_and_P_holds by metis
  show ?thesis
    using X Y Z by simp
qed

lemma execute_wfs4:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 wfs4"
using assms proof (induction rule: execute.inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (simp add: Ball4_single)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have length_append: "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step(4) execute_length_eq by blast
  then have length: "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
    by auto
  then have wfs4_start: "wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    using Step(5) zip4_appendi_last_holds by metis
  have "wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    using wfs4_start by (meson Step.hyps(1) step4_wfs4)
  then show ?case
    using Step zip4_append2_P length_append by metis
qed


lemma execute_inducts[consumes 1, case_names Init Step]:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4.
              hd \<pi>\<^sub>1 = s\<^sub>1 \<Longrightarrow>
              hd \<pi>\<^sub>2 = s\<^sub>2 \<Longrightarrow>
              hd \<pi>\<^sub>3 = s\<^sub>3 \<Longrightarrow>
              hd \<pi>\<^sub>4 = s\<^sub>4 \<Longrightarrow>
              execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
            P [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4])"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4.
              valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<Longrightarrow> 
              length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<Longrightarrow>
              hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<Longrightarrow>
              hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<Longrightarrow>
              hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              P (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
            P (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']))" 
    shows "P \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
using assms proof (induction rule: execute.inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    by (simp add: Init.hyps Init.prems(1) Init.prems(2) execute.Init)  
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have length_append: "length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step(4) execute_length_eq by blast
  then have length: "length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4"
    by auto
  have "hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<and>
        hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<and>
        hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<and>
        hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4'])"
    using Step by (simp add: hd_append2)
  then show ?case
    apply (subgoal_tac "violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<and> tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4")
    apply (subgoal_tac "wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4")
    apply (subgoal_tac "wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
    apply (subgoal_tac "P (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])")
    using length_append length Step apply blast 
    using Step apply presburger
    using Step step4_wfs4 apply presburger
    apply (subgoal_tac "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4")
    using Step(3) length apply (smt (z3) zip4_appendi_last_holds)
    using Step(4) execute_wfs4 apply presburger
    using Step execute_Step_case by presburger
qed

subsection \<open>Proof of trivial invariants when not speculating\<close>

lemma execute_Ball4_inducts[consumes 2, case_names Init Step]:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "Q \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4.
              hd \<pi>\<^sub>1 = s\<^sub>1 \<Longrightarrow>
              hd \<pi>\<^sub>2 = s\<^sub>2 \<Longrightarrow>
              hd \<pi>\<^sub>3 = s\<^sub>3 \<Longrightarrow>
              hd \<pi>\<^sub>4 = s\<^sub>4 \<Longrightarrow>
              execute [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              Q [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] \<Longrightarrow>
            Ball4 [s\<^sub>1] [s\<^sub>2] [s\<^sub>3] [s\<^sub>4] P)"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4.
              valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<Longrightarrow> 
              length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
              hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<Longrightarrow>
              hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<Longrightarrow>
              hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<Longrightarrow>
              hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<Longrightarrow>
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              Q (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              (Q (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
                Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P \<and>
                Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<and>
                P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4) \<Longrightarrow>
            Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P)" 
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P"
using assms(1-2) proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case using assms(3) by simp
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case
    using Step assms(4) by (metis Ball4_append_\<pi>_and_P_holds)
qed


lemma execute_Ball4_inducts_init[consumes 3, case_names Step]:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)" 
      and "Q \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4.
              valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<Longrightarrow> 
              length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<Longrightarrow>
              hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<Longrightarrow>
              hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<Longrightarrow>
              hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              Q (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              (Q (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
                Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P \<and>
                Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P \<and>
                P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4) \<Longrightarrow>
            Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P)" 
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P"
proof -
  obtain Q'
    where precon: "Q' = (\<lambda>\<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4. P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4) \<and> Q \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)"
    by simp
  then have x: "Q' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
    using assms(2, 3) by simp
  show ?thesis
    using assms(1) x proof (induction rule: execute_Ball4_inducts)
    case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
    then show ?case by (metis Ball4_single assms(2))
  next
    case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
    then show ?case
    proof -
      have "\<forall>ms msa msb msc. Q' ms msa msb msc = (P (hd ms) (hd msa) (hd msb) (hd msc) \<and> Q ms msa msb msc)"
        using precon by blast
      then show ?thesis
        using Step assms(4) by presburger
    qed
  qed
qed

lemma execute_Ball4_inducts_init_Ball4[consumes 3, case_names Step]:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)" 
      and "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 Q"
      and "(\<And>s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4.
              valid_step4 (s\<^sub>1, s\<^sub>2, s\<^sub>3, s\<^sub>4) (s\<^sub>1', s\<^sub>2', s\<^sub>3', s\<^sub>4') \<Longrightarrow>
              wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              wfs4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              violation\<^sub>\<L> s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              tpod_step\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<Longrightarrow>
              length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<Longrightarrow> 
              length \<pi>\<^sub>1 = length \<pi>\<^sub>4 \<Longrightarrow>
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>2 @ [s\<^sub>2]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>3 @ [s\<^sub>3]) \<Longrightarrow> 
              length (\<pi>\<^sub>1 @ [s\<^sub>1]) = length (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              hd (\<pi>\<^sub>1 @ [s\<^sub>1]) = hd (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) \<Longrightarrow>
              hd (\<pi>\<^sub>2 @ [s\<^sub>2]) = hd (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) \<Longrightarrow>
              hd (\<pi>\<^sub>3 @ [s\<^sub>3]) = hd (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) \<Longrightarrow>
              hd (\<pi>\<^sub>4 @ [s\<^sub>4]) = hd (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) \<Longrightarrow>
              execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) \<Longrightarrow>
              Q s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
              Q s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<Longrightarrow>
              Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 Q \<Longrightarrow>
              Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) Q \<Longrightarrow>             
              Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) Q \<Longrightarrow>
              Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) P \<Longrightarrow>
              P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 \<Longrightarrow>
            Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1, s\<^sub>1']) (\<pi>\<^sub>2 @ [s\<^sub>2, s\<^sub>2']) (\<pi>\<^sub>3 @ [s\<^sub>3, s\<^sub>3']) (\<pi>\<^sub>4 @ [s\<^sub>4, s\<^sub>4']) P)" 
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 P"
using assms(1-3) proof (induction rule: execute_Ball4_inducts_init)
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case 
    using assms(4) by (metis Ball4_append_\<pi>'_and_P'_holds Ball4_append_\<pi>_P_and_P'_holds)
qed


lemma execute_ns_wfs4_triv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)" 
      and "\<not> speculates \<pi>\<^sub>3 \<and> \<not> speculates \<pi>\<^sub>4"
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 wfs4_triv"
using assms proof (induct rule: execute_Ball4_inducts_init)
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not>speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not>speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (metis not_speculates_append not_speculates_append2)
  moreover have "wfs4_triv s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    using wfs4_ns_triv Step calculation by (metis not_speculates_append2)
  ultimately show ?case
    using Step zip4_append2_P by metis
qed
    
subsection \<open>Proof of non speculative invariants when not speculating\<close>

lemma execute_ns_wfs4_ns:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)" 
      and "\<not> speculates \<pi>\<^sub>3 \<and> \<not> speculates \<pi>\<^sub>4"
    shows "Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 wfs4_ns"
using assms proof (induction rule: execute_Ball4_inducts_init)
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not>speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not>speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (metis not_speculates_append not_speculates_append2)
  moreover have "wfs4_ns s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'"
    using wfs4_ns Step calculation by (metis not_speculates_append2)
  then show ?case
    using Step zip4_append2_P by (metis not_speculates_append not_speculates_append2)
qed

subsection \<open>Proof that if one trace is not speculating then the other is also not speculating\<close>

lemma execute4_ns\<pi>\<^sub>3_ns\<pi>\<^sub>4:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculating (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3"
    shows "\<not>speculates \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (auto simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have ns: "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])"
    using Step by (auto simp add: speculates_def)
  have "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4_triv"
    using Step execute_ns_wfs4_triv ns by presburger
  then have invs_ns: "wfs4_triv s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    by (meson Step.hyps(10) Step.hyps(8) Step.hyps(9) zip4_appendi_last_holds)
  have "\<not>speculating s\<^sub>3 \<and> \<not>speculating s\<^sub>4 \<and> \<not>speculating s\<^sub>3'"
    using Step by (auto simp add: speculates_def)
  then have step4_ns_after: "\<not>speculating s\<^sub>4'"
    using Step invs_ns wfs4_ns_not_or by blast
  then show ?case    
    using Step by (metis not_speculates_append not_speculates_append2)
qed

lemma execute4_ns\<pi>\<^sub>4_ns\<pi>\<^sub>3:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculating (hd \<pi>\<^sub>3)"
      and "\<not>speculates \<pi>\<^sub>4"
    shows "\<not>speculates \<pi>\<^sub>3"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (auto simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have ns: "\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (auto simp add: speculates_def)
  have "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4_triv"
    using Step execute_ns_wfs4_triv ns by presburger
  then have invs_ns: "wfs4_triv s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    by (meson Step.hyps(10) Step.hyps(8) Step.hyps(9) zip4_appendi_last_holds)
  have "\<not>speculating s\<^sub>3 \<and> \<not>speculating s\<^sub>4 \<and> \<not>speculating s\<^sub>4'"
    using Step by (auto simp add: speculates_def)
  then have step4_ns_after: "\<not>speculating s\<^sub>3'"
    using Step invs_ns wfs4_ns_not_or by blast
  then show ?case    
    using Step by (metis not_speculates_append not_speculates_append2)
qed

subsection \<open>Proof that if one trace is always low then all traces are always low\<close>

lemma execute4_op\<^sub>\<L>_\<pi>:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>1) \<longrightarrow> i \<noteq> \<bottom>) \<or>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>2) \<longrightarrow> i \<noteq> \<bottom>) \<or>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>3) \<longrightarrow> i \<noteq> \<bottom>) \<or>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>4) \<longrightarrow> i \<noteq> \<bottom>)"
    shows "(\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>1) \<longrightarrow> i \<noteq> \<bottom>) \<and>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>2) \<longrightarrow> i \<noteq> \<bottom>) \<and>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>3) \<longrightarrow> i \<noteq> \<bottom>) \<and>
           (\<forall>i. i \<in> set (op\<^sub>\<L> \<pi>\<^sub>4) \<longrightarrow> i \<noteq> \<bottom>)"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case    
    apply (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def)
    by metis
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "(\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>1 @ [s\<^sub>1])) \<longrightarrow> i \<noteq> \<bottom>) \<or>
        (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>2 @ [s\<^sub>2])) \<longrightarrow> i \<noteq> \<bottom>) \<or>
        (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>3 @ [s\<^sub>3])) \<longrightarrow> i \<noteq> \<bottom>) \<or>
        (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>4 @ [s\<^sub>4])) \<longrightarrow> i \<noteq> \<bottom>)"
    using Step(20) by auto

  then have x: "(\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>1 @ [s\<^sub>1])) \<longrightarrow> i \<noteq> \<bottom>) \<and>
                (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>2 @ [s\<^sub>2])) \<longrightarrow> i \<noteq> \<bottom>) \<and>
                (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>3 @ [s\<^sub>3])) \<longrightarrow> i \<noteq> \<bottom>) \<and>
                (\<forall>i. i \<in> set (op\<^sub>\<L> (\<pi>\<^sub>4 @ [s\<^sub>4])) \<longrightarrow> i \<noteq> \<bottom>)"
    using Step(19) by blast

  have "op\<^sub>\<L> s\<^sub>1' \<noteq> \<bottom> \<or> op\<^sub>\<L> s\<^sub>2' \<noteq> \<bottom> \<or> op\<^sub>\<L> s\<^sub>3' \<noteq> \<bottom> \<or> op\<^sub>\<L> s\<^sub>4' \<noteq> \<bottom>"
    using Step(20) by auto
  
  then have "op\<^sub>\<L> s\<^sub>1' \<noteq> \<bottom> \<and> op\<^sub>\<L> s\<^sub>2' \<noteq> \<bottom> \<and> op\<^sub>\<L> s\<^sub>3' \<noteq> \<bottom> \<and> op\<^sub>\<L> s\<^sub>4' \<noteq> \<bottom>"
    using Step(7) apply (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def) 
    by metis

  then show ?case    
    using x by auto
qed

subsection \<open>Proof that if one trace is always high then all traces are always high\<close>

lemma execute4_op\<^sub>\<H>_\<pi>:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<forall>s. s \<in> set \<pi>\<^sub>1 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<or>
           (\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
      and "(\<forall>s. s \<in> set \<pi>\<^sub>2 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<or>
           (\<forall>s. s \<in> set \<pi>\<^sub>4 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    shows "(\<forall>s. s \<in> set \<pi>\<^sub>1 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
           (\<forall>s. s \<in> set \<pi>\<^sub>2 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
           (\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
           (\<forall>s. s \<in> set \<pi>\<^sub>4 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def op\<^sub>\<H>_sub_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have s\<^sub>1\<^sub>3_inst\<^sub>\<T>: "(\<forall>s. s \<in> set (\<pi>\<^sub>1 @ [s\<^sub>1]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<or>
                  (\<forall>s. s \<in> set (\<pi>\<^sub>3 @ [s\<^sub>3]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    using Step(20) by auto

  have s\<^sub>2\<^sub>4_inst\<^sub>\<T>: "(\<forall>s. s \<in> set (\<pi>\<^sub>2 @ [s\<^sub>2]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<or>
                  (\<forall>s. s \<in> set (\<pi>\<^sub>4 @ [s\<^sub>4]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    using Step(21) by auto

  have x: "(\<forall>s. s \<in> set (\<pi>\<^sub>1 @ [s\<^sub>1]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
        (\<forall>s. s \<in> set (\<pi>\<^sub>2 @ [s\<^sub>2]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
        (\<forall>s. s \<in> set (\<pi>\<^sub>3 @ [s\<^sub>3]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and>
        (\<forall>s. s \<in> set (\<pi>\<^sub>4 @ [s\<^sub>4]) \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    using s\<^sub>1\<^sub>3_inst\<^sub>\<T> s\<^sub>2\<^sub>4_inst\<^sub>\<T> Step(19) by fastforce

  have "(inst\<^sub>\<T> s\<^sub>1' \<noteq> \<bottom> \<or> inst\<^sub>\<T> s\<^sub>3' \<noteq> \<bottom>) \<and> (inst\<^sub>\<T> s\<^sub>2' \<noteq> \<bottom> \<or> inst\<^sub>\<T> s\<^sub>4' \<noteq> \<bottom>)"
    using Step(20) Step(21) by auto
  
  then have "inst\<^sub>\<T> s\<^sub>1' \<noteq> \<bottom> \<and> inst\<^sub>\<T> s\<^sub>2' \<noteq> \<bottom> \<and> inst\<^sub>\<T> s\<^sub>3' \<noteq> \<bottom> \<and> inst\<^sub>\<T> s\<^sub>4' \<noteq> \<bottom>"
    using Step(7) by (simp add: tpod_step\<^sub>4_def tpod_step\<^sub>2_def op\<^sub>\<H>_sub_def)

  then show ?case
    using x by auto
qed
end
