theory ExecuteTrivial
  imports Execute
begin

section \<open>Trivial discharges of TPOD for partial executions\<close>

subsection \<open>No speculation\<close>

text \<open>Neither traces \<pi>$_3$ or \<pi>$_4$ speculate at the same time\<close>


lemma execute_strict_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (auto simp add: low_equiv_sub_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have ns: "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (auto simp add: speculates_def)
  have step4_before_invs: "s\<^sub>3 \<approx>\<^sub>\<L> s\<^sub>4"
    using Step apply (simp)    
    by (metis list.rel_inject(2) list_all2_append ns)
  have execute_wfs4_ns: "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4_triv"
    using Step by (simp add: execute_ns_wfs4_triv ns)
  then have invs_ns: "wfs4_triv s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    using zip4_appendi_last_holds Step by metis
  have "\<not>speculating s\<^sub>3 \<and> \<not>speculating s\<^sub>4 \<and> \<not>speculating s\<^sub>3' \<and> \<not>speculating s\<^sub>4'"
    using Step by (auto simp add: speculates_def)
  then have step4_after_invs: "s\<^sub>3' \<approx>\<^sub>\<L> s\<^sub>4'"
    using Step.hyps(1) Step.hyps(2) invs_ns step4_before_invs step4_ns_triv_and_triv_low_equiv by blast
  show ?case 
    using Step.IH Step.hyps(14) Step.hyps(15) Step.hyps(16) Step.hyps(17) Step.prems(1) Step.prems(2) low_equiv_append2 ns step4_after_invs by presburger
qed

lemma execute_strict_ns_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms(1) assms(2) assms(3) assms(4) execute_strict_ns_low_equiv wfs4_ns_def by blast


lemma execute_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms by (meson execute4_ns\<pi>\<^sub>3_ns\<pi>\<^sub>4 execute4_ns\<pi>\<^sub>4_ns\<pi>\<^sub>3 execute_strict_ns_low_equiv)

lemma execute_ns_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_ns_low_equiv wfs4_ns_def by blast

lemma execute_s\<^sub>3_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3"
      and "\<not>speculating (hd \<pi>\<^sub>3)"
    shows "\<not>speculates \<pi>\<^sub>3"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
    then show ?case by (simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not> mispred (n s\<^sub>3) (\<beta> s\<^sub>3) (pc s\<^sub>3)"
    using Step by (auto simp add: mispredicts_def)
  have "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])"
    apply (subgoal_tac "\<not> speculating (hd (\<pi>\<^sub>3 @ [s\<^sub>3]))")
    apply (subgoal_tac "\<not> mispredicts (\<pi>\<^sub>3 @ [s\<^sub>3])")
    using Step apply blast
    using Step apply (simp add: mispredicts_def)
    apply metis  
    by (simp add: Step.hyps(16) Step.hyps(17) Step.prems(2))
  then have "\<not> speculating s\<^sub>3"
    using \<open>\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])\<close> not_speculates_append by auto
  then have "\<not> speculating s\<^sub>3'"
    using step4_s\<^sub>3_ns_s_mispred 
    using Step.hyps(1) \<open>\<not> mispred (n s\<^sub>3) (\<beta> s\<^sub>3) (pc s\<^sub>3)\<close> by blast
  then show ?case
    by (meson \<open>\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])\<close> not_speculates_append not_speculates_append2)
qed

lemma execute_s\<^sub>4_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
    then show ?case by (simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not> mispred (n s\<^sub>4) (\<beta> s\<^sub>4) (pc s\<^sub>4)"
    using Step by (auto simp add: mispredicts_def)
  have "\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    apply (subgoal_tac "\<not> speculating (hd (\<pi>\<^sub>4 @ [s\<^sub>4]))")
    apply (subgoal_tac "\<not> mispredicts (\<pi>\<^sub>4 @ [s\<^sub>4])")
    using Step apply blast
    using Step apply (simp add: mispredicts_def)
    apply metis  
    by (simp add: Step.hyps(16) Step.hyps(17) Step.prems(2))
  then have "\<not> speculating s\<^sub>4"
    using \<open>\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])\<close> not_speculates_append by auto
  then have "\<not> speculating s\<^sub>4'"
    using step4_s\<^sub>4_ns_s_mispred 
    using Step.hyps(1) \<open>\<not> mispred (n s\<^sub>4) (\<beta> s\<^sub>4) (pc s\<^sub>4)\<close> by blast
  then show ?case
    by (meson \<open>\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])\<close> not_speculates_append not_speculates_append2)
qed

lemma execute_no_mispred_or_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3 \<or> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
  using assms execute_s\<^sub>3_no_mispred_ns_init_implies_speculating 
        execute_s\<^sub>4_no_mispred_ns_init_implies_speculating by blast

lemma execute_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
  using assms execute_s\<^sub>3_no_mispred_ns_init_implies_speculating 
        execute_s\<^sub>4_no_mispred_ns_init_implies_speculating by auto


lemma execute_no_mispred_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_no_mispred_ns_init_implies_speculating execute_ns_low_equiv by blast


lemma execute_no_mispred_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_no_mispred_low_equiv wfs4_ns_def by meson





subsection \<open>Execution in the high state is trivially low equivalent\<close>

lemma set_P_append: 
  "(\<forall>s. s \<in> set (\<pi> @ [s']) \<longrightarrow> P s) \<longleftrightarrow> (\<forall>s. s \<in> set \<pi> \<longrightarrow> P s) \<and> P s'"
  by auto

lemma set_P_append2: 
  "(\<forall>s. s \<in> set (\<pi> @ [s', s'']) \<longrightarrow> P s) \<longleftrightarrow> (\<forall>s. s \<in> set (\<pi> @ [s']) \<longrightarrow> P s) \<and> P s''"
  by auto


lemma execute_hi_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    apply (simp add: low_equiv_sub_def)
    by (metis option.discI)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case 
    using low_equiv_s_op_high low_equiv_append2 by auto
qed

lemma execute_hi_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and> (\<forall>s. s \<in> set \<pi>\<^sub>4 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  oops


subsection \<open>Program analysis\<close>

abbreviation \<open>forward_jumps4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 
    \<equiv> forward_jumps s\<^sub>1 \<and> forward_jumps s\<^sub>2 \<and> forward_jumps s\<^sub>3 \<and> forward_jumps s\<^sub>4\<close>

lemma execute_forward_jumps:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
  using assms apply (induct_tac rule: execute_Ball4_inducts_init)
  apply simp
  apply simp
  apply simp
  using step4_forward_jumps by (metis zip4_append2_P)

abbreviation \<open>future\<^sub>\<rho>_P4 P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4
    \<equiv> future\<^sub>\<rho>_P s\<^sub>1 P \<and> future\<^sub>\<rho>_P s\<^sub>2 P \<and> future\<^sub>\<rho>_P s\<^sub>3 P \<and> future\<^sub>\<rho>_P s\<^sub>4 P\<close>

lemma execute_future\<^sub>\<rho>_Ball_forward_jumps_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<rho>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<rho>_P4 P)\<close>
  using assms apply (induction rule: execute_Ball4_inducts_init_Ball4)
  apply (subgoal_tac "future\<^sub>\<rho>_P4 P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
  using zip4_append2_P apply metis
  using step4_future\<^sub>\<rho>_P by blast

lemma execute_future\<^sub>\<rho>_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<rho>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<rho>_P4 P)\<close>
  using assms by (simp add: execute_forward_jumps execute_future\<^sub>\<rho>_Ball_forward_jumps_P)

abbreviation \<open>future\<^sub>\<iota>_P4 P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4
    \<equiv> future\<^sub>\<iota>_P s\<^sub>1 P \<and> future\<^sub>\<iota>_P s\<^sub>2 P \<and> future\<^sub>\<iota>_P s\<^sub>3 P \<and> future\<^sub>\<iota>_P s\<^sub>4 P\<close>

lemma execute_future\<^sub>\<iota>_Ball_forward_jumps_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 P)\<close>
  using assms apply (induction rule: execute_Ball4_inducts_init_Ball4)
  apply (subgoal_tac "future\<^sub>\<iota>_P4 P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
  using zip4_append2_P apply metis
  using step4_future\<^sub>\<iota>_P by blast

lemma execute_future\<^sub>\<iota>_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 P)\<close>
  using assms by (simp add: execute_forward_jumps execute_future\<^sub>\<iota>_Ball_forward_jumps_P)

end