theory HoareTriple
  imports Main
begin

subsection \<open>Generic Hoare Triple\<close>

text \<open>Definition of a generic Hoare Triple in this simple state model.\<close>

type_synonym ('s) Predicate = \<open>'s \<Rightarrow> bool\<close>
type_synonym ('s) Transformer = \<open>'s \<Rightarrow> 's\<close>


definition triple :: 
  \<open>('s) Predicate \<Rightarrow> ('s) Transformer \<Rightarrow> ('s) Predicate \<Rightarrow> bool\<close> (\<open>\<lbrace>_\<rbrace>/ _ /\<lbrace>_\<rbrace>\<close>) 
where 
  \<open>\<lbrace>P\<rbrace> f \<lbrace>Q\<rbrace> \<equiv> \<forall>s . P s \<longrightarrow> Q (f s)\<close>


lemma name_me: \<open>\<lbrace>\<lambda>s'. s' = s\<rbrace> f \<lbrace>\<lambda>s'. s' = f s\<rbrace>\<close>
  by (simp add: triple_def)

lemma valid_wp: \<open>\<lbrace>\<lambda>s. Q (f s)\<rbrace> f \<lbrace>Q\<rbrace>\<close>
  by (simp add: triple_def)

lemma valid_wp_equiv_name_me: \<open>\<lbrace>\<lambda>s. Q (f s)\<rbrace> f \<lbrace>Q\<rbrace> \<longleftrightarrow> \<lbrace>\<lambda>s'. s' = s\<rbrace> f \<lbrace>\<lambda>s'. s' = f s\<rbrace>\<close>
  by (simp add: name_me valid_wp)


lemma excluded_miracle: \<open>\<lbrace>P\<rbrace> S \<lbrace>\<lambda>_. False\<rbrace> \<longleftrightarrow> (\<forall>s . \<not>P s)\<close>
  by (simp add: triple_def)

lemma termin: \<open>\<lbrace>P\<rbrace> S \<lbrace>\<lambda>_. True\<rbrace>\<close>
  by (simp add: triple_def)

lemma strengthen_pre: \<open>(\<And>s . P' s \<longrightarrow> P s) \<Longrightarrow> \<lbrace>P\<rbrace> S \<lbrace>Q\<rbrace> \<Longrightarrow> \<lbrace>P'\<rbrace> S \<lbrace>Q\<rbrace>\<close>
  by (simp add: triple_def)

lemma weakening_post: \<open>(\<And>s . Q s \<longrightarrow> Q' s) \<Longrightarrow> \<lbrace>P\<rbrace> S \<lbrace>Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> S \<lbrace>Q'\<rbrace>\<close>
  by (simp add: triple_def)

lemma post_conj: \<open>(\<lbrace>P\<rbrace> S \<lbrace>Q1\<rbrace> \<and> \<lbrace>P\<rbrace> S \<lbrace>Q2\<rbrace>) \<longleftrightarrow> \<lbrace>P\<rbrace> S \<lbrace>\<lambda>s. Q1 s \<and> Q2 s\<rbrace>\<close>
  by (auto simp add: triple_def)

lemma pre_disj: \<open>(\<lbrace>P1\<rbrace> S \<lbrace>Q\<rbrace> \<and> \<lbrace>P2\<rbrace> S \<lbrace>Q\<rbrace>) \<longleftrightarrow> \<lbrace>\<lambda>s. P1 s \<or> P2 s\<rbrace> S \<lbrace>Q\<rbrace>\<close>
  by (auto simp add: triple_def)

lemma post_disj: \<open>\<lbrace>P\<rbrace> S \<lbrace>Q1\<rbrace> \<or> \<lbrace>P\<rbrace> S \<lbrace>Q2\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> S \<lbrace>\<lambda>s. Q1 s \<or> Q2 s\<rbrace>\<close>
  by (auto simp add: triple_def)

lemma triple_concat: \<open>\<lbrace>P\<rbrace> S0 \<lbrace>R\<rbrace> \<Longrightarrow> \<lbrace>R\<rbrace> S1 \<lbrace>Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> (\<lambda>s. S1 (S0 s)) \<lbrace>Q\<rbrace>\<close>
  by (auto simp add: triple_def)

end