theory ExpressionSemantics
  imports Main MachineState
begin

section \<open>Expression Semantics\<close>


primrec eval_exp_rec :: \<open>Exp \<Rightarrow> Registers \<Rightarrow> Const\<close>
  where
\<open>eval_exp_rec (Const v) _ = v\<close> |
\<open>eval_exp_rec (Reg r) rs = the (rs r)\<close> |
\<open>eval_exp_rec (UnOp f e) rs = f (eval_exp_rec e rs)\<close> |
\<open>eval_exp_rec (BinOp e\<^sub>1 f e\<^sub>2) rs = f (eval_exp_rec e\<^sub>1 rs) (eval_exp_rec e\<^sub>2 rs)\<close> 


definition \<open>as_bool c \<longleftrightarrow> \<not>(c = 0x0)\<close> 

text \<open>Define an overloaded syntax for Expression evaluation\<close>

consts eval_exp :: \<open>Exp \<Rightarrow> Registers \<Rightarrow> 'a\<close> (\<open>\<lbrakk>_\<rbrakk>_\<close> [0, 10] 10) 

text \<open>Use pretty syntax when outputting evaluations in latex \<lbrakk>e\<rbrakk>$_{\<Delta>_n}$ \<close>

notation (latex output)
  eval_exp  (\<open>\<lbrakk>_\<rbrakk>\<^bsub>_\<^esub>\<close> [111,110] 110)

text \<open>Overload \<lbrakk>e\<rbrakk>$_{\<Delta>_n}$ to return Const and bool from eval_exp_rec using as_bool\<close>

overloading
  eval_exp_const \<equiv> \<open>eval_exp :: Exp \<Rightarrow> Registers \<Rightarrow> Const\<close>
  eval_exp_bool  \<equiv> \<open>eval_exp :: Exp \<Rightarrow> Registers \<Rightarrow> bool\<close>
begin

fun eval_exp_const :: \<open>Exp \<Rightarrow> Registers \<Rightarrow> Const\<close> 
  where \<open>eval_exp_const e rs = eval_exp_rec e rs\<close>

fun eval_exp_bool :: \<open>Exp \<Rightarrow> Registers \<Rightarrow> bool\<close> 
  where \<open>eval_exp_bool e rs = as_bool (eval_exp_rec e rs)\<close>

end

text \<open>Verify Constant evaluations correspond to correct boolean evaluations\<close>

lemma eval_exp_false_zero: \<open>(\<lbrakk>e\<rbrakk>rs) = (0::Const) \<longleftrightarrow> \<not>(\<lbrakk>e\<rbrakk>rs)\<close>
  by (simp add: as_bool_def)

lemma eval_exp_true_non_zero: \<open>(\<lbrakk>e\<rbrakk>rs) \<noteq> (0::Const) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>rs)\<close>
  by (simp add: as_bool_def)

text \<open>Ensure the operational semantics are correct/consistent with those defined in the paper.\<close>

lemma eval_Const: \<open>c = (\<lbrakk>Const c\<rbrakk>rs)\<close>
  by auto

lemma eval_reg: \<open>the (rs r) = v \<Longrightarrow> v = (\<lbrakk>Reg r\<rbrakk>rs)\<close>  
  by auto

lemma eval_unop: \<open>\<lbrakk>v' = (\<lbrakk>e\<rbrakk>rs); v = f v'\<rbrakk> \<Longrightarrow> v = (\<lbrakk>UnOp f e\<rbrakk>rs)\<close>
  by auto

lemma eval_binop: \<open>\<lbrakk>v\<^sub>1 = (\<lbrakk>e\<^sub>1\<rbrakk>rs); v\<^sub>2 = (\<lbrakk>e\<^sub>2\<rbrakk>rs); v = f v\<^sub>1 v\<^sub>2\<rbrakk> \<Longrightarrow> v = (\<lbrakk>BinOp e\<^sub>1 f e\<^sub>2\<rbrakk>rs)\<close>
  by auto

end
