theory StatementSemantics 
  imports Main AdversaryModel HoareTriple AIRSyntax
begin

section \<open>AIR Instructions\<close>

subsection \<open>Statement Semantics\<close>

subsubsection \<open>RegisterUpdate\<close>

text \<open>Assigns the value v to the Register r adding just the program address to the trace of accessed
 addresses omega. After the update the program counter will be incremented and the next Instruction 
chosen. This will only modify the program state at the current speculative level.\<close>

definition 
  trans_register_update :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Reg \<Rightarrow> Exp \<Rightarrow> MachineStateTransformer\<close>
where
  \<open>trans_register_update \<rho>' r e \<equiv> \<lambda>s. 
    s\<lparr>
      \<Delta> := \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s),
      pc := pc s ((n s) \<mapsto> next_\<rho> \<rho>'),
      \<omega> := \<omega> s @ [(\<rho>', None)]
    \<rparr>
  \<close>

subsubsection \<open>Load\<close>

text \<open>Updates the Register r with the value stored at address a adding the program and memory
addresses to the trace of accessed addresses omega. After the update the program counter will be 
incremented and the next Instruction chosen. This will only modify the program state at the current
speculative level.\<close>

definition trans_load :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Reg \<Rightarrow> Exp \<Rightarrow> MachineStateTransformer\<close> where
"trans_load \<rho>' r e \<equiv> \<lambda>s.
  let a = \<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s in
    s\<lparr>
      \<Delta> := \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s a),
      pc := pc s (n s \<mapsto> next_\<rho> \<rho>'),
      \<omega> := \<omega> s @ [(\<rho>', Some a)]
    \<rparr>
"

subsubsection \<open>Store\<close>

text \<open>Updates the memory address a with the value v adding the program and memory addresses to the 
trace of accessed addresses omega. After the update the program counter will be  incremented and the 
next Instruction chosen. This will only modify the program state at the current speculative level.\<close>

definition trans_store :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Exp \<Rightarrow> Exp \<Rightarrow> MachineStateTransformer\<close> where
"trans_store \<rho>' e\<^sub>1 e\<^sub>2 \<equiv> \<lambda>s.
  let a = \<lbrakk>e\<^sub>1\<rbrakk>\<Delta>\<^sub>n s in 
    let v = \<lbrakk>e\<^sub>2\<rbrakk>\<Delta>\<^sub>n s in
      s\<lparr>
        \<mu> := \<mu>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s a v,
        pc := pc s (n s \<mapsto> next_\<rho> \<rho>'),
        \<omega> := \<omega> s @ [(\<rho>', Some a)]
      \<rparr>
"

subsubsection \<open>T-Pred\<close>

text \<open>A correct prediction on a branch statement that evaluates true. The program counter
will advance to the specified program address and the branch predictor will be updated with the 
evaluation.\<close>

definition trans_tpred :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where
"trans_tpred \<rho>' c \<equiv> \<lambda>s.
  s\<lparr>
    \<beta> := update (n s) (\<rho>') False (\<beta> s),
    pc := pc s (n s \<mapsto> c),
    \<omega> := \<omega> s @ [(\<rho>', None)]
   \<rparr>
"

subsubsection \<open>T-Mispred\<close>

text \<open>An incorrect prediction on a branch statement that evaluates true. This triggers 
miss-speculation (increments n), jumping the program counter to the address specified (c) and 
updating the branch predictor.\<close>

definition trans_tmispred :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where  
"trans_tmispred \<rho>' c \<equiv> \<lambda>s.
  let n' = Suc (n s) in 
    s\<lparr>
      \<Delta> := \<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s,
      \<mu> := \<mu>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s,
      \<beta> := update (n s) (\<rho>') True (\<beta> s),
      pc := pc s((n s) \<mapsto> c, n' \<mapsto> next_\<rho> \<rho>'),
      n := n',
      \<omega> := \<omega> s @ [(\<rho>', None)]
     \<rparr>
"

subsubsection \<open>NT-Pred\<close>

text \<open>A correct prediction on a branch statement that evaluates false. The program counter will move
to the next Instruction and update the branch predictor.\<close>

definition trans_ntpred :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where 
"trans_ntpred \<rho>' \<equiv> \<lambda>s.
  s\<lparr>
    \<beta> := update (n s) (\<rho>') False (\<beta> s),
    pc := pc s(n s \<mapsto> next_\<rho> \<rho>'),
    \<omega> := \<omega> s @ [(\<rho>', None)]
   \<rparr>
"

subsubsection \<open>NT-Mispred\<close>

text \<open>An incorrect prediction on a branch statement that evaluates false. This triggers 
miss-speculation (increments n), increments the program counter to point to the next Instruction and
updates the branch predictor.\<close>

definition trans_ntmispred :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where 
"trans_ntmispred \<rho>' c \<equiv> \<lambda>s.
  let n' = Suc (n s) in 
    s\<lparr>
      \<Delta> := \<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s,
      \<mu> := \<mu>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s,
      \<beta> := update (n s) (\<rho>') True (\<beta> s),
      \<omega> := \<omega> s @ [(\<rho>', None)],
      n := n',
      pc := pc s (n s \<mapsto> next_\<rho> \<rho>', n' \<mapsto> c)
     \<rparr>
"

subsubsection \<open>Goto\<close>

text \<open>Advances the program counter to the given Instruction address, updating the branch predictor
in the process.\<close>

definition trans_goto :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where
"trans_goto \<rho>' c \<equiv> \<lambda>s.
  s\<lparr>
    \<beta> := update (n s) (\<rho>') False (\<beta> s),
    pc := pc s (n s \<mapsto> c),
    \<omega> := \<omega> s @ [(\<rho>', None)]
   \<rparr>
"

subsubsection \<open>SpecFence\<close>

text \<open>Immediately resolves all miss-predictions, effectively terminating speculative execution.\<close>

definition trans_specfence :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where
 "trans_specfence \<rho>' \<equiv> 
  \<lambda>s. s
  \<lparr>
    n := 0, 
    \<Delta> := ((\<Delta> s) |` {0}), 
    \<mu> := ((\<mu> s) |` {0}), 
    pc := ((pc s(n s \<mapsto> next_\<rho> \<rho>')) |` {0}),
    \<omega> := \<omega> s @ [(\<rho>', None)]
  \<rparr>
"

lemma map_eq_all_keys: 
  assumes "\<forall>k. (m::(('a,'b) map)) k = m' k"
    shows "m = m'"
  using assms by blast

lemma specfence_ns: 
  assumes "\<not>speculating s"
      and "wfs s"
      and "s' = trans_specfence \<rho>' s"
    shows "s' = s (|pc := pc s(n s \<mapsto> next_\<rho> \<rho>'), \<omega> := \<omega> s @ [(\<rho>', None)]|)"
proof -
  have ns_\<Delta>_eq: "\<Delta> s' 0 = \<Delta> s 0"
    using assms by (auto simp add: speculating_def trans_specfence_def)
  have s_\<Delta>_eq: "\<forall>n. n \<noteq> 0 \<longrightarrow> \<Delta> s' n = \<Delta> s n"
    using assms by (auto simp add: speculating_def trans_specfence_def wfs_def wfn_def wf\<Delta>_def)
  have \<Delta>_eq: "\<Delta> s' = \<Delta> s"
    using ns_\<Delta>_eq s_\<Delta>_eq map_eq_all_keys by metis

  have ns_\<mu>_eq: "\<mu> s' 0 = \<mu> s 0"
    using assms by (auto simp add: speculating_def trans_specfence_def)
  have s_\<mu>_eq: "\<forall>n. n \<noteq> 0 \<longrightarrow> \<mu> s' n = \<mu> s n"
    using assms by (auto simp add: speculating_def trans_specfence_def wfs_def wfn_def wf\<mu>_def)
  have \<mu>_eq: "\<mu> s' = \<mu> s"
    using ns_\<mu>_eq s_\<mu>_eq map_eq_all_keys by metis

  have pc_next: "pc s' = pc s(n s \<mapsto> next_\<rho> \<rho>')"
    using assms by (auto simp add: speculating_def trans_specfence_def wfs_def wfn_def wfpc_def)

  show ?thesis
    using pc_next \<Delta>_eq \<mu>_eq assms by (auto simp add: speculating_def trans_specfence_def)
qed

subsubsection \<open>Resolve\<close>

text \<open>Resolves the current branch miss-prediction, decreasing the speculation level and updating the 
branch predictor.\<close>

definition trans_resolve :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where
"trans_resolve \<rho>' \<equiv> 
  \<lambda>s. s\<lparr>
    n := (n s) - 1,
    \<beta> := update (n s) \<rho>' False (\<beta> s),
    \<Delta> := \<Delta>\<^sub>r\<^sub>e\<^sub>s s,
    \<mu> := \<mu>\<^sub>r\<^sub>e\<^sub>s s,
    pc := (pc s)((n s) := None)
   \<rparr>
"

subsubsection \<open>Noop\<close>

text \<open>Indicates this is a no operation (instruction).\<close>

definition trans_noop :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> MachineStateTransformer\<close> where
\<open>trans_noop \<rho>' \<equiv>
  \<lambda>s. s\<lparr>
    pc := pc s ((n s) \<mapsto> next_\<rho> \<rho>'),
    \<omega> := \<omega> s @ [(\<rho>', None)]
   \<rparr>
\<close>

lemmas trans_def = trans_resolve_def trans_specfence_def trans_goto_def trans_ntmispred_def 
                   trans_tmispred_def trans_tpred_def trans_ntpred_def trans_store_def 
                   trans_load_def trans_register_update_def trans_noop_def Let_def

subsection \<open>Branch Evaluation\<close>

fun trans_branch :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Addr\<^sub>\<Pi> \<Rightarrow> bool \<Rightarrow> bool \<Rightarrow> MachineStateTransformer\<close> where
"trans_branch \<rho>' c True  True  = trans_tmispred \<rho>' c" |
"trans_branch \<rho>' c True  False = trans_tpred \<rho>' c" |
"trans_branch \<rho>' c False True  = trans_ntmispred \<rho>' c" |
"trans_branch \<rho>' _ False False = trans_ntpred \<rho>'"

text \<open>Cases rule for trans_branch\<close>

lemma trans_branch_cases:
  assumes "s' = trans_branch \<rho>' c e m s"
  obtains 
    (TMispred) e c where
          "s' = trans_tmispred \<rho>' c s"
          "e"
          "m"
  | (TPred) e c where
          "s' = trans_tpred \<rho>' c s"
          "e"
          "\<not>m"
  | (NTMispred) e c where
          "s' = trans_ntmispred \<rho>' c s"
          "\<not>e"
          "m"
  | (NTPred) e c where
          "s' = trans_ntpred \<rho>' s"
          "\<not>e"
          "\<not>m"
  using assms by (metis (full_types) trans_branch.simps(1) trans_branch.simps(2) trans_branch.simps(3) trans_branch.simps(4))

lemma trans_branch_\<omega>_ne_preserved: 
  assumes "\<omega> s\<^sub>1 \<noteq> \<omega> s\<^sub>2"
      and "s\<^sub>1' = trans_branch p\<^sub>1 c\<^sub>1 e\<^sub>1 m\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = trans_branch p\<^sub>2 c\<^sub>2 e\<^sub>2 m\<^sub>2 s\<^sub>2"
    shows "\<omega> s\<^sub>1' \<noteq> \<omega> s\<^sub>2'"
  using assms(2) apply (induction rule: trans_branch_cases)
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 5
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 9
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 13
  using assms(3) apply (induction rule: trans_branch_cases)
  using assms(1) by (simp_all add: trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def)

lemma trans_branch_length_\<omega>_preserved: 
  assumes "length (\<omega> s\<^sub>1) = length (\<omega> s\<^sub>2)"
      and "s\<^sub>1' = trans_branch p\<^sub>1 c\<^sub>1 e\<^sub>1 m\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = trans_branch p\<^sub>2 c\<^sub>2 e\<^sub>2 m\<^sub>2 s\<^sub>2"
    shows "length (\<omega> s\<^sub>1') = length (\<omega> s\<^sub>2')"
  using assms(2) apply (induction rule: trans_branch_cases)
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 5
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 9
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 13
  using assms(3) apply (induction rule: trans_branch_cases)
  using assms(1) by (simp_all add: trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def)

lemma trans_branch_length_\<beta>\<^sub>\<omega>_preserved: 
  assumes "length (\<beta>\<^sub>\<omega> s\<^sub>1) = length (\<beta>\<^sub>\<omega> s\<^sub>2)"
      and "s\<^sub>1' = trans_branch p\<^sub>1 c\<^sub>1 e\<^sub>1 m\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = trans_branch p\<^sub>2 c\<^sub>2 e\<^sub>2 m\<^sub>2 s\<^sub>2"
    shows "length (\<beta>\<^sub>\<omega> s\<^sub>1') = length (\<beta>\<^sub>\<omega> s\<^sub>2')"
  using assms(2) apply (induction rule: trans_branch_cases)
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 5
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 9
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 13
  using assms(3) apply (induction rule: trans_branch_cases)
  using assms(1) by (simp_all add: trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def update_def)

lemma trans_branch_\<beta>_ne_preserved: 
  assumes "\<beta>\<^sub>\<omega> s\<^sub>1 \<noteq> \<beta>\<^sub>\<omega> s\<^sub>2"
      and "s\<^sub>1' = trans_branch p\<^sub>1 c\<^sub>1 e\<^sub>1 m\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = trans_branch p\<^sub>2 c\<^sub>2 e\<^sub>2 m\<^sub>2 s\<^sub>2"
    shows "\<beta>\<^sub>\<omega> s\<^sub>1' \<noteq> \<beta>\<^sub>\<omega> s\<^sub>2'"
  using assms(2) apply (induction rule: trans_branch_cases)
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 5
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 9
  using assms(3) apply (induction rule: trans_branch_cases)
  prefer 13
  using assms(3) apply (induction rule: trans_branch_cases)
  using assms(1) by (simp_all add: trans_tmispred_def trans_ntpred_def trans_tpred_def 
                                   trans_ntmispred_def update_def)

subsection \<open>Instruction Evaluation\<close>

primrec eval_instr :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Instr \<Rightarrow> MachineStateTransformer\<close> where
"eval_instr \<rho>' (RegisterUpdate r e) = trans_register_update \<rho>' r e" |
"eval_instr \<rho>' (Load r e) = trans_load \<rho>' r e" |
"eval_instr \<rho>' (Store e\<^sub>1 e\<^sub>2) = trans_store \<rho>' e\<^sub>1 e\<^sub>2" |
"eval_instr \<rho>' (Branch e c) = (\<lambda>s. trans_branch \<rho>' c (\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s) (mispred (n s) (\<beta> s) (pc s)) s)" |
"eval_instr \<rho>' (Goto c) = trans_goto \<rho>' c" |
"eval_instr \<rho>' (SpecFence) = trans_specfence \<rho>'" |
"eval_instr \<rho>' (NoOp) = trans_noop \<rho>'"

text \<open>Cases rule for eval_instr that extents primitive induction on i to include the induction of 
      trans_branch\<close>

lemma eval_instr_cases:
  assumes "s' = eval_instr \<rho>' i s"
  obtains 
    (SpecFence) 
          "i = SpecFence" 
          "s' = trans_specfence \<rho>' s"
  | (TMispred) e c where
          "i = Branch e c"
          "s' = trans_tmispred \<rho>' c s"
          "(\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s)"
          "mispred (n s) (\<beta> s) (pc s)"
  | (TPred) e c where
          "i = Branch e c"
          "s' = trans_tpred \<rho>' c s"
          "(\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s)"
          "\<not> mispred (n s) (\<beta> s) (pc s)"
  | (NTMispred) e c where
          "i = Branch e c"
          "s' = trans_ntmispred \<rho>' c s"
          "\<not>(\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s)"
          "mispred (n s) (\<beta> s) (pc s)"
  | (NTPred) e c where
          "i = Branch e c"
          "s' = trans_ntpred \<rho>' s"
          "\<not>(\<lbrakk>e\<rbrakk>\<Delta>\<^sub>n s)"
          "\<not> mispred (n s) (\<beta> s) (pc s)"
  | (Goto) c where
          "i = Goto c"
          "s' = trans_goto \<rho>' c s"
  | (RegisterUpdate) r e where
          "i = RegisterUpdate r e"
          "s' = trans_register_update \<rho>' r e s"
  | (Load) r e where
          "i = Load r e"
          "s' = trans_load \<rho>' r e s"
  | (Store) e\<^sub>1 e\<^sub>2 where
          "i = Store e\<^sub>1 e\<^sub>2"
          "s' = trans_store \<rho>' e\<^sub>1 e\<^sub>2 s"
  | (Noop)
          "i = NoOp"
          "s' = trans_noop \<rho>' s"
  using assms apply (induction i)
  apply (metis eval_instr.simps(1))
  apply (metis eval_instr.simps(2))
  apply (metis eval_instr.simps(3))
  apply force
  apply (metis eval_instr.simps(5))  
  apply (metis eval_instr.simps(6))
  by (metis eval_instr.simps(7))


lemma eval_instr_\<omega>_ne_preserved: 
  assumes "\<omega> s\<^sub>1  \<noteq> \<omega> s\<^sub>2"
      and "s\<^sub>1' = eval_instr \<rho>'\<^sub>1 i\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = eval_instr \<rho>'\<^sub>2 i\<^sub>2 s\<^sub>2"
    shows "\<omega> s\<^sub>1' \<noteq> \<omega> s\<^sub>2'"
  using assms(2) apply (induction rule: eval_instr_cases)
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 11
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 21
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 31
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 41
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 51
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 61
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 71
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 81
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 91
  using assms(3) apply (induction rule: eval_instr_cases)
  using assms(1) by (simp_all add: Let_def trans_noop_def trans_store_def trans_load_def trans_specfence_def trans_goto_def trans_register_update_def trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def)

lemma eval_instr_length_\<omega>_preserved: 
  assumes "length (\<omega> s\<^sub>1) = length (\<omega> s\<^sub>2)"
      and "s\<^sub>1' = eval_instr \<rho>'\<^sub>1 i\<^sub>1 s\<^sub>1"
      and "s\<^sub>2' = eval_instr \<rho>'\<^sub>2 i\<^sub>2 s\<^sub>2"
    shows "length (\<omega> s\<^sub>1') = length (\<omega> s\<^sub>2')"
  using assms(2) apply (induction rule: eval_instr_cases)
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 11
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 21
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 31
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 41
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 51
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 61
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 71
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 81
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 91
  using assms(3) apply (induction rule: eval_instr_cases)
  using assms(1) by (simp_all add: Let_def trans_noop_def trans_store_def trans_load_def trans_specfence_def trans_goto_def trans_register_update_def trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def)

lemma eval_instr_length_\<omega>_greater: 
  assumes "s' = eval_instr \<rho>' i s"
    shows "length (\<omega> s) < length (\<omega> s')"
  using assms apply (elim eval_instr_cases)
  by (simp_all add: Let_def trans_store_def trans_load_def trans_specfence_def trans_goto_def 
                    trans_register_update_def trans_tmispred_def trans_ntpred_def trans_tpred_def 
                    trans_ntmispred_def trans_noop_def)

lemma eval_instr_length_\<beta>_preserved_same_i: 
  assumes "length (\<beta>\<^sub>\<omega> s\<^sub>1) = length (\<beta>\<^sub>\<omega> s\<^sub>2)"
      and "s\<^sub>1' = eval_instr \<rho>'\<^sub>1 i s\<^sub>1"
      and "s\<^sub>2' = eval_instr \<rho>'\<^sub>2 i s\<^sub>2"
    shows "length (\<beta>\<^sub>\<omega> s\<^sub>1') = length (\<beta>\<^sub>\<omega> s\<^sub>2')"
  using assms(2) apply (induction rule: eval_instr_cases)
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 11
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 21
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 31
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 41
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 51
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 61
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 71
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 81
  using assms(3) apply (induction rule: eval_instr_cases)
  prefer 91
  using assms(3) apply (induction rule: eval_instr_cases)
  using assms(1) by (simp_all add: Let_def trans_noop_def trans_store_def trans_load_def trans_specfence_def trans_goto_def trans_register_update_def trans_tmispred_def trans_ntpred_def trans_tpred_def trans_ntmispred_def update_def)  

end
