theory Step2Semantics
  imports StepSemantics ProgramAnalysis
begin

type_synonym MachineTrace = "MachineState list"


type_synonym MachineState2 = "(MachineState \<times> MachineState)"
type_synonym MachineState2Predicate = "MachineState2 \<Rightarrow> bool"
type_synonym MachineState2Transformer = "MachineState2 \<Rightarrow> MachineState2"

subsection \<open>Well formedness of a state pair\<close>

abbreviation
  BranchPredictorTrace\<^sub>\<rho> :: "BranchPredictorTrace \<Rightarrow> Addr\<^sub>\<Pi> list"
where
 "BranchPredictorTrace\<^sub>\<rho> xs \<equiv> map (\<lambda>x. fst (snd x)) xs"

abbreviation "\<beta>\<^sub>\<omega>\<^sub>\<rho> s\<^sub>n\<^sub>s \<equiv> BranchPredictorTrace\<^sub>\<rho> (\<beta>\<^sub>\<omega> s\<^sub>n\<^sub>s)"

abbreviation "\<beta>\<^sub>\<omega>\<^sub>\<rho>_ns s\<^sub>n\<^sub>s \<equiv> BranchPredictorTrace\<^sub>\<rho> (filter (\<lambda>x. fst x = 0) (\<beta>\<^sub>\<omega> s\<^sub>n\<^sub>s))"

definition "wf\<beta>\<^sub>\<omega>2 s\<^sub>n\<^sub>s s\<^sub>s \<longleftrightarrow>
  \<beta>\<^sub>\<omega>\<^sub>\<rho> s\<^sub>n\<^sub>s = \<beta>\<^sub>\<omega>\<^sub>\<rho>_ns s\<^sub>s
"

text \<open>All elements in the first list must appear in order in the second list with speculative gaps.\<close>

fun ordered_sublist :: "AddressTrace \<Rightarrow> AddressTrace \<Rightarrow> bool" where
"ordered_sublist [] ys = True" |
"ordered_sublist xs [] = False" |
"ordered_sublist (x # xs) (y # ys) = (if x = y then ordered_sublist xs ys else ordered_sublist (x # xs) ys)"

lemma ordered_sublist_inducts[consumes 1, case_names Empty Match NotMatch]:
"ordered_sublist xs ys \<Longrightarrow>
(\<And>ys.
    ordered_sublist [] ys
  \<Longrightarrow> P [] ys) \<Longrightarrow>
(\<And>x y xs ys.
    x = y \<Longrightarrow>
    ordered_sublist xs ys \<Longrightarrow>
    P xs ys \<Longrightarrow>
    P (x # xs) (y # ys)) \<Longrightarrow>
(\<And>x y xs ys.
    x \<noteq> y \<Longrightarrow>
    ordered_sublist (x # xs) ys \<Longrightarrow>
    P (x # xs) ys \<Longrightarrow>
    P (x # xs) (y # ys)) \<Longrightarrow>
P xs ys"
  apply (induction rule: ordered_sublist.induct, simp_all)
  by (case_tac "x = y", simp_all)

lemma ordered_sublist_xs_empty_true: "ordered_sublist [] ys"
  by auto

lemma ordered_sublist_ys_empty_false: "xs \<noteq> [] \<Longrightarrow> \<not>ordered_sublist xs []"
  using ordered_sublist.elims(2) by blast

lemma ordered_sublist_ys_empty_xs_empty: "ordered_sublist xs [] \<Longrightarrow> xs = []"
  by (induct xs, simp_all)

lemma ordered_sublist_ys_empty_xs_empty2: "ordered_sublist (x # xs) [] \<Longrightarrow> False"
  by simp

lemma ordered_sublist_single: "ordered_sublist [x] [x]"
  by simp

lemma ordered_sublist_single_start: "ordered_sublist [x] (x # xs)"
  by simp

lemma ordered_sublist_single_end: "ordered_sublist [x] (xs @ [x])"
  by (induct xs, simp_all)

lemma ordered_sublist_equal: "ordered_sublist xs xs"
  by (induct xs, simp_all)


lemma ordered_sublist_length_ys_greater: "ordered_sublist xs ys \<Longrightarrow> length xs \<le> length ys"
  by (induct rule: ordered_sublist_inducts, simp_all)

lemma ordered_subset_length_ys_greater: "ordered_sublist xs ys \<Longrightarrow> set xs \<subseteq> set ys"
  by (induct rule: ordered_sublist_inducts, simp_all) auto

lemma ordered_sublist_ys_trivial: 
  assumes "ordered_sublist xs ys"
    shows "ordered_sublist xs (ys @ [y])"
  using assms by (induct rule: ordered_sublist_inducts, simp_all)

lemma ordered_sublist_add_equal: 
  assumes "ordered_sublist xs ys"
    shows "ordered_sublist (xs @ [x]) (ys @ [x])"
  using assms apply (induct rule: ordered_sublist_inducts, simp_all)
  by (simp add: ordered_sublist_single_end)


definition "wf\<omega>2 s\<^sub>n\<^sub>s s\<^sub>s \<longleftrightarrow>
  ordered_sublist (\<omega> s\<^sub>n\<^sub>s) (\<omega> s\<^sub>s)
"

definition "wfs2 s\<^sub>n\<^sub>s s\<^sub>s \<longleftrightarrow>
  pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s \<and>
  \<Delta>\<^sub>0 s\<^sub>n\<^sub>s = \<Delta>\<^sub>0 s\<^sub>s \<and>
  \<mu>\<^sub>0 s\<^sub>n\<^sub>s = \<mu>\<^sub>0 s\<^sub>s \<and>
  wfs s\<^sub>n\<^sub>s \<and> wfs s\<^sub>s \<and>
  \<Pi> s\<^sub>n\<^sub>s = \<Pi> s\<^sub>s \<and>
  \<U>wr s\<^sub>n\<^sub>s = \<U>wr s\<^sub>s \<and>
  \<U>rd s\<^sub>n\<^sub>s = \<U>rd s\<^sub>s \<and>
  \<S>\<^sub>\<T> s\<^sub>n\<^sub>s = \<S>\<^sub>\<T> s\<^sub>s \<and>
  \<E>\<P> s\<^sub>n\<^sub>s = \<E>\<P> s\<^sub>s \<and>
  \<T>\<^sub>\<rho> s\<^sub>n\<^sub>s = \<T>\<^sub>\<rho> s\<^sub>s \<and>
  wf\<omega>2 s\<^sub>n\<^sub>s s\<^sub>s \<and>
  wf\<beta>\<^sub>\<omega>2 s\<^sub>n\<^sub>s s\<^sub>s \<and>
  \<not>\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s \<and> \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"

definition "wfs2_ns s\<^sub>n\<^sub>s s\<^sub>s \<longleftrightarrow> 
  \<omega> s\<^sub>n\<^sub>s = \<omega> s\<^sub>s \<and> 
  \<beta>\<^sub>\<omega> s\<^sub>n\<^sub>s = \<beta>\<^sub>\<omega> s\<^sub>s
"

lemma wfs2_s\<^sub>n\<^sub>s_ns: 
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<not>speculating s\<^sub>n\<^sub>s"
  using assms by (simp add: wfs2_def wfs_def wfn_def speculating_def)

lemma wfs2_s\<^sub>n\<^sub>s_no_mispred:
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<not>mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
  using assms by (simp add: wfs2_def \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred)

(*lemma wfs2_s\<^sub>n\<^sub>s_no_resolve:
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
  sorry*)

text \<open>Constraints for step2\<close>

lemma wfs2_no_halt_\<iota>\<^sub>0_same:
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>halting (pc\<^sub>0 s\<^sub>n\<^sub>s) (\<Pi> s\<^sub>n\<^sub>s) \<and> \<not>halting (pc\<^sub>0 s\<^sub>s) (\<Pi> s\<^sub>s)"
    shows "\<iota>\<^sub>0 s\<^sub>n\<^sub>s = \<iota>\<^sub>0 s\<^sub>s"
  using assms by (simp add: wfs2_def)



lemma wfs2_ns_\<iota>_same:
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<not>halting (pc\<^sub>0 s\<^sub>n\<^sub>s) (\<Pi> s\<^sub>n\<^sub>s) \<and> \<not>halting (pc\<^sub>0 s\<^sub>s) (\<Pi> s\<^sub>s)"
    shows "\<iota> s\<^sub>n\<^sub>s = \<iota> s\<^sub>s"
  by (metis \<iota>\<^sub>0_\<iota>_no_spec assms(1) assms(2) assms(3) wfs2_no_halt_\<iota>\<^sub>0_same wfs2_s\<^sub>n\<^sub>s_ns)

lemma wfs2_ns_\<rho>_equal: 
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
    shows "\<rho> s\<^sub>n\<^sub>s = \<rho> s\<^sub>s"
  using assms apply (simp add: wfs2_def speculating_def)
  by (metis assms(1) pc\<^sub>0_pc_no_spec wfs2_s\<^sub>n\<^sub>s_ns)

lemma wfs2_eval_equiv:
  assumes "\<not>speculating s\<^sub>s"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
  using assms eval_exp_same_\<Delta>\<^sub>0_equiv wfs2_s\<^sub>n\<^sub>s_ns wfs2_def by blast


subsection \<open>Step 2 function\<close>

fun step\<^sub>2 :: \<open>MachineState2Transformer\<close> where
\<open>step\<^sub>2 (s\<^sub>n\<^sub>s, s\<^sub>s) = (if speculating s\<^sub>s then s\<^sub>n\<^sub>s else step s\<^sub>n\<^sub>s, step s\<^sub>s)\<close>

subsection \<open>TPOD step constraints\<close>

definition \<open>tpod_step\<^sub>2 s\<^sub>n\<^sub>s s\<^sub>s \<longleftrightarrow> tpod_step s\<^sub>n\<^sub>s \<and> tpod_step s\<^sub>s \<and> \<not> speculating s\<^sub>n\<^sub>s \<and> 
  ((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s \<and> ((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s\<close>


text \<open>A valid step of two states with s2 unconstrained and s1 not able to speculate\<close>

fun valid_step2 :: "MachineState2 \<Rightarrow> MachineState2 \<Rightarrow> bool" (infixl "\<leadsto>\<^sub>2" 101) where
\<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s') = (tpod_step\<^sub>2 s\<^sub>n\<^sub>s s\<^sub>s \<and> tpod_step\<^sub>2 s\<^sub>n\<^sub>s' s\<^sub>s' \<and> (s\<^sub>n\<^sub>s', s\<^sub>s') = step\<^sub>2 (s\<^sub>n\<^sub>s, s\<^sub>s) 
  \<and> conformantEntrypoint s\<^sub>n\<^sub>s s\<^sub>n\<^sub>s' \<and> conformantEntrypoint s\<^sub>s s\<^sub>s')\<close>

lemma valid_step2_s\<^sub>s_always_step:
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
    shows \<open>s\<^sub>s \<leadsto> s\<^sub>s'\<close>
  using assms valid_step2.simps tpod_step\<^sub>2_def valid_step_def by auto

lemma valid_step2_s\<^sub>n\<^sub>s_ns_step:
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
      and \<open>\<not>speculating s\<^sub>s\<close>
    shows \<open>s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'\<close>
  using assms valid_step2.simps tpod_step\<^sub>2_def valid_step_def by auto

lemma valid_step2_s\<^sub>n\<^sub>s_s_equal:
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
      and \<open>speculating s\<^sub>s\<close>
    shows \<open>s\<^sub>n\<^sub>s = s\<^sub>n\<^sub>s'\<close>
  using assms valid_step2.simps tpod_step\<^sub>2_def valid_step_def by auto

lemma valid_step2_step:
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
    shows \<open>s\<^sub>s \<leadsto> s\<^sub>s' \<and> (\<not>speculating s\<^sub>s \<longrightarrow> s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s') \<and> (speculating s\<^sub>s \<longrightarrow> s\<^sub>n\<^sub>s = s\<^sub>n\<^sub>s')\<close>
  using assms valid_step2_s\<^sub>n\<^sub>s_ns_step valid_step2_s\<^sub>s_always_step by auto


(*
  ((op\<^sub>\<L> s\<^sub>n\<^sub>s')::low_operation) = op\<^sub>\<L> s\<^sub>s' \<and> ((op\<^sub>\<H> s\<^sub>n\<^sub>s')::high_operation) = op\<^sub>\<H> s\<^sub>s' \<and>
  s\<^sub>s \<leadsto> s\<^sub>s' \<and>
  (if speculating s\<^sub>s then s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s else s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s')
*)

lemma step2_nospec_pc\<^sub>0_NeverHalts:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
    shows "\<not>halting (pc\<^sub>0 s\<^sub>n\<^sub>s) (\<Pi> s\<^sub>n\<^sub>s) \<and> \<not>halting (pc\<^sub>0 s\<^sub>s) (\<Pi> s\<^sub>s)"
  using assms no_valid_transition_nospec_adversary_halt wfs2_s\<^sub>n\<^sub>s_ns valid_step2_step by blast

lemma step2_ns_\<iota>_same:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
    shows "\<iota> s\<^sub>n\<^sub>s = \<iota> s\<^sub>s"
  using assms step2_nospec_pc\<^sub>0_NeverHalts wfs2_ns_\<iota>_same by auto

lemma step2_spec_s\<^sub>n\<^sub>s_stutters: 
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "speculating s\<^sub>s"
      and "P s\<^sub>n\<^sub>s"
    shows "P s\<^sub>n\<^sub>s'"
  using assms by simp

lemma step2_spec_retains_lo_hi: 
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "speculating s\<^sub>s"
    shows "(((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s' \<and> ((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s') \<and> (((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s' \<and> ((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s')"
  using assms by (auto simp add: wfs2_def tpod_step\<^sub>2_def)
    
lemma step2_spec_retains_lo_s_s': 
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "speculating s\<^sub>s"
    shows "(op\<^sub>\<L> s\<^sub>n\<^sub>s \<noteq> \<bottom> \<and> op\<^sub>\<L> s\<^sub>s \<noteq> \<bottom>) \<longleftrightarrow> (op\<^sub>\<L> s\<^sub>n\<^sub>s' \<noteq> \<bottom> \<and> op\<^sub>\<L> s\<^sub>s' \<noteq> \<bottom>)"
  using assms(1) assms(2) assms(3) wfs2_def tpod_step\<^sub>2_def by fastforce

lemma step2_spec_retains_hi_s_s': 
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "speculating s\<^sub>s"
    shows "(inst\<^sub>\<T> s\<^sub>n\<^sub>s \<noteq> None \<and> inst\<^sub>\<T> s\<^sub>s \<noteq> None) \<longleftrightarrow> (inst\<^sub>\<T> s\<^sub>n\<^sub>s' \<noteq> None \<and> inst\<^sub>\<T> s\<^sub>s' \<noteq> None)"
  using assms(1) assms(2) assms(3) by (metis Pair_inject op\<^sub>\<H>\<^sub>s.simps op\<^sub>\<H>_sub_def step2_spec_retains_lo_hi)

lemma step2_ns_s\<^sub>s'_no_mispred:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<not>speculating s\<^sub>s'"
      and "\<iota> s\<^sub>s = Branch e c"
      and "\<not>resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
    shows "\<not>mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
  using assms mispred_speculating valid_step2_s\<^sub>s_always_step by blast

lemma wfs2_ns_eval_equiv_F:
  assumes "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
    shows "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) \<longleftrightarrow> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
  using assms(1) assms(2) wfs2_eval_equiv by auto


subsubsection \<open>Case simplification for the valid transition\<close>

text \<open>Cases for step2\<close>

lemma valid_step2_cases [consumes 2]:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
obtains
    (NS_SpecFence) 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = SpecFence"
          "\<iota> s\<^sub>s = SpecFence"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_TPred) e c where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Branch e c"
          "\<iota> s\<^sub>s = Branch e c"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)"
          "\<not> mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_NTPred) e c where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Branch e c"
          "\<iota> s\<^sub>s = Branch e c"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
          "\<not> mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_Goto) c where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Goto c"
          "\<iota> s\<^sub>s = Goto c"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_RegisterUpdate) r e where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e"
          "\<iota> s\<^sub>s = RegisterUpdate r e"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_Load) r e where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Load r e"
          "\<iota> s\<^sub>s = Load r e"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_Store) e\<^sub>1 e\<^sub>2 where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2"
          "\<iota> s\<^sub>s = Store e\<^sub>1 e\<^sub>2"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_TMispred) e c where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Branch e c"
          "\<iota> s\<^sub>s = Branch e c"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)"
          "\<not>mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_NTMispred) e c where 
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = Branch e c"
          "\<iota> s\<^sub>s = Branch e c"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
          "\<not>mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (NS_NoOp)
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>n\<^sub>s = NoOp"
          "\<iota> s\<^sub>s = NoOp"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s'"
          "\<not>speculating s\<^sub>s'"
          "\<not>speculating s\<^sub>s"
  | (S_SpecFence)
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = SpecFence"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "\<not>speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_RegisterUpdate) r e where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = RegisterUpdate r e"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_Load) r e where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Load r e"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_Store) e\<^sub>1 e\<^sub>2 where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Store e\<^sub>1 e\<^sub>2"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_TPred) e c where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Branch e c"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)"
          "\<not>mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_TMispred) e c where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Branch e c"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)"
          "mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_NTPred) e c where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Branch e c"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
          "\<not> mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_NTMispred) e c where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Branch e c"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
          "mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_Goto) c where
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = Goto c"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_NoOp)
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "\<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
          "\<iota> s\<^sub>s = NoOp"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_TrueResolve)
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "n s\<^sub>s = 1"
          "\<not>speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
  | (S_Resolve_S)
          "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
          "resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
          "s\<^sub>s \<leadsto> s\<^sub>s'"
          "s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s"
          "n s\<^sub>s > 1"
          "speculating s\<^sub>s'"
          "speculating s\<^sub>s"
          "((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
          "((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s'"
          "((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s'"
proof -
  have conds: "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s) \<and> s\<^sub>s \<leadsto> s\<^sub>s'"
    apply (rule conjI)
    using assms(1) tpod_step\<^sub>2_def tpod_step_def valid_step2.simps apply blast
    using assms(1) valid_step2_s\<^sub>s_always_step by blast
  show ?thesis
  proof (cases "speculating s\<^sub>s")
    case True
    have s_conds: "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s) \<and> s\<^sub>s \<leadsto> s\<^sub>s' \<and> s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> 
                  ((op\<^sub>\<H> s\<^sub>n\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>n\<^sub>s' \<and> ((op\<^sub>\<H> s\<^sub>s)::high_operation) = op\<^sub>\<H> s\<^sub>s' \<and> 
                  ((op\<^sub>\<L> s\<^sub>n\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>n\<^sub>s' \<and> ((op\<^sub>\<L> s\<^sub>s)::low_operation) = op\<^sub>\<L> s\<^sub>s'" 
      apply (meson True assms(1-2) conds step2_spec_retains_lo_hi valid_step2.simps)
      using True assms(1) by fastforce
    have elims: "\<iota> s\<^sub>s \<noteq> SpecFence \<and> \<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s) \<longrightarrow> speculating s\<^sub>s'"
      using True by (metis const_n not_speculating_no_mispred s_conds speculating_def)
    show ?thesis
      using True s_conds apply auto
      apply (elim valid_step_cases)
      using s_conds S_Resolve_S            S_TrueResolve apply (metis (full_types) cancel_comm_monoid_add_class.diff_cancel dec_n_resolve less_one n_1_resolve_nospec nat_neq_iff speculating_def)
      using s_conds S_SpecFence            speculating_no_specfence apply blast
      using s_conds S_TMispred             not_speculating_no_mispred apply blast
      using s_conds S_TPred                const_n speculating_def apply fastforce      
      using s_conds S_NTMispred            not_speculating_no_mispred apply blast
      using s_conds S_NTPred               const_n speculating_def apply fastforce
      using s_conds elims S_Goto           S_SpecFence apply blast
      using s_conds elims S_RegisterUpdate S_SpecFence apply blast
      using s_conds elims S_Load           S_SpecFence apply blast
      using s_conds elims S_Store          S_SpecFence apply blast
      using s_conds elims S_NoOp           S_SpecFence by blast
  next
    case False
    have ns_conds: "\<not> resolve (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s) \<and> \<not> resolve (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s) \<and> \<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>n\<^sub>s \<and> \<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s \<and> \<iota> s\<^sub>n\<^sub>s = \<iota> s\<^sub>s \<and> s\<^sub>s \<leadsto> s\<^sub>s' \<and> s\<^sub>n\<^sub>s \<leadsto> s\<^sub>n\<^sub>s' \<and> \<not>speculating s\<^sub>s \<and> \<not>mispred (n s\<^sub>n\<^sub>s) (\<beta> s\<^sub>n\<^sub>s) (pc s\<^sub>n\<^sub>s)"
      using False conds assms(1, 2) apply (auto simp add: resolve_must_be_speculating no_valid_transition_halt step2_ns_\<iota>_same wfs2_s\<^sub>n\<^sub>s_no_mispred)      
      using always_resolve_on_halt no_valid_transition_halt assms(1) valid_step2_s\<^sub>n\<^sub>s_ns_step by blast+
    have ns_ns_elims: "\<not>mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s) \<longrightarrow> \<not> speculating s\<^sub>s'"
      using ns_conds speculating_mispred by auto
    have ns_s_elims: "\<And>e c. mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s) \<and> \<iota> s\<^sub>s = Branch e c \<longrightarrow> speculating s\<^sub>s'"
      using not_speculating_no_mispred ns_conds by auto
    show ?thesis
      using assms ns_conds apply auto
      apply (erule valid_step_cases)
      using ns_conds apply blast
      apply (metis NS_SpecFence less_not_refl n_0_fence ns_conds speculating_def)
      using NS_TMispred ns_conds ns_s_elims wfs2_eval_equiv apply fastforce
      using NS_TPred ns_conds ns_ns_elims wfs2_eval_equiv apply metis
      apply (metis NS_NTMispred ns_conds ns_s_elims wfs2_eval_equiv)
      using NS_NTPred ns_conds ns_ns_elims wfs2_eval_equiv apply metis
      using NS_Goto mispred_ns ns_conds ns_ns_elims apply force
      using NS_RegisterUpdate mispred_ns ns_conds ns_ns_elims apply force
      using NS_Load mispred_ns ns_conds ns_ns_elims apply force
      using NS_Store mispred_ns ns_conds ns_ns_elims apply force
      using NS_NoOp mispred_ns ns_conds ns_ns_elims by force
  qed
qed


section \<open>Transitions\<close>

subsubsection \<open>Trace of instructions and memory accessed (\<omega>)\<close>

text \<open>Instructions which do not access memory only add their program address to the trace.\<close>

lemma step2_\<omega>_\<rho>_None:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e \<or> \<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c \<or> \<iota> s\<^sub>n\<^sub>s = SpecFence \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
    shows "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>n\<^sub>s @ [(\<rho> s\<^sub>n\<^sub>s, None)] \<and> \<omega> s\<^sub>s' = \<omega> s\<^sub>s @ [(\<rho> s\<^sub>s, None)]"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  by (simp_all add: \<omega>_\<rho>_None)

lemma step2_\<omega>_\<rho>_None_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e \<or> \<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
      and "\<omega> s\<^sub>n\<^sub>s = \<omega> s\<^sub>s"
    shows "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>s'"
proof -
  have "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>n\<^sub>s @ [(\<rho> s\<^sub>n\<^sub>s, None)] \<and> \<omega> s\<^sub>s' = \<omega> s\<^sub>s @ [(\<rho> s\<^sub>s, None)]"
    using step2_\<omega>_\<rho>_None assms by blast
  moreover have "pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s"
    using assms(2) by (simp add: wfs2_def)
  ultimately show ?thesis
    by (simp add: assms wfs2_ns_\<rho>_equal)
qed

text \<open>Instructions that access memory add this to the trace along with their program address.\<close>

lemma step2_\<omega>_\<rho>_a:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Load r e \<or> \<iota> s\<^sub>n\<^sub>s = Store e e\<^sub>2"
      and "a = (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))"
    shows "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>n\<^sub>s @ [(\<rho> s\<^sub>n\<^sub>s, Some a)] \<and> \<omega> s\<^sub>s' = \<omega> s\<^sub>s @ [(\<rho> s\<^sub>s, Some a)]"
proof -
  have equal_eval: "(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) = (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"  
    using eval_exp_same_\<Delta>\<^sub>0_equiv wfs2_def apply (simp only: wfs2_def)
    using assms(2-3) by (smt (z3) eval_exp_same_\<Delta>\<^sub>0_equiv wfs2_def wfs2_s\<^sub>n\<^sub>s_ns)
  show ?thesis 
    using assms apply (induct rule: valid_step2_cases, simp_all)
    apply (metis (no_types, hide_lams) \<omega>_\<rho>_a equal_eval eval_exp_const.simps)
    by (metis \<omega>_\<rho>_a equal_eval eval_exp_const.simps)
qed



lemma step2_\<omega>_\<rho>_a_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<not>speculating s\<^sub>s"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Load r e \<or> \<iota> s\<^sub>n\<^sub>s = Store e e\<^sub>2"
      and "(a::64word) = (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))"
      and "\<omega> s\<^sub>n\<^sub>s = \<omega> s\<^sub>s"
    shows "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>s'"
proof -
  have "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>n\<^sub>s @ [(\<rho> s\<^sub>n\<^sub>s, Some a)] \<and> \<omega> s\<^sub>s' = \<omega> s\<^sub>s @ [(\<rho> s\<^sub>s, Some a)]"
    using assms step2_\<omega>_\<rho>_a by blast
  moreover have "pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s"
    using assms(3) by (simp add: wfs2_def)
  ultimately show ?thesis (* lemma to help here + sledgehammer all \<omega> again *)
    by (simp add: assms wfs2_ns_\<rho>_equal)
qed

lemma step2_\<omega>_equal_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "wfs2_ns s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<omega> s\<^sub>n\<^sub>s' = \<omega> s\<^sub>s'"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  apply (simp add: \<omega>_\<rho>_None wfs2_ns_\<rho>_equal wfs2_ns_def)
  using assms(1) assms(4) step2_\<omega>_\<rho>_None_equal wfs2_ns_def apply blast+  
  using assms(1) assms(4) step2_\<omega>_\<rho>_a_equal wfs2_ns_def apply blast+
  using assms(1) assms(4) step2_\<omega>_\<rho>_None_equal wfs2_ns_def by blast+

lemma step2_\<omega>_subset:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "wf\<omega>2 s\<^sub>n\<^sub>s' s\<^sub>s'"
  using assms apply (induction rule: valid_step2_cases, simp_all)
  using assms(2) apply (simp_all add: wfs2_def wf\<omega>2_def const_\<omega> \<omega>_\<rho>_None \<omega>_\<rho>_a)
  using assms(2) ordered_sublist_add_equal wfs2_ns_\<rho>_equal apply presburger+
  apply (simp add: \<Delta>\<^sub>0_\<Delta>_no_spec ordered_sublist_add_equal pc\<^sub>0_pc_no_spec wfn_def wfs_def)
  apply (metis neq0_conv ordered_sublist_add_equal speculating_def wfn_def wfs_def)
  using assms(2) ordered_sublist_add_equal wfs2_ns_\<rho>_equal apply presburger+
  using ordered_sublist_ys_trivial by blast+

subsubsection \<open>Branch predictor (\<beta>)\<close>


text \<open>Non branching instructions do not change the branch predictor.\<close>


lemma step2_const_\<beta>_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s'"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e \<or> \<iota> s\<^sub>n\<^sub>s = Load r e \<or> \<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s\<^sub>n\<^sub>s = SpecFence \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
    shows "\<beta> s\<^sub>n\<^sub>s' = \<beta> s\<^sub>n\<^sub>s \<and> \<beta> s\<^sub>s' = \<beta> s\<^sub>s"
  using assms(1-5) apply (induct rule: valid_step2_cases, simp_all)
  by (simp_all add: const_\<beta> wfs2_def)

(*maybe not needed*)
lemma step2_const_\<beta>_ns_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<not>speculating s\<^sub>s'"
      and "\<not>speculating s\<^sub>s"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e \<or> \<iota> s\<^sub>n\<^sub>s = Load r e \<or> \<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s\<^sub>n\<^sub>s = SpecFence \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
      and "\<beta> s\<^sub>n\<^sub>s = \<beta> s\<^sub>s"
    shows "\<beta> s\<^sub>n\<^sub>s' = \<beta> s\<^sub>s'"
  using assms by (metis step2_const_\<beta>_ns)


text \<open>Branching instructions add accesses to the branch predictor.\<close>


lemma step2_update_\<beta>_False_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s'"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<longrightarrow> t_pred s\<^sub>n\<^sub>s e \<or> nt_pred s\<^sub>n\<^sub>s e"
    shows "\<beta> s\<^sub>n\<^sub>s' = update (n s\<^sub>n\<^sub>s) (\<rho> s\<^sub>n\<^sub>s) False (\<beta> s\<^sub>n\<^sub>s) \<and> \<beta> s\<^sub>s' = update (n s\<^sub>s) (\<rho> s\<^sub>s) False (\<beta> s\<^sub>s)"
  using assms(1-6) apply (induct rule: valid_step2_cases, simp_all)
  using assms(6) by (simp_all add: update_\<beta>_False)  


lemma step2_update_\<beta>_False_ns_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<not>speculating s\<^sub>s'"
      and "\<not>speculating s\<^sub>s"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<longrightarrow> t_pred s\<^sub>n\<^sub>s e \<or> nt_pred s\<^sub>n\<^sub>s e"
      and "\<beta> s\<^sub>n\<^sub>s = \<beta> s\<^sub>s"
    shows "\<beta> s\<^sub>n\<^sub>s' = \<beta> s\<^sub>s'"
proof - 
  have "\<beta> s\<^sub>n\<^sub>s' = update (n s\<^sub>n\<^sub>s) (\<rho> s\<^sub>n\<^sub>s) False (\<beta> s\<^sub>n\<^sub>s) \<and> \<beta> s\<^sub>s' = update (n s\<^sub>s) (\<rho> s\<^sub>s) False (\<beta> s\<^sub>s)"
    using assms step2_update_\<beta>_False_ns by blast
  moreover have "pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s"
    using assms(4) by (simp add: wfs2_def)
  ultimately show ?thesis
    by (metis assms(3) assms(4) assms(7) gr0I speculating_def wfs2_s\<^sub>n\<^sub>s_ns)
qed

lemma step2_\<beta>\<^sub>\<omega>_equal_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s'"
      and "\<not>speculating s\<^sub>s"
      and "wfs2_ns s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<beta>\<^sub>\<omega> s\<^sub>n\<^sub>s' = \<beta>\<^sub>\<omega> s\<^sub>s'"
proof -
  show ?thesis
    using assms(1-5) apply (induct rule: valid_step2_cases, simp_all)
    apply (simp_all add: wfs2_ns_def const_\<beta> update_\<beta>\<^sub>\<omega>_False halting_def)
    using speculating_def wfs2_ns_\<rho>_equal wfs2_s\<^sub>n\<^sub>s_ns by auto
qed

lemma step2_wf2\<beta>\<^sub>\<omega>:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "wf\<beta>\<^sub>\<omega>2 s\<^sub>n\<^sub>s' s\<^sub>s'"
  using assms apply (induction rule: valid_step2_cases, simp_all)
  using assms(2) apply (simp_all add: wfs2_def wf\<beta>\<^sub>\<omega>2_def const_\<beta>)
  using assms(2) wfs2_s\<^sub>n\<^sub>s_ns by (simp_all add: update_\<beta>\<^sub>\<omega>_True update_\<beta>\<^sub>\<omega>_False speculating_def)



subsubsection \<open>Adversary observable memory (\<mu>\<A>)\<close>

lemma step2_const_\<mu>\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s \<longrightarrow> (\<forall>e\<^sub>1 e\<^sub>2. \<iota> s\<^sub>n\<^sub>s \<noteq> Store e\<^sub>1 e\<^sub>2)"
    shows "\<mu>\<^sub>0 s\<^sub>n\<^sub>s' = \<mu>\<^sub>0 s\<^sub>n\<^sub>s \<and> \<mu>\<^sub>0 s\<^sub>s' = \<mu>\<^sub>0 s\<^sub>s"
  using assms apply auto
  using const_\<mu>\<^sub>0 assms(1) valid_step2_step apply blast
  using const_\<mu>\<^sub>0 assms(1) valid_step2_step apply blast
  using const_\<mu>\<^sub>0 assms(1) valid_step2_step apply blast
  by (metis assms(1) const_\<mu>\<^sub>0 step2_ns_\<iota>_same valid_step2_step)

lemma step2_store_\<mu>\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2"
    shows "\<mu>\<^sub>0 s\<^sub>n\<^sub>s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) (\<mu>\<^sub>0 s\<^sub>n\<^sub>s) \<and> 
           \<mu>\<^sub>0 s\<^sub>s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)) (\<mu>\<^sub>0 s\<^sub>s)"
  using assms \<Delta>\<^sub>0_\<Delta>_no_spec nonspec_store_\<mu>\<^sub>0 step2_ns_\<iota>_same wfs2_s\<^sub>n\<^sub>s_ns valid_step2_s\<^sub>n\<^sub>s_ns_step 
        valid_step2_s\<^sub>s_always_step by auto


lemma step2_store_\<mu>\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2"
    shows "\<mu>\<^sub>0 s\<^sub>n\<^sub>s' = \<mu>\<^sub>0 s\<^sub>s'"
proof - 
  have step2: "\<mu>\<^sub>0 s\<^sub>n\<^sub>s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) (\<mu>\<^sub>0 s\<^sub>n\<^sub>s) \<and> 
               \<mu>\<^sub>0 s\<^sub>s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)) (\<mu>\<^sub>0 s\<^sub>s)"
    using assms step2_store_\<mu>\<^sub>0 by blast
  have eval_e1: "(\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) = (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
    using assms(2-4) eval_exp_same_\<Delta>\<^sub>0_equiv apply (simp only: wfs2_def)
    using assms(2) wfs2_s\<^sub>n\<^sub>s_ns by blast
  have eval_e2: "(\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s)) = (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s))"
    using assms(2-4) eval_exp_same_\<Delta>\<^sub>0_equiv apply (simp only: wfs2_def)
    using assms(2) wfs2_s\<^sub>n\<^sub>s_ns by blast
  show ?thesis
    using assms(2-4) apply (simp add: step2 wfs2_def)
    by (metis eval_e1 eval_e2 eval_exp_const.simps)
qed


subsubsection \<open>Adversary observable registers (\<Delta>\<A>)\<close>


lemma step2_const_\<Delta>\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s \<longrightarrow> (\<forall>e r. \<iota> s\<^sub>n\<^sub>s \<noteq> Load r e \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> RegisterUpdate r e)"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>n\<^sub>s \<and> \<Delta>\<^sub>0 s\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s"
  apply (rule conjI)
  using assms valid_step2_step const_\<Delta>\<^sub>0 apply blast
  using assms valid_step2_step const_\<Delta>\<^sub>0 step2_ns_\<iota>_same by metis


lemma step2_load_\<Delta>\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Load r e"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>n\<^sub>s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>n\<^sub>s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))) 
         \<and> \<Delta>\<^sub>0 s\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)))"
  apply (rule conjI)
  using assms(2, 3) apply (auto simp add: wfs2_def) (* TODO lemma this for all delta *)
  using assms nonspec_load_\<Delta>\<^sub>0  step2_ns_\<iota>_same apply auto
  apply (metis assms(1) valid_step2_s\<^sub>n\<^sub>s_ns_step wfs2_s\<^sub>n\<^sub>s_ns)
  by (metis (no_types, hide_lams) assms(1) valid_step2_s\<^sub>s_always_step)

lemma step2_load_\<Delta>\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Load r e"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s'"
proof -
  have "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>n\<^sub>s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>n\<^sub>s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))) \<and> \<Delta>\<^sub>0 s\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s\<^sub>s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)))"
    using assms step2_load_\<Delta>\<^sub>0 by blast
  moreover have "\<mu>\<^sub>0 s\<^sub>n\<^sub>s = \<mu>\<^sub>0 s\<^sub>s \<and> wfs s\<^sub>n\<^sub>s \<and> wfs s\<^sub>s \<and> \<Delta>\<^sub>0 s\<^sub>n\<^sub>s = \<Delta>\<^sub>0 s\<^sub>s"
    using assms(2) by (simp add: wfs2_def)
  ultimately show ?thesis
    using eval_exp_same_\<Delta>\<^sub>0_equiv assms(2-4) apply (simp only: wfs2_def)
    using \<Delta>\<^sub>0_\<Delta>_no_spec assms(2) wfs2_s\<^sub>n\<^sub>s_ns by auto
qed

lemma step2_reg_update_\<Delta>\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>n\<^sub>s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))) \<and> \<Delta>\<^sub>0 s\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)))"
  apply (rule conjI)
  using assms(1-3) apply (induct rule: valid_step2_cases)
  using assms(2) assms(4) apply (simp_all add: wfs2_def)
  using assms(2) nonspec_reg_update_\<Delta>\<^sub>0 wfs2_s\<^sub>n\<^sub>s_ns apply auto[1]
  using assms(1-3) nonspec_reg_update_\<Delta>\<^sub>0 wfs2_ns_\<rho>_equal apply auto
  using assms(1) valid_step2_s\<^sub>s_always_step by presburger


lemma step2_reg_update_\<Delta>\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s'"
proof - 
  have "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>n\<^sub>s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>n\<^sub>s))) \<and> \<Delta>\<^sub>0 s\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s\<^sub>s)))"
    using assms(1-4) step2_reg_update_\<Delta>\<^sub>0 by blast
  moreover have "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s = \<Delta>\<^sub>0 s\<^sub>s \<and> wfs s\<^sub>n\<^sub>s \<and> wfs s\<^sub>s"
    using assms(2) by (simp add: wfs2_def)
  ultimately show ?thesis
    using eval_exp_same_\<Delta>\<^sub>0_equiv assms(2-4) apply (simp only: wfs2_def)
    by (smt (z3) \<Delta>\<^sub>0_\<Delta>_no_spec assms(2) wfs2_s\<^sub>n\<^sub>s_ns)
qed


subsubsection \<open>Adversary observable program counter (pc\<A>)\<close>


lemma step2_const_pc\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s \<longrightarrow> (\<forall>r e\<^sub>1 e\<^sub>2 c. \<iota> s\<^sub>n\<^sub>s \<noteq> Store e\<^sub>1 e\<^sub>2 \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> Load r e\<^sub>1 \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> RegisterUpdate r e\<^sub>1 \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> Branch e\<^sub>1 c \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> Goto c \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> SpecFence \<and> \<iota> s\<^sub>n\<^sub>s \<noteq> NoOp)"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = pc\<^sub>0 s\<^sub>n\<^sub>s \<and> pc\<^sub>0 s\<^sub>s' = pc\<^sub>0 s\<^sub>s"
  using assms(1-3) apply (induct rule: valid_step2_cases, simp_all)
  using assms(3) by (simp_all add: const_pc\<^sub>0 wfs2_def)


lemma step2_next_pc\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s\<^sub>n\<^sub>s = Load r e\<^sub>1 \<or> \<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s\<^sub>n\<^sub>s = Branch e\<^sub>1 c \<or> \<iota> s\<^sub>n\<^sub>s = SpecFence \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s\<^sub>n\<^sub>s e\<^sub>1 \<or> nt_mispred s\<^sub>n\<^sub>s e\<^sub>1)"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = next_\<rho> (\<rho> s\<^sub>n\<^sub>s) \<and> pc\<^sub>0 s\<^sub>s' = next_\<rho> (\<rho> s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  by (simp_all add: next_pc\<^sub>0 wfs2_s\<^sub>n\<^sub>s_ns)


lemma step2_next_pc\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s\<^sub>n\<^sub>s = Load r e\<^sub>1 \<or> \<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s\<^sub>n\<^sub>s = Branch e\<^sub>1 c \<or> \<iota> s\<^sub>n\<^sub>s = SpecFence \<or> \<iota> s\<^sub>n\<^sub>s = NoOp"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s\<^sub>n\<^sub>s e\<^sub>1 \<or> nt_mispred s\<^sub>n\<^sub>s e\<^sub>1)"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = pc\<^sub>0 s\<^sub>s'"
proof - 
  have "pc\<^sub>0 s\<^sub>n\<^sub>s' = next_\<rho> (\<rho> s\<^sub>n\<^sub>s) \<and> pc\<^sub>0 s\<^sub>s' = next_\<rho> (\<rho> s\<^sub>s)"
    using assms step2_next_pc\<^sub>0 by auto
  moreover have "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s = \<Delta>\<^sub>0 s\<^sub>s \<and> wfs s\<^sub>n\<^sub>s \<and> wfs s\<^sub>s \<and> pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s"
    using assms(2) by (simp add: wfs2_def)
  ultimately show ?thesis
    using assms(2) assms(3) wfs2_ns_\<rho>_equal by auto
qed


lemma step2_jmp_pc\<^sub>0:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<longrightarrow> (t_pred s\<^sub>n\<^sub>s e \<or> t_mispred s\<^sub>n\<^sub>s e)"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = c \<and> pc\<^sub>0 s\<^sub>s' = c"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  by (simp_all add: jmp_pc\<^sub>0 wfs2_s\<^sub>n\<^sub>s_ns)


lemma step2_jmp_pc\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<or> \<iota> s\<^sub>n\<^sub>s = Goto c"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c \<longrightarrow> (t_pred s\<^sub>n\<^sub>s e \<or> t_mispred s\<^sub>n\<^sub>s e)"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = pc\<^sub>0 s\<^sub>s'"
proof - 
  have "pc\<^sub>0 s\<^sub>n\<^sub>s' = c \<and> pc\<^sub>0 s\<^sub>s' = c"
    using assms step2_jmp_pc\<^sub>0 by blast
  then show ?thesis
    by auto
qed


lemma step2_\<T>\<^sub>\<rho>_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<T>\<^sub>\<rho> s\<^sub>n\<^sub>s = \<T>\<^sub>\<rho> s\<^sub>n\<^sub>s' \<and> \<T>\<^sub>\<rho> s\<^sub>s = \<T>\<^sub>\<rho> s\<^sub>s'"
  using const_\<T>\<^sub>\<rho> valid_step2_step by blast

lemma step2_\<E>\<P>_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<E>\<P> s\<^sub>n\<^sub>s = \<E>\<P> s\<^sub>n\<^sub>s' \<and> \<E>\<P> s\<^sub>s = \<E>\<P> s\<^sub>s'"
  using const_\<E>\<P> valid_step2_step by metis

lemma step2_\<S>\<^sub>\<T>_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<S>\<^sub>\<T> s\<^sub>n\<^sub>s = \<S>\<^sub>\<T> s\<^sub>n\<^sub>s' \<and> \<S>\<^sub>\<T> s\<^sub>s = \<S>\<^sub>\<T> s\<^sub>s'"
  using const_\<S>\<^sub>\<T> valid_step2_step by blast

lemma step2_\<U>rd_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<U>rd s\<^sub>n\<^sub>s = \<U>rd s\<^sub>n\<^sub>s' \<and> \<U>rd s\<^sub>s = \<U>rd s\<^sub>s'"
  using const_\<U>rd valid_step2_step by blast

lemma step2_\<U>wr_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<U>wr s\<^sub>n\<^sub>s = \<U>wr s\<^sub>n\<^sub>s' \<and> \<U>wr s\<^sub>s = \<U>wr s\<^sub>s'"
  using const_\<U>wr valid_step2_step by blast

lemma step2_\<Pi>_const:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> \<Longrightarrow> \<Pi> s\<^sub>n\<^sub>s = \<Pi> s\<^sub>n\<^sub>s' \<and> \<Pi> s\<^sub>s = \<Pi> s\<^sub>s'"
  using const_\<Pi> valid_step2_step by metis

subsubsection \<open>Invariants\<close>

lemma step2_wf: 
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs s\<^sub>n\<^sub>s \<and> wfs s\<^sub>s"
    shows "wfs s\<^sub>n\<^sub>s' \<and> wfs s\<^sub>s'"
  using assms apply simp 
  by (metis tpod_step\<^sub>2_def valid_step_def wfs_trans)


lemma step2_\<Pi>_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<Pi> s\<^sub>n\<^sub>s = \<Pi> s\<^sub>s"
    shows "\<Pi> s\<^sub>n\<^sub>s' = \<Pi> s\<^sub>s'"
  using assms by (metis step2_\<Pi>_const)


lemma step2_\<T>\<^sub>\<rho>_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<T>\<^sub>\<rho> s\<^sub>n\<^sub>s = \<T>\<^sub>\<rho> s\<^sub>s"
    shows "\<T>\<^sub>\<rho> s\<^sub>n\<^sub>s' = \<T>\<^sub>\<rho> s\<^sub>s'"
  using assms by (metis step2_\<T>\<^sub>\<rho>_const)


lemma step2_\<E>\<P>_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<E>\<P> s\<^sub>n\<^sub>s = \<E>\<P> s\<^sub>s"
    shows "\<E>\<P> s\<^sub>n\<^sub>s' = \<E>\<P> s\<^sub>s'"
  using assms by (metis step2_\<E>\<P>_const)


lemma step2_\<S>\<^sub>\<T>_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<S>\<^sub>\<T> s\<^sub>n\<^sub>s = \<S>\<^sub>\<T> s\<^sub>s"
    shows "\<S>\<^sub>\<T> s\<^sub>n\<^sub>s' = \<S>\<^sub>\<T> s\<^sub>s'"
  using assms by (metis step2_\<S>\<^sub>\<T>_const)


lemma step2_\<U>rd_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<U>rd s\<^sub>n\<^sub>s = \<U>rd s\<^sub>s" 
    shows "\<U>rd s\<^sub>n\<^sub>s' = \<U>rd s\<^sub>s'"
  using assms by (metis step2_\<U>rd_const)


lemma step2_\<U>wr_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<U>wr s\<^sub>n\<^sub>s = \<U>wr s\<^sub>s"
    shows "\<U>wr s\<^sub>n\<^sub>s' = \<U>wr s\<^sub>s'"
  using assms by (metis step2_\<U>wr_const)


lemma step2_\<Delta>\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<Delta>\<^sub>0 s\<^sub>n\<^sub>s' = \<Delta>\<^sub>0 s\<^sub>s'"
  using assms apply (induction rule: valid_step2_cases, simp_all)
  using assms(2) apply (simp_all add: wfs2_def const_\<Delta>\<^sub>0)
  apply (metis assms(1) assms(2) step2_reg_update_\<Delta>\<^sub>0_equal)
  by (metis assms(1) assms(2) step2_load_\<Delta>\<^sub>0_equal)


lemma step2_\<mu>\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "\<mu>\<^sub>0 s\<^sub>n\<^sub>s' = \<mu>\<^sub>0 s\<^sub>s'"
  using assms apply (induction rule: valid_step2_cases, simp_all)
  using assms(2) apply (simp_all add: wfs2_def const_\<Delta>\<^sub>0)
  apply (simp_all add: const_\<mu>\<^sub>0)
  using assms(2) step2_store_\<mu>\<^sub>0_equal by auto


lemma step2_pc\<^sub>0_equal:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "pc\<^sub>0 s\<^sub>n\<^sub>s' = pc\<^sub>0 s\<^sub>s'"
proof (cases "speculating s\<^sub>s")
  case True
  then show ?thesis 
    using assms(2) apply (auto simp add: wfs2_def)
    using assms step2_const_pc\<^sub>0 by presburger
next
  case False
    have ns_now_s1: "\<not> speculating s\<^sub>n\<^sub>s"
      using assms(2) wfs2_s\<^sub>n\<^sub>s_ns by auto
    then show ?thesis 
      proof (cases "speculating s\<^sub>s'")
        case True
        show ?thesis
          using assms apply (induction rule: valid_step2_cases, simp_all)
          using assms(2) apply (simp_all only: True False wfs2_def)
          using assms step2_jmp_pc\<^sub>0_equal step2_next_pc\<^sub>0_equal by (simp_all add: False)
      next
        case False
        show ?thesis
          using assms apply (induction rule: valid_step2_cases, simp_all)
          using assms(2) apply (simp_all only: False `\<not>speculating s\<^sub>s` wfs2_def)
          using assms step2_jmp_pc\<^sub>0_equal step2_next_pc\<^sub>0_equal by (simp_all add: `\<not>speculating s\<^sub>s` wfs2_def)
      qed
    qed


lemma step2_wfs2:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
    shows "wfs2 s\<^sub>n\<^sub>s' s\<^sub>s'"
  apply (simp only: wfs2_def) 
  using assms apply (intro conjI)
  using step2_pc\<^sub>0_equal apply blast
  using step2_\<Delta>\<^sub>0_equal apply blast
  using step2_\<mu>\<^sub>0_equal apply blast
  using wfs2_def step2_wf apply blast
  using wfs2_def step2_wf apply blast
  using wfs2_def step2_\<Pi>_equal apply blast
  using wfs2_def step2_\<U>wr_equal apply blast
  using wfs2_def step2_\<U>rd_equal apply blast
  using wfs2_def step2_\<S>\<^sub>\<T>_equal apply blast
  using wfs2_def step2_\<E>\<P>_equal apply blast
  using wfs2_def step2_\<T>\<^sub>\<rho>_equal apply blast
  using step2_\<omega>_subset apply blast
  using step2_wf2\<beta>\<^sub>\<omega> apply blast
  using wfs2_def const_\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d valid_step2_step apply blast
  using wfs2_def const_\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d valid_step2_s\<^sub>s_always_step by blast


lemma step2_wfs2_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not> speculating s\<^sub>s'"
      and "\<not> speculating s\<^sub>s"
      and "wfs2_ns s\<^sub>n\<^sub>s s\<^sub>s"
    shows "wfs2_ns s\<^sub>n\<^sub>s' s\<^sub>s'"
  apply (simp add: wfs2_ns_def)
  using assms step2_\<beta>\<^sub>\<omega>_equal_ns step2_\<omega>_equal_ns by blast


lemma step2_conformantStutter:
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> 
    \<Longrightarrow> conformantStutter s\<^sub>s s\<^sub>n\<^sub>s s\<^sub>n\<^sub>s'"
  by (auto simp add: conformantStutter_def)

lemma step2_conformantStoreAddr: 
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s'); conformantStoreAddr s\<^sub>n\<^sub>s\<rbrakk> 
    \<Longrightarrow> conformantStoreAddr s\<^sub>n\<^sub>s' \<and> conformantStoreAddr s\<^sub>s'"
  using s'_conformantStoreAddr valid_step2_step by blast

lemma step2_conformantEntrypoint: 
  "\<lbrakk>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<rbrakk> 
    \<Longrightarrow> conformantEntrypoint s\<^sub>n\<^sub>s s\<^sub>n\<^sub>s' \<and> conformantEntrypoint s\<^sub>s s\<^sub>s'"
  using valid_step2.simps by blast

lemma step2_forward_jumps: 
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
      and \<open>forward_jumps s\<^sub>n\<^sub>s\<close> and \<open>forward_jumps s\<^sub>s\<close>
    shows \<open>forward_jumps s\<^sub>n\<^sub>s' \<and> forward_jumps s\<^sub>s'\<close>
  using assms valid_step2_step by (metis step_forward_jumps)

lemma step2_future\<^sub>\<rho>_P: 
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
      and \<open>forward_jumps s\<^sub>n\<^sub>s\<close> and \<open>forward_jumps s\<^sub>s\<close>
      and \<open>future\<^sub>\<rho>_P s\<^sub>n\<^sub>s P\<close> and \<open>future\<^sub>\<rho>_P s\<^sub>s P\<close>
    shows \<open>future\<^sub>\<rho>_P s\<^sub>n\<^sub>s' P \<and> future\<^sub>\<rho>_P s\<^sub>s' P\<close>
  using assms valid_step2_step by (metis step_future\<^sub>\<rho>_P_holds)

lemma step2_future\<^sub>\<iota>_P: 
  assumes \<open>(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')\<close>
      and \<open>forward_jumps s\<^sub>n\<^sub>s\<close> and \<open>forward_jumps s\<^sub>s\<close>
      and \<open>future\<^sub>\<iota>_P s\<^sub>n\<^sub>s P\<close> and \<open>future\<^sub>\<iota>_P s\<^sub>s P\<close>
    shows \<open>future\<^sub>\<iota>_P s\<^sub>n\<^sub>s' P \<and> future\<^sub>\<iota>_P s\<^sub>s' P\<close>
  using assms valid_step2_step by (metis step_future\<^sub>\<iota>_P_holds)

(*JUNK*)
(* TODO move and weaken *)

lemma next_\<rho>_equal: "\<not>speculating s\<^sub>s \<and> \<not>speculating s\<^sub>n\<^sub>s \<and> pc\<^sub>0 s\<^sub>n\<^sub>s = pc\<^sub>0 s\<^sub>s  \<Longrightarrow> next_\<rho> (\<rho> s\<^sub>s) = next_\<rho> (\<rho> s\<^sub>n\<^sub>s)"
  by (metis ns_pc\<^sub>0_\<rho>_equal)



subsection \<open>Instructions \<iota>\<close>

lemma step2_Load_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Load r e"
    shows "s\<^sub>n\<^sub>s' = trans_load (\<rho> s\<^sub>n\<^sub>s) r e s\<^sub>n\<^sub>s \<and>
           s\<^sub>s' = trans_load (\<rho> s\<^sub>s) r e s\<^sub>s"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  using assms(4) by (simp_all add: step_Load)

lemma step2_Load_s:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "speculating s\<^sub>s"
      and "\<iota> s\<^sub>s = Load r e"
    shows "(s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_load (\<rho> s\<^sub>s) r e s\<^sub>s) \<or>
           (s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_resolve (\<rho> s\<^sub>s) s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  using step_Load apply blast
  using valid_step_defs apply presburger
  using step_Resolve by auto

lemma step2_SpecFence_s:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "speculating s\<^sub>s"
      and "\<iota> s\<^sub>s = SpecFence"
    shows "(s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_specfence (\<rho> s\<^sub>s) s\<^sub>s) \<or> 
           (s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_resolve (\<rho> s\<^sub>s) s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  using step_SpecFence apply blast
  apply (metis step_Resolve)
  using step_Resolve by blast

lemma step2_RegisterUpdate_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = RegisterUpdate r e"
    shows "s\<^sub>n\<^sub>s' = trans_register_update (\<rho> s\<^sub>n\<^sub>s) r e s\<^sub>n\<^sub>s \<and>
           s\<^sub>s' = trans_register_update (\<rho> s\<^sub>s) r e s\<^sub>s"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  using assms(4) by (simp_all add: step_RegisterUpdate)

lemma step2_BranchMispred_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c"
    shows "(s\<^sub>n\<^sub>s' = trans_tpred (\<rho> s\<^sub>n\<^sub>s) c s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_tmispred (\<rho> s\<^sub>s) c s\<^sub>s) \<or>
           (s\<^sub>n\<^sub>s' = trans_ntpred (\<rho> s\<^sub>n\<^sub>s) s\<^sub>n\<^sub>s  \<and> s\<^sub>s' = trans_ntmispred (\<rho> s\<^sub>s) c s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  using assms(4,5) by (simp_all add: step_BranchFalseMispred step_BranchFalsePred
                                     step_BranchTrueMispred step_BranchTruePred)

lemma step2_Branch_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "\<not>speculating s\<^sub>s"
      and "\<iota> s\<^sub>n\<^sub>s = Branch e c"
    shows "(s\<^sub>n\<^sub>s' = trans_tpred (\<rho> s\<^sub>n\<^sub>s) c s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_tpred (\<rho> s\<^sub>s) c s\<^sub>s) \<or>
           (s\<^sub>n\<^sub>s' = trans_tpred (\<rho> s\<^sub>n\<^sub>s) c s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_tmispred (\<rho> s\<^sub>s) c s\<^sub>s) \<or>
           (s\<^sub>n\<^sub>s' = trans_ntpred (\<rho> s\<^sub>n\<^sub>s) s\<^sub>n\<^sub>s  \<and> s\<^sub>s' = trans_ntpred (\<rho> s\<^sub>s) s\<^sub>s) \<or>
           (s\<^sub>n\<^sub>s' = trans_ntpred (\<rho> s\<^sub>n\<^sub>s) s\<^sub>n\<^sub>s  \<and> s\<^sub>s' = trans_ntmispred (\<rho> s\<^sub>s) c s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  by (simp_all add: valid_step_defs)

lemma step2_Halt_resolve:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "wfs2 s\<^sub>n\<^sub>s s\<^sub>s"
      and "halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>s"
    shows "(s\<^sub>n\<^sub>s' = s\<^sub>n\<^sub>s \<and> s\<^sub>s' = trans_resolve (\<rho> s\<^sub>s) s\<^sub>s)"
  using assms apply (induct rule: valid_step2_cases, simp_all)
  apply (metis step_Resolve)
  using step_Resolve by force+

lemma step2_no_mispred_ns:
  assumes "(s\<^sub>n\<^sub>s, s\<^sub>s) \<leadsto>\<^sub>2 (s\<^sub>n\<^sub>s', s\<^sub>s')"
      and "\<not>mispred (n s\<^sub>s) (\<beta> s\<^sub>s) (pc s\<^sub>s)"
      and "\<not>speculating s\<^sub>s"
    shows "\<not>speculating s\<^sub>s'"
  apply (subgoal_tac "s\<^sub>s \<leadsto> s\<^sub>s'")
  using assms(2) assms(3) step_no_mispred_ns apply blast
  using assms(1) valid_step2_step by blast

end
