theory StepSemantics
    imports StatementSemantics
begin


section \<open>System Model for Speculative AIR\<close>

abbreviation "t_pred     s e \<equiv>  (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> \<not>mispred (n s) (\<beta> s) (pc s)"
abbreviation "nt_pred    s e \<equiv> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> \<not>mispred (n s) (\<beta> s) (pc s)"
abbreviation "t_mispred  s e \<equiv>  (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and>  mispred (n s) (\<beta> s) (pc s)"
abbreviation "nt_mispred s e \<equiv> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and>  mispred (n s) (\<beta> s) (pc s)"


subsection \<open>Step function\<close>

definition step :: \<open>MachineStateTransformer\<close> where
\<open>step s \<equiv> 
  if resolve (n s) (\<beta> s) (pc s) then trans_resolve (\<rho> s) s
  else if halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s then s
  else eval_instr (\<rho> s) (\<iota> s) s
\<close>

subsection \<open>TPOD step constraints\<close>

definition \<open>tpod_step s \<longleftrightarrow> 
  (\<not>speculating s \<longrightarrow> \<not>resolve (n s) (\<beta> s) (pc s)) \<and>
  (speculating s \<and> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<longrightarrow> resolve (n s) (\<beta> s) (pc s)) \<and>
  \<not>terminating s \<and>
  conformantStoreAddr s
\<close>

subsection \<open>Valid Transition\<close>

definition valid_step 
  :: \<open>MachineState \<Rightarrow> MachineState \<Rightarrow> bool\<close> (infixl \<open>\<leadsto>\<close> 101) 
where
  \<open>s \<leadsto> s' \<equiv> tpod_step s \<and> tpod_step s' \<and> conformantEntrypoint s s' \<and> s' = step s\<close>

lemmas valid_step_defs = valid_step_def step_def tpod_step_def terminating_def

text \<open>Case simplification for the valid transition\<close>

lemma valid_step_cases [consumes 1]:
  assumes "s \<leadsto> s'"
  obtains 
    (Resolve) 
          "resolve (n s) (\<beta> s) (pc s)" 
          "speculating s"
          "s' = trans_resolve (\<rho> s) s"
  | (SpecFence) 
          "\<not>resolve (n s) (\<beta> s) (pc s)"  
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s" 
          "\<iota> s = SpecFence" 
          "s' = trans_specfence (\<rho> s) s"
  | (TMispred) e c where "\<not>resolve (n s) (\<beta> s) (pc s)" "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s" "\<iota> s = Branch e c"
          "s' = trans_tmispred (\<rho> s) c s"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)"
          "mispred (n s) (\<beta> s) (pc s)"
  | (TPred) e c where "\<not> resolve (n s) (\<beta> s) (pc s)" 
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Branch e c"
          "s' = trans_tpred (\<rho> s) c s"
          "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)"
          "\<not> mispred (n s) (\<beta> s) (pc s)"
  | (NTMispred) e c where "\<not> resolve (n s) (\<beta> s) (pc s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Branch e c"
          "s' = trans_ntmispred (\<rho> s) c s"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
          "mispred (n s) (\<beta> s) (pc s)"
  | (NTPred) e c where "\<not> resolve (n s) (\<beta> s) (pc s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Branch e c"
          "s' = trans_ntpred (\<rho> s) s"
          "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
          "\<not> mispred (n s) (\<beta> s) (pc s)"
  | (Goto) c where "\<not> resolve (n s) (\<beta> s) (pc s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Goto c"
          "s' = trans_goto (\<rho> s) c s"
  | (RegisterUpdate) r e where 
          "\<not> resolve (n s) (\<beta> s) (pc s) "
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = RegisterUpdate r e"
          "s' = trans_register_update (\<rho> s) r e s"
  | (Load) r e where 
          "\<not> resolve (n s) (\<beta> s) (pc s) "
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Load r e"
          "s' = trans_load (\<rho> s) r e s"
  | (Store) e\<^sub>1 e\<^sub>2 where 
          "\<not> resolve (n s) (\<beta> s) (pc s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = Store e\<^sub>1 e\<^sub>2"
          "s' = trans_store (\<rho> s) e\<^sub>1 e\<^sub>2 s"
  | (NoOp) 
          "\<not> resolve (n s) (\<beta> s) (pc s)"
          "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
          "\<iota> s = NoOp"
          "s' = trans_noop (\<rho> s) s"
  using assms
  apply (simp add: valid_step_defs)
  apply (elim conjE)
  apply (case_tac "resolve (n s) (\<beta> s) (pc s)")
  apply (case_tac "halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s")
  apply fastforce
  apply fastforce
  by (metis \<iota>_some_Instr)

subsubsection \<open>Instructions (\<iota>)\<close>

lemma step_SpecFence:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
      and "\<iota> s = SpecFence"
    shows "s' = trans_specfence (\<rho> s) s"
  using assms valid_step_defs by auto

lemma step_Resolve:
  assumes "s \<leadsto> s'"
      and "resolve (n s) (\<beta> s) (pc s)"
    shows "s' = trans_resolve (\<rho> s) s"
  using assms valid_step_defs by auto

lemma step_Load:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Load r e"
    shows "s' = trans_load (\<rho> s) r e s"
  using assms valid_step_defs by auto

lemma step_RegisterUpdate:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = RegisterUpdate r e"
    shows "s' = trans_register_update (\<rho> s) r e s"
  using assms valid_step_defs by auto

lemma step_BranchTrueMispred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "mispred (n s) (\<beta> s) (pc s)"
      and "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)"
    shows "s' = trans_tmispred (\<rho> s) c s"
  using assms valid_step_defs by auto

lemma step_BranchFalseMispred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "mispred (n s) (\<beta> s) (pc s)"
      and "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
    shows "s' = trans_ntmispred (\<rho> s) c s"
  using assms valid_step_defs by auto

lemma step_BranchTruePred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "\<not>mispred (n s) (\<beta> s) (pc s)"
      and "\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)"
    shows "s' = trans_tpred (\<rho> s) c s"
  using assms valid_step_defs by auto

lemma step_BranchFalsePred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "\<not>mispred (n s) (\<beta> s) (pc s)"
      and "\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
    shows "s' = trans_ntpred (\<rho> s) s"
  using assms valid_step_defs by auto

lemma step_BranchPred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "\<not>mispred (n s) (\<beta> s) (pc s)"
    shows "s' = trans_tpred (\<rho> s) c s \<or> s' = trans_ntpred (\<rho> s) s"
  using assms step_BranchFalsePred step_BranchTruePred by blast

lemma step_BranchMispred:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "mispred (n s) (\<beta> s) (pc s)"
    shows "s' = trans_tmispred (\<rho> s) c s \<or> s' = trans_ntmispred (\<rho> s) c s"
  using assms step_BranchFalseMispred step_BranchTrueMispred by blast

lemma step_Branch:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
    shows "s' = trans_tpred (\<rho> s) c s \<or> s' = trans_ntpred (\<rho> s) s \<or> s' = trans_tmispred (\<rho> s) c s \<or> s' = trans_ntmispred (\<rho> s) c s"
  using assms step_BranchMispred step_BranchPred by blast

lemma step_Goto:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Goto c"
    shows "s' = trans_goto (\<rho> s) c s"
  using assms(1) assms(2) assms(3) valid_step_defs by auto

lemma step_NoOp:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = NoOp"
    shows "s' = trans_noop (\<rho> s) s"
  using assms(1) assms(2) assms(3) valid_step_defs by auto


subsubsection \<open>Speculation level (n)\<close>

text \<open>Correct prediction and non spec fence instructions do not change the spec level.\<close>

lemma const_n:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s \<noteq> SpecFence"
  and "(\<And>e c . \<iota> s = Branch e c \<longrightarrow> \<not>mispred (n s) (\<beta> s) (pc s))"
shows "n s' = n s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                trans_goto_def trans_tpred_def trans_ntpred_def trans_noop_def)

text \<open>Only speculate on mispredictions.\<close>

lemma inc_n_mispred:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "mispred (n s) (\<beta> s) (pc s)"
shows "n s' = n s + 1"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tmispred_def trans_ntmispred_def)

text \<open>Specfence resets all speculation.\<close>

lemma n_0_fence: 
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = SpecFence"
shows "n s' = 0"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_specfence_def Let_def)

text \<open>Resolve will resolve a miss-predicted branch by reducing the speculation level.\<close>

lemma dec_n_resolve: 
  assumes "s \<leadsto> s'" 
      and "resolve (n s) (\<beta> s) (pc s)"
    shows "n s' = n s - 1"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms(2) by (simp_all add: trans_resolve_def Let_def)

lemmas n_trans = const_n n_0_fence dec_n_resolve inc_n_mispred

subsubsection \<open>Program memory (\<Pi>)\<close>

text \<open>The program memory is always invariant.\<close>

lemma const_\<Pi>: 
  assumes "s \<leadsto> s'"
    shows "\<Pi> s' = \<Pi> s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Register state (\<Delta>)\<close>

text \<open>Register state is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma const_\<Delta>:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = NoOp"
  and "\<iota> s = Branch e c \<longrightarrow> \<not> mispred (n s) (\<beta> s) (pc s)"
shows "\<Delta> s' = \<Delta> s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def trans_goto_def trans_tpred_def 
                                trans_ntpred_def trans_noop_def)

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma \<Delta>_resolve:
  assumes "s \<leadsto> s'"
  and "resolve (n s) (\<beta> s) (pc s)"
shows "\<Delta> s' = \<Delta>\<^sub>r\<^sub>e\<^sub>s s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_resolve_def)

text \<open>Updating a register (r) with the RegisterUpdate instruction sets its value to v.\<close>

lemma \<Delta>_specfence:
  assumes "s \<leadsto> s'" 
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = SpecFence"
shows "\<Delta> s' = \<Delta> s |` {0}"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)

text \<open>Updating a register (r) with the RegisterUpdate instruction sets its value to v.\<close>

lemma \<Delta>_n_r_v:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = RegisterUpdate r e"
shows "\<Delta> s' = \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_register_update_def Let_def)

text \<open>Updating a register (r) with the Load instruction sets its value to that of the data located 
      at address a in \<mu>.\<close>

lemma \<Delta>_n_r_v_at_a:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Load r e"
shows "\<Delta> s' = \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_load_def Let_def)

text \<open>If miss prediction occurs on a Branch instruction then the registers at the new spec level 
      equal the registers at the old spec level.\<close>

lemma \<Delta>_n_equals_\<Delta>'_n':
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "mispred (n s) (\<beta> s) (pc s)"
  and "wf\<Delta> s"
shows "\<Delta> s' = \<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def wf\<Delta>_def trans_tmispred_def trans_ntmispred_def)

lemmas \<Delta>_trans = const_\<Delta> \<Delta>_n_r_v \<Delta>_n_r_v_at_a \<Delta>_n_equals_\<Delta>'_n' \<Delta>_resolve \<Delta>_specfence



subsubsection \<open>Data memory (\<mu>)\<close>

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma const_\<mu>:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "(\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = NoOp)"
  and "\<iota> s = Branch e c \<longrightarrow> \<not> mispred (n s) (\<beta> s) (pc s)"
shows "\<mu> s' = \<mu> s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tpred_def trans_ntpred_def trans_load_def 
                                trans_goto_def trans_register_update_def trans_noop_def)

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma \<mu>_resolve:
  assumes "s \<leadsto> s'"
  and "resolve (n s) (\<beta> s) (pc s)"
shows "\<mu> s' = (\<mu> s)(n s := None)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_resolve_def)

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma \<mu>_specfence:
  assumes "s \<leadsto> s'" 
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = SpecFence"
shows "\<mu> s' = \<mu> s |` {0}"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)


text \<open>Updating a memory address (a) with the Store instruction sets its value to that of the value 
      v\<close>

lemma \<mu>_n_a_v:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Store e\<^sub>1 e\<^sub>2"
shows "\<mu> s' = \<mu>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def)

text \<open>If miss prediction occurs on a Branch instruction then the memory at the new spec level equal 
      the memory at the old spec level.\<close>

lemma \<mu>_n_equals_\<mu>'_n':
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "mispred (n s) (\<beta> s) (pc s)"
  and "wf\<mu> s"
shows "\<mu> s' = ((\<mu> s)(n s' \<mapsto> the ((\<mu> s)(n s))))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def wf\<mu>_def trans_tmispred_def trans_ntmispred_def)

lemmas \<mu>_trans = const_\<mu> \<mu>_n_a_v \<mu>_n_equals_\<mu>'_n' \<mu>_resolve \<mu>_specfence



subsubsection \<open>Program Counter (pc)\<close>

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma pc_resolve:
  assumes "s \<leadsto> s'" 
      and "resolve (n s) (\<beta> s) (pc s)"
    shows "pc s' = (pc s)(n s := None)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def trans_resolve_def)

lemma \<rho>_resolve_less:
  assumes "s \<leadsto> s'" 
      and "resolve (n s) (\<beta> s) (pc s)"
      and "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
    shows "\<rho> s' = (the (pc s (n s - 1)))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms pc_resolve apply (simp add: halting_def)
  apply (metis One_nat_def dec_n_resolve less_not_refl minus_eq n_not_Suc_n speculating_def)
  using assms(2) by simp_all


text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma pc_specfence:
  assumes "s \<leadsto> s'" 
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = SpecFence"
shows "pc s' = pc s ((n s) \<mapsto> next_\<rho> (pc\<^sub>0 s)) |` {0}"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)

lemma \<rho>_spec_specfence_0:
  assumes "s \<leadsto> s'" 
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = SpecFence"
      and "speculating s"
      and "\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"      
    shows "\<rho> s' = pc\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)
  prefer 2 using assms pc_specfence n_0_fence apply (simp add: speculating_def)
  using assms(2,3) by auto

text \<open>Predicted false evaluation sets the program counter at the current speculation level to \<rho> + 1\<close>

lemma pc_n'_next_\<rho>:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Branch e c \<or> \<iota> s = SpecFence"
  and "\<iota> s = SpecFence \<longrightarrow> \<not> speculating s"
  and "\<iota> s = Branch e c \<longrightarrow> nt_pred s e"
  and "wfpc s"
shows "pc s' = (pc s)(n s' \<mapsto> next_\<rho> (\<rho> s))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                   trans_ntpred_def speculating_def trans_specfence_def)
  by (auto simp add: wfpc_def Let_def)

text \<open>Predicted true evaluation sets the program counter at the current speculation level to c\<close>

lemma pc_n'_c:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Goto c \<or> \<iota> s = Branch e c"
  and "\<iota> s = Branch e c \<longrightarrow> t_pred s e"
shows "pc s' = (pc s)(n s' \<mapsto> c)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tpred_def trans_goto_def trans_noop_def)

text \<open>Miss-predicted true evaluation sets the program counter at the new speculation level to \<rho> + 1
      and to c at the current speculation level.\<close>

lemma pc_n_c_n'_next_\<rho>:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "t_mispred s e"
shows "pc s' = (pc s)([n s, n s'] [\<mapsto>] [c, next_\<rho> (\<rho> s)])"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tmispred_def)

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c and
      to \<rho> + 1 at the current speculation level.\<close>

lemma pc_n_next_\<rho>_n'_c:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "nt_mispred s e"
shows "pc s' = (pc s)([n s, n s'] [\<mapsto>] [next_\<rho> (\<rho> s), c])"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: Let_def trans_ntmispred_def)

lemmas pc_trans = pc_resolve pc_specfence pc_n'_next_\<rho> pc_n_c_n'_next_\<rho> pc_n'_c pc_n_next_\<rho>_n'_c



subsubsection \<open>Trace of instructions and memory accessed (\<omega>)\<close>

text \<open>Resolve and specfence instructions do not leave a trace.\<close>

lemma const_\<omega>:
  assumes "s \<leadsto> s'"
  and "resolve (n s) (\<beta> s) (pc s)"
shows "\<omega> s' = \<omega> s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_resolve_def)

text \<open>Instructions which do not access memory only add their program address to the trace.\<close>

lemma \<omega>_\<rho>_None:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = RegisterUpdate r e \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp"
shows "\<omega> s' = \<omega> s @ [(\<rho> s, None)]"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_def)

text \<open>Instructions that access memory add this to the trace along with their program address.\<close>

lemma \<omega>_\<rho>_a:
  assumes "s \<leadsto> s'"
  and "\<iota> s = Load r e \<or> \<iota> s = Store e e\<^sub>2"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "a = (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))"
shows "\<omega> s' = \<omega> s @ [(\<rho> s, Some a)]"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def trans_load_def)

text \<open>\<omega> is always a prefix of \<omega>'\<close>

lemma \<omega>_always_prefix:
  assumes "s \<leadsto> s'"
    shows "prefix (\<omega> s) (\<omega> s')"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)


text \<open>Program address is always logged and the memory address can be logged\<close>

lemma \<omega>_always_appends_\<rho>:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
    shows "\<exists>b. \<omega> s' = \<omega> s @ [(\<rho> s, b)]"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms(1,2) by (simp_all add: \<omega>_\<rho>_None \<omega>_\<rho>_a)



subsubsection \<open>Branch predictor (\<beta>)\<close>

text \<open>Non branching instructions do not change the branch predictor.\<close>

lemma const_\<beta>:
  assumes "s \<leadsto> s'"
  and "\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
shows "\<beta> s' = \<beta> s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                trans_specfence_def trans_noop_def)

text \<open>The prediction state on a branch predictor never changes.\<close>

lemma const_\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d:
  assumes "s \<leadsto> s'"
    shows "\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s' = \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: trans_def)
  using \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const by blast+


text \<open>Branching instructions add accesses to the branch predictor.\<close>

lemma update_\<beta>_True:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s)"
  and "\<iota> s = Branch e c"
  and "t_mispred s e \<or> nt_mispred s e"
shows "\<beta> s' = update (n s) (\<rho> s) True (\<beta> s)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_ntpred_def 
                  trans_tmispred_def trans_ntmispred_def halting_def)

lemma update_\<beta>\<^sub>\<omega>_True:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
      and "t_mispred s e \<or> nt_mispred s e"
    shows "\<beta>\<^sub>\<omega> s' = \<beta>\<^sub>\<omega> s @ [(n s, \<rho> s, True)]"
proof -
  have "\<beta> s' = update (n s) (\<rho> s) True (\<beta> s)"
    using assms update_\<beta>_True by simp
  then show ?thesis by (simp add: update_def)
qed

lemma update_\<beta>_False:
  assumes "s \<leadsto> s'"
  and "\<not> resolve (n s) (\<beta> s) (pc s) \<longrightarrow> \<iota> s = Branch e c \<or> \<iota> s = Goto c"
  and "(\<not> resolve (n s) (\<beta> s) (pc s) \<and> \<iota> s = Branch e c) \<longrightarrow> t_pred s e \<or> nt_pred s e"
shows "\<beta> s' = update (n s) (\<rho> s) False (\<beta> s)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_resolve_def trans_goto_def update_def trans_ntpred_def
                  trans_tpred_def halting_def)

lemma update_\<beta>\<^sub>\<omega>_False:
  assumes "s \<leadsto> s'"
      and "\<not> resolve (n s) (\<beta> s) (pc s) \<longrightarrow> \<iota> s = Branch e c \<or> \<iota> s = Goto c"
      and "(\<not> resolve (n s) (\<beta> s) (pc s) \<and> \<iota> s = Branch e c) \<longrightarrow> t_pred s e \<or> nt_pred s e"
    shows "\<beta>\<^sub>\<omega> s' = \<beta>\<^sub>\<omega> s @ [(n s, \<rho> s, False)]"
proof -
  have "\<beta> s' = update (n s) (\<rho> s) False (\<beta> s)"
    using assms update_\<beta>_False by simp
  then show ?thesis by (simp add: update_def)
qed

lemmas \<beta>_trans = const_\<beta> update_\<beta>_True update_\<beta>_False

subsection \<open>Adversary Observables\<close>
(* TODO prove lemmas like "the adversary can not read or store data they can't access"*)

subsubsection \<open>Trusted program instructions (\<T>\<rho>)\<close>

text \<open>The trusted set is always invariant.\<close>

lemma const_\<T>\<^sub>\<rho>: 
  assumes "s \<leadsto> s'"
    shows "\<T>\<^sub>\<rho> s' = \<T>\<^sub>\<rho> s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Adversary writable addresses (\<U>wr)\<close>

text \<open>The adversary writable set is always invariant.\<close>

lemma const_\<U>wr: 
  assumes "s \<leadsto> s'"
    shows "\<U>wr s' = \<U>wr s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Adversary readable addresses (\<U>rd)\<close>

text \<open>The adversary readable set is always invariant.\<close>

lemma const_\<U>rd: 
  assumes "s \<leadsto> s'"
    shows "\<U>rd s' = \<U>rd s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Trusted program instructions (\<S>\<T>)\<close>

text \<open>The trusted set is always invariant.\<close>

lemma const_\<S>\<^sub>\<T>: 
  assumes "s \<leadsto> s'"
    shows "\<S>\<^sub>\<T> s' = \<S>\<^sub>\<T> s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Trusted program entry point (\<E>\<P>)\<close>

text \<open>The entry point is always invariant.\<close>

lemma const_\<E>\<P>: 
  assumes "s \<leadsto> s'"
    shows "\<E>\<P> s' = \<E>\<P> s"
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)


(*TODO format, look at correcting (via lemma) the defs *)
subsubsection \<open>Adversary observable program counter (pc\<A>)\<close>

lemma const_pc\<^sub>0:
  assumes "s \<leadsto> s'"
      and "speculating s"
    shows "pc\<^sub>0 s' = pc\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: pc_resolve pc_specfence speculating_def)
  by (auto simp add: trans_def)

lemma next_pc\<^sub>0:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
      and "\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Load r e\<^sub>1 \<or> \<iota> s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s = Branch e\<^sub>1 c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp"
      and "\<iota> s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s e\<^sub>1 \<or> nt_mispred s e\<^sub>1)"
    shows "pc\<^sub>0 s' = next_\<rho> (\<rho> s)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_ntmispred_def trans_load_def Let_def trans_store_def 
                                trans_ntpred_def trans_register_update_def trans_goto_def 
                                trans_specfence_def speculating_def trans_noop_def)

lemma pc\<^sub>0_greater:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
      and "\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Load r e\<^sub>1 \<or> \<iota> s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s = Branch e\<^sub>1 c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp"
      and "\<iota> s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s e\<^sub>1 \<or> nt_mispred s e\<^sub>1)"
    shows "pc\<^sub>0 s' > pc\<^sub>0 s"
  using assms next_pc\<^sub>0 next_\<rho>_greater pc\<^sub>0_pc_no_spec by auto


lemma jmp_pc\<^sub>0:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
      and "\<iota> s = Branch e c \<or> \<iota> s =  Goto c"
      and "\<iota> s = Branch e c \<longrightarrow> (t_pred s e \<or> t_mispred s e)"
    shows "pc\<^sub>0 s' = c"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_tmispred_def Let_def trans_tpred_def trans_goto_def 
                                speculating_def)  

subsubsection \<open>Adversary observable registers (\<Delta>\<A>)\<close>


lemma const_\<Delta>\<^sub>0:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s \<longrightarrow> (\<forall>e r. \<iota> s \<noteq> Load r e \<and> \<iota> s \<noteq> RegisterUpdate r e)"
    shows "\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)  
  using assms by (auto simp add: const_\<Delta> trans_ntmispred_def trans_tmispred_def
                                 trans_register_update_def trans_load_def Let_def speculating_def 
                                 trans_specfence_def trans_resolve_def)


lemma nonspec_load_\<Delta>\<^sub>0:
  assumes "s \<leadsto> s'"
  and "\<not>speculating s"
  and "\<iota> s = Load r e"
  and "wfs s"
shows "\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_load_def speculating_def Let_def 
                                 wfs_def wf\<Delta>_def wf\<mu>_def)

lemma nonspec_reg_update_\<Delta>\<^sub>0:
  assumes "s \<leadsto> s'"
  and "\<not>speculating s"
  and "\<iota> s = RegisterUpdate r e"
  and "wfs s"
shows "\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_register_update_def speculating_def 
                                 Let_def wfs_def wf\<Delta>_def)


lemma \<Delta>\<^sub>0_always_defined: (*TODO can loosen this to say that \<Delta> is never undefined - make it part of wfs *)
  assumes "s \<leadsto> s'"
  and "\<Delta> s 0 \<noteq> None"
shows "\<Delta> s' 0 \<noteq> None"
  using assms(1) apply (induct rule: valid_step_cases)
  using \<Delta>_resolve assms(1) assms(2) speculating_def apply auto[1]
  using assms by (auto simp add: \<Delta>_trans trans_ntmispred_def trans_tmispred_def Let_def 
                                 resolve_def)

subsubsection \<open>Adversary observable memory (\<mu>\<A>)\<close>

lemma const_\<mu>\<^sub>0:
  assumes "s \<leadsto> s'"
  and "\<not>speculating s \<longrightarrow> (\<forall>e\<^sub>1 e\<^sub>2. \<iota> s \<noteq> Store e\<^sub>1 e\<^sub>2)"
  shows "\<mu>\<^sub>0 s' = \<mu>\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: const_\<mu> trans_ntmispred_def trans_tmispred_def trans_resolve_def
                                 trans_store_def speculating_def Let_def trans_specfence_def)

lemma nonspec_store_\<mu>\<^sub>0:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
      and "\<iota> s = Store e\<^sub>1 e\<^sub>2"
    shows "\<mu>\<^sub>0 s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>0 s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>0 s)) (\<mu>\<^sub>0 s)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_store_def speculating_def Let_def 
                                 wf\<mu>_def)

lemma \<mu>\<^sub>0_always_defined:
  assumes "s \<leadsto> s'"
  and "\<mu> s 0 \<noteq> None"
shows "\<mu> s' 0 \<noteq> None"
  using assms(1) apply (induct rule: valid_step_cases)
  using \<mu>_resolve assms(1) assms(2) speculating_def apply auto[1]
  using assms by (auto simp add: \<mu>_trans trans_ntmispred_def trans_tmispred_def Let_def resolve_def)

(* probably not needed - can work it out from pcA
subsubsection \<open>Adversary observable program address (\<rho>\<A>)\<close>

lemma const_\<rho>\<^sub>0:
  assumes "s \<leadsto> s'"
    shows "\<rho>\<^sub>0 s' = \<rho>\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: \<rho>\<^sub>0_def const_pc\<^sub>0)
  apply (simp_all add: trans_ntmispred_def trans_load_def Let_def trans_store_def trans_ntpred_def 
                    trans_register_update_def trans_goto_def trans_specfence_def trans_resolve_def)



subsubsection \<open>Adversary observable instruction (\<iota>\<A>)\<close>

lemma const_\<iota>\<^sub>0:
  assumes "s \<leadsto> s'"
    shows "\<iota>\<^sub>0 s' = \<iota>\<^sub>0 s"
  using assms(1) apply (induct rule: valid_step_cases)
  apply (simp_all add: pc\<^sub>0_def const_pc)
  by (simp_all add: trans_ntmispred_def trans_load_def Let_def trans_store_def trans_ntpred_def 
                    trans_register_update_def trans_goto_def)
*)


subsection \<open>Lemmas for the System Model\<close>

text \<open>Next states store addresses are conformant.\<close>

lemma s'_conformantStoreAddr:
  assumes "s \<leadsto> s'"
    shows "conformantStoreAddr s'"
  using assms apply (simp add: valid_step_defs)
  by metis

text \<open>Next states entry points are conformant.\<close>

lemma s'_conformantEntrypoint:
  assumes "s \<leadsto> s'"
    shows "conformantEntrypoint s s'"
  using assms apply (simp add: valid_step_defs)
  by metis

text \<open>Never resolve if executing speculatively.\<close>

lemma never_resolve_not_speculating:
  assumes "s \<leadsto> s'"
  and "\<not>speculating s"
shows "\<not>resolve (n s) (\<beta> s) (pc s)"
  using assms by (simp add: valid_step_defs)

text \<open>Only resolve if speculating.\<close>

lemma resolve_must_be_speculating:
  assumes "s \<leadsto> s'"
  and "resolve (n s) (\<beta> s) (pc s)"
shows "speculating s"
  using assms(1) assms(2) never_resolve_not_speculating by blast

text \<open>If halting but still speculating always resolve.\<close>

lemma always_resolve_on_halt:
  assumes "s \<leadsto> s'"
  and "speculating s"
  and "halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
shows "resolve (n s) (\<beta> s) (pc s)"
  using assms by (simp add:  valid_step_defs)

text \<open>If halting but still speculating always resolve.\<close>

lemma n_1_resolve_nospec:
  assumes "s \<leadsto> s'"
  and "speculating s \<and> \<not> speculating s'"
  and "resolve (n s) (\<beta> s) (pc s)"
shows "n s = 1"
  using assms by (simp add: speculating_def dec_n_resolve)


text \<open>If halting then no valid transition.\<close>

lemma no_valid_transition_halt:
  assumes "\<not> speculating s"
  and "halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
shows "\<not> s \<leadsto> s'"
  apply (rule notI)  
  using assms by (meson valid_step_defs)

text \<open>Performing a jump to HALT will halt the program counter in the next state\<close>

lemma jmp_halt:
  assumes "s \<leadsto> s'"
      and "\<not>resolve (n s) (\<beta> s) (pc s)"
      and "\<iota> s \<in> {Goto c, Branch e c}"
      and "\<iota> s = Branch e c \<longrightarrow> (t_pred s e \<or> nt_mispred s e)"
      and "halting c (\<Pi> s)"
    shows "halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def trans_tpred_def trans_ntmispred_def trans_goto_def)

text \<open>If we mispredict then in the next state we speculate\<close>

lemma mispred_speculating:
  assumes "s \<leadsto> s'"
      and "\<not>resolve (n s) (\<beta> s) (pc s)"
      and "mispred (n s) (\<beta> s) (pc s)"
      and "\<iota> s = Branch e c"
    shows "speculating s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: speculating_def inc_n_mispred)

lemma ns_no_resolve:
  assumes "s \<leadsto> s'"
      and "\<not> speculating s"
    shows "\<not>resolve (n s) (\<beta> s) (pc s)"
  using assms(1) assms(2) resolve_must_be_speculating by auto

lemma mispred_ns:
  assumes "s \<leadsto> s'"
      and "mispred (n s) (\<beta> s) (pc s)"
      and "(\<forall>e c.  \<iota> s \<noteq> Branch e c)"
      and "\<not>speculating s"
    shows "\<not>speculating s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: const_n n_0_fence speculating_def inc_n_mispred)

lemma step_no_mispred_ns:
  assumes "s \<leadsto> s'"
      and "\<not>mispred (n s) (\<beta> s) (pc s)"
      and "\<not>speculating s"
    shows "\<not>speculating s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: const_n n_0_fence speculating_def inc_n_mispred)


text \<open>If we are not speculating in the next state then we cannot mispredict in this state\<close>

lemma not_speculating_no_mispred:
  assumes "s \<leadsto> s'"
      and "\<not>resolve (n s) (\<beta> s) (pc s)"
      and "\<not>speculating s'"
      and "\<iota> s = Branch e c"
    shows "\<not>mispred (n s) (\<beta> s) (pc s)"
  using assms mispred_speculating by auto

text \<open>If we are not speculating in the next state then we cannot mispredict in this state\<close>

lemma speculating_mispred:
  assumes "s \<leadsto> s'"
  and "\<not>speculating s"
  and "speculating s'"
shows "mispred (n s) (\<beta> s) (pc s)"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: speculating_def n_trans)

lemma speculating_branch: (* TODO NOT SURE *)
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
      and "speculating s'"
    shows "(\<exists> e c . \<iota> s = Branch e c)"
  using assms(1-3) mispred_ns speculating_mispred resolve_must_be_speculating by blast

lemma speculating_no_specfence:
  assumes "s \<leadsto> s'"
  and "speculating s"
  and "speculating s'"
  and "\<not>resolve (n s) (\<beta> s) (pc s)"
shows "\<iota> s \<noteq> SpecFence"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: speculating_def n_trans) 

lemma no_valid_transition_nospec_adversary_halt:
  assumes "s \<leadsto> s'"
      and "\<not>speculating s"
    shows "\<not>halting (pc\<^sub>0 s) (\<Pi> s)"
  using assms(1) assms(2) halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def no_valid_transition_halt pc\<^sub>0_pc_no_spec by auto


lemma low_equiv_props_no_step:
  assumes "\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s"
      and "(\<forall>a . a \<in> \<U>rd s \<longrightarrow> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s' a = \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s a)"
      and "\<omega> s' = \<omega> s"
      and "\<beta> s' = \<beta> s"
    shows "\<not>(s \<leadsto> s')"
proof - 
  have "\<beta> s\<lparr>trace := \<beta>\<^sub>\<omega> s @ [(n s, \<rho> s, False)]\<rparr> \<noteq> \<beta> s"
    apply auto
    by (metis (no_types, lifting) BranchPredictorState.select_convs(2) BranchPredictorState.simps(5) BranchPredictorState.surjective append_Nil2 list.simps(3) same_append_eq)
  then show ?thesis
  apply auto
  apply (erule valid_step_cases)
  using assms apply auto
  by (simp_all add: trans_def update_def)
qed

subsection \<open>Well formed transitions\<close>

text \<open>These lemmas prove that s' is always well formed given a valid transition from s to s' where s
      is a well formed state.\<close>
       
lemma wf\<Delta>_trans:
  assumes "s \<leadsto> s'"
  and "wf\<Delta> s"
shows "wf\<Delta> s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<Delta>_def n_trans \<Delta>_trans)
  using never_resolve_not_speculating speculating_def by fastforce

lemma wf\<mu>_trans:
  assumes "s \<leadsto> s'"
  and "wf\<mu> s"
shows "wf\<mu> s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<mu>_def n_trans \<mu>_trans)
  using never_resolve_not_speculating speculating_def by fastforce

lemma wfpc_trans: 
  assumes "s \<leadsto> s'"
  and "wfpc s"
shows "wfpc s'"
  using assms apply (simp_all add: wfpc_def Let_def)
  apply (intro allI impI conjI)
  using assms(1) apply (induct rule: valid_step_cases)
  prefer 12 using assms(1) apply (induct rule: valid_step_cases)
  apply (simp_all add: halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def n_trans trans_store_def trans_resolve_def Let_def
                       trans_specfence_def trans_tmispred_def trans_tpred_def trans_ntmispred_def
                       trans_ntpred_def trans_goto_def trans_register_update_def trans_load_def
                       trans_noop_def)
  by (meson nat_le_Suc_less nat_less_le speculating_def)
    

lemma wf\<omega>_trans: 
  assumes "s \<leadsto> s'"
  and "wf\<omega> s"
  and "wfpc s"
shows "wf\<omega> s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<omega>_def wfpc_def const_\<Pi> Let_def const_\<omega> \<omega>_\<rho>_None \<omega>_\<rho>_a n_trans halting_def)
  using assms(3) not_halting_\<rho>_in_\<Pi> by presburger+


lemma wfn_trans: 
  assumes "s \<leadsto> s'"
      and "wfn s"
    shows "wfn s'"
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (auto simp add: wfn_def const_\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred trans_ntmispred_def
                                    n_trans trans_tmispred_def)
  using \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const apply blast
  apply (simp add: update_def)
  apply (metis \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred)
  using \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const apply blast
  by (metis \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred)

lemma wf\<beta>_trans: 
  assumes "s \<leadsto> s'"
  and "wfpc s"
  and "wfn s"
  and "wf\<beta> s"
shows "wf\<beta> s'"
proof -
  have wfpc': "wfpc s'"
    using assms(1-3) wfpc_trans by blast
  show ?thesis
    using assms(1) apply (induct rule: valid_step_cases)
    using assms apply (auto simp add: update_\<beta>\<^sub>\<omega>_False update_\<beta>\<^sub>\<omega>_True update_\<beta>_True wf\<beta>_def const_\<beta> const_\<Pi> 
                                      const_\<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d)
    using assms(3) apply (simp_all add: wfpc_def Let_def wfn_def)
    using \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred apply blast
    using \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred apply blast
    apply (simp add: update_def)
    apply (metis \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred)
    by (meson \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_const \<beta>\<^sub>p\<^sub>r\<^sub>e\<^sub>d_disabled_no_mispred)
qed

lemma wf\<T>\<^sub>\<rho>_trans:
  assumes "s \<leadsto> s'"
  and "wf\<T>\<^sub>\<rho> s"
shows "wf\<T>\<^sub>\<rho> s'"
  using assms by (simp add: wf\<T>\<^sub>\<rho>_def const_\<T>\<^sub>\<rho> const_\<Pi>)

lemma wf\<E>\<P>_trans:
  assumes "s \<leadsto> s'"
  and "wf\<E>\<P> s"
shows "wf\<E>\<P> s'"
  using assms by (simp add: wf\<E>\<P>_def const_\<T>\<^sub>\<rho> const_\<E>\<P>)

lemma wf\<U>rd_trans:
  assumes "s \<leadsto> s'"
  and "wf\<U>rd s"
  and "wf\<mu> s"
shows "wf\<U>rd s'"
  using assms by (simp add: wf\<U>rd_def const_\<S>\<^sub>\<T> const_\<U>rd)


lemma wf\<U>wr_trans:
  assumes "s \<leadsto> s'"
  and "wf\<U>wr s"
  and "wf\<mu> s"
shows "wf\<U>wr s'"
  using assms by (simp add: wf\<U>wr_def const_\<S>\<^sub>\<T> const_\<U>wr)


lemma wfs_trans: 
  assumes "s \<leadsto> s'"
  and "wfs s"
shows "wfs s'"
  using assms by (simp add: wfs_def wf\<Delta>_trans wf\<mu>_trans wfpc_trans wf\<omega>_trans wf\<beta>_trans
                            wfn_trans wf\<T>\<^sub>\<rho>_trans wf\<E>\<P>_trans wf\<U>rd_trans wf\<U>wr_trans)

end
