theory ExecuteTrivial
  imports Execute
begin

section \<open>Trivial discharges of TPOD for partial executions\<close>

subsection \<open>No speculation\<close>

text \<open>Neither traces \<pi>$_3$ or \<pi>$_4$ speculate at the same time\<close>


lemma execute_strict_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case
    by (auto simp add: low_equiv_sub_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have ns: "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (auto simp add: speculates_def)
  have step4_before_invs: "s\<^sub>3 \<approx>\<^sub>\<L> s\<^sub>4"
    using Step apply (simp)    
    by (metis list.rel_inject(2) list_all2_append ns)
  have execute_wfs4_ns: "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4_triv"
    using Step by (simp add: execute_ns_wfs4_triv ns)
  then have invs_ns: "wfs4_triv s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4"
    using Ball4_appendi_last_holds Step by metis
  have "\<not>speculating s\<^sub>3 \<and> \<not>speculating s\<^sub>4 \<and> \<not>speculating s\<^sub>3' \<and> \<not>speculating s\<^sub>4'"
    using Step by (auto simp add: speculates_def)
  then have step4_after_invs: "s\<^sub>3' \<approx>\<^sub>\<L> s\<^sub>4'"
    using Step.hyps(1) Step.hyps(2) invs_ns step4_before_invs step4_ns_triv_and_triv_low_equiv by blast
  show ?case 
    using Step.IH Step.hyps(14) Step.hyps(15) Step.hyps(16) Step.hyps(17) Step.prems(1) Step.prems(2) low_equiv_append2 ns step4_after_invs by presburger
qed

lemma execute_strict_ns_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms(1) assms(2) assms(3) assms(4) execute_strict_ns_low_equiv wfs4_ns_def by blast


lemma execute_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms by (meson execute4_ns\<pi>\<^sub>3_ns\<pi>\<^sub>4 execute4_ns\<pi>\<^sub>4_ns\<pi>\<^sub>3 execute_strict_ns_low_equiv)

lemma execute_ns_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_ns_low_equiv wfs4_ns_def by blast

lemma execute_s\<^sub>3_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3"
      and "\<not>speculating (hd \<pi>\<^sub>3)"
    shows "\<not>speculates \<pi>\<^sub>3"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
    then show ?case by (simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not> mispred (n s\<^sub>3) (\<beta> s\<^sub>3) (pc s\<^sub>3)"
    using Step by (auto simp add: mispredicts_def)
  have "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])"
    apply (subgoal_tac "\<not> speculating (hd (\<pi>\<^sub>3 @ [s\<^sub>3]))")
    apply (subgoal_tac "\<not> mispredicts (\<pi>\<^sub>3 @ [s\<^sub>3])")
    using Step apply blast
    using Step apply (simp add: mispredicts_def)
    apply metis  
    by (simp add: Step.hyps(16) Step.hyps(17) Step.prems(2))
  then have "\<not> speculating s\<^sub>3"
    using \<open>\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])\<close> not_speculates_append by auto
  then have "\<not> speculating s\<^sub>3'"
    using step4_s\<^sub>3_ns_s_mispred 
    using Step.hyps(1) \<open>\<not> mispred (n s\<^sub>3) (\<beta> s\<^sub>3) (pc s\<^sub>3)\<close> by blast
  then show ?case
    by (meson \<open>\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3])\<close> not_speculates_append not_speculates_append2)
qed

lemma execute_s\<^sub>4_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
    then show ?case by (simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have "\<not> mispred (n s\<^sub>4) (\<beta> s\<^sub>4) (pc s\<^sub>4)"
    using Step by (auto simp add: mispredicts_def)
  have "\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    apply (subgoal_tac "\<not> speculating (hd (\<pi>\<^sub>4 @ [s\<^sub>4]))")
    apply (subgoal_tac "\<not> mispredicts (\<pi>\<^sub>4 @ [s\<^sub>4])")
    using Step apply blast
    using Step apply (simp add: mispredicts_def)
    apply metis  
    by (simp add: Step.hyps(16) Step.hyps(17) Step.prems(2))
  then have "\<not> speculating s\<^sub>4"
    using \<open>\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])\<close> not_speculates_append by auto
  then have "\<not> speculating s\<^sub>4'"
    using step4_s\<^sub>4_ns_s_mispred 
    using Step.hyps(1) \<open>\<not> mispred (n s\<^sub>4) (\<beta> s\<^sub>4) (pc s\<^sub>4)\<close> by blast
  then show ?case
    by (meson \<open>\<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])\<close> not_speculates_append not_speculates_append2)
qed

lemma execute_no_mispred_or_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3 \<or> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>3 \<or> \<not>speculates \<pi>\<^sub>4"
  using assms execute_s\<^sub>3_no_mispred_ns_init_implies_speculating 
        execute_s\<^sub>4_no_mispred_ns_init_implies_speculating by blast

lemma execute_no_mispred_ns_init_implies_speculating:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4"
  using assms execute_s\<^sub>3_no_mispred_ns_init_implies_speculating 
        execute_s\<^sub>4_no_mispred_ns_init_implies_speculating by auto


lemma execute_no_mispred_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_no_mispred_ns_init_implies_speculating execute_ns_low_equiv by blast


lemma execute_no_mispred_wfs_ns_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)"
      and "wfs4_ns (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)"
      and "\<not>mispredicts \<pi>\<^sub>3 \<and> \<not>mispredicts \<pi>\<^sub>4"
      and "\<not>speculating (hd \<pi>\<^sub>3) \<and> \<not>speculating (hd \<pi>\<^sub>4)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  using assms execute_no_mispred_low_equiv wfs4_ns_def by blast





subsection \<open>Execution in the high state is trivially low equivalent\<close>

lemma set_P_append: 
  "(\<forall>s. s \<in> set (\<pi> @ [s']) \<longrightarrow> P s) \<longleftrightarrow> (\<forall>s. s \<in> set \<pi> \<longrightarrow> P s) \<and> P s'"
  by auto

lemma set_P_append2: 
  "(\<forall>s. s \<in> set (\<pi> @ [s', s'']) \<longrightarrow> P s) \<longleftrightarrow> (\<forall>s. s \<in> set (\<pi> @ [s']) \<longrightarrow> P s) \<and> P s''"
  by auto


lemma execute_hi_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>" and "\<forall>s. s \<in> set \<pi>\<^sub>4 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case 
    apply (simp add: low_equiv_sub_def)
    by (metis option.discI)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  then show ?case 
    using low_equiv_s_op_high low_equiv_append2 by auto
qed

lemma execute_hi_low_equiv:
  assumes "execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4"
      and "(\<forall>s. s \<in> set \<pi>\<^sub>3 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>) \<and> (\<forall>s. s \<in> set \<pi>\<^sub>4 \<longrightarrow> inst\<^sub>\<T> s \<noteq> \<bottom>)"
    shows "\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4"
  oops


subsection \<open>Program analysis\<close>

abbreviation \<open>forward_jumps4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 
    \<equiv> forward_jumps s\<^sub>1 \<and> forward_jumps s\<^sub>2 \<and> forward_jumps s\<^sub>3 \<and> forward_jumps s\<^sub>4\<close>

lemma execute_forward_jumps:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
  using assms apply (induct_tac rule: execute_Ball4_inducts_init)
  apply simp
  apply simp
  apply simp
  using step4_forward_jumps by (metis zip4_append2_P)

lemma execute_last_wfs4:
  assumes \<open>execute (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4])\<close>
    shows \<open>wfs4 s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4\<close>
  apply (subgoal_tac \<open>Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4\<close>) defer
  using assms execute_wfs4 apply simp
  apply (subgoal_tac \<open>length \<pi>\<^sub>1 = length \<pi>\<^sub>2 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>3 \<and> length \<pi>\<^sub>1 = length \<pi>\<^sub>4\<close>)
  apply (simp add: Ball4_appendi_last_holds)  
  using assms(1) execute_length_append_eq by blast

abbreviation \<open>future\<^sub>\<rho>_P4 P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4
    \<equiv> future\<^sub>\<rho>_P s\<^sub>1 P \<and> future\<^sub>\<rho>_P s\<^sub>2 P \<and> future\<^sub>\<rho>_P s\<^sub>3 P \<and> future\<^sub>\<rho>_P s\<^sub>4 P\<close>

lemma execute_future\<^sub>\<rho>_Ball_forward_jumps_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<rho>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<rho>_P4 P)\<close>
  using assms apply (induction rule: execute_Ball4_inducts_init_Ball4)
  apply (subgoal_tac "future\<^sub>\<rho>_P4 P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
  using zip4_append2_P apply metis
  using step4_future\<^sub>\<rho>_P by blast

lemma execute_future\<^sub>\<rho>_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<rho>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<rho>_P4 P)\<close>
  using assms by (simp add: execute_forward_jumps execute_future\<^sub>\<rho>_Ball_forward_jumps_P)

abbreviation \<open>future\<^sub>\<iota>_P4 P s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4
    \<equiv> future\<^sub>\<iota>_P s\<^sub>1 P \<and> future\<^sub>\<iota>_P s\<^sub>2 P \<and> future\<^sub>\<iota>_P s\<^sub>3 P \<and> future\<^sub>\<iota>_P s\<^sub>4 P\<close>

lemma execute_future\<^sub>\<iota>_Ball_forward_jumps_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 forward_jumps4\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 P)\<close>
  using assms apply (induction rule: execute_Ball4_inducts_init_Ball4)
  apply (subgoal_tac "future\<^sub>\<iota>_P4 P s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4'")
  using zip4_append2_P apply metis
  using step4_future\<^sub>\<iota>_P by blast

lemma execute_future\<^sub>\<iota>_P:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 P (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
    shows \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 P)\<close>
  using assms by (simp add: execute_forward_jumps execute_future\<^sub>\<iota>_Ball_forward_jumps_P)

lemma execute_future\<^sub>\<iota>_\<iota>_not_branch_no_spec:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 \<iota>_not_branch (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>\<not>speculating (hd \<pi>\<^sub>3)\<close> and \<open> \<not>speculating (hd \<pi>\<^sub>4)\<close>
    shows \<open>\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4\<close>
  apply (subgoal_tac \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 \<iota>_not_branch)\<close>)
   defer
  using assms(1-3) apply (simp add: execute_future\<^sub>\<iota>_P)
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case by (auto simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have hd_speculating: \<open>\<not> speculating (hd (\<pi>\<^sub>3 @ [s\<^sub>3])) \<and> \<not> speculating (hd (\<pi>\<^sub>4 @ [s\<^sub>4]))\<close>
    using Step by (simp add: speculating_def)
  have hd_forward_jumps4: \<open>forward_jumps4 (hd (\<pi>\<^sub>1 @ [s\<^sub>1])) (hd (\<pi>\<^sub>2 @ [s\<^sub>2])) (hd (\<pi>\<^sub>3 @ [s\<^sub>3])) (hd (\<pi>\<^sub>4 @ [s\<^sub>4]))\<close>
    by (simp add: Step.hyps(14-17) Step.prems(3))
  have hd_future\<^sub>\<iota>_P4: \<open>future\<^sub>\<iota>_P4 \<iota>_not_branch (hd (\<pi>\<^sub>1 @ [s\<^sub>1])) (hd (\<pi>\<^sub>2 @ [s\<^sub>2])) (hd (\<pi>\<^sub>3 @ [s\<^sub>3])) (hd (\<pi>\<^sub>4 @ [s\<^sub>4]))\<close>
    by (simp add: Step.IH Step.hyps(1-18) Step.prems(1-4))
  have current_speculates: \<open>\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])\<close>
    using Step.IH hd_speculating hd_future\<^sub>\<iota>_P4 hd_forward_jumps4 Step.hyps(18) execute_future\<^sub>\<iota>_P by blast
  then have current_speculating: \<open>\<not> speculating s\<^sub>3 \<and> \<not> speculating s\<^sub>4\<close>
    using not_speculates_append by blast
  have not_halting: \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>3 \<and> \<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<^sub>4\<close>
    using current_speculating Step terminating_def by auto
  have \<open>future\<^sub>\<iota>_P4 \<iota>_not_branch s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4\<close>
    using Step step4_future\<^sub>\<iota>_P Ball4_append_\<pi>_P_and_P'_holds by (smt (z3))
  then have current_\<iota>_not_branch: "\<iota>_not_branch (\<iota> s\<^sub>3) \<and> \<iota>_not_branch (\<iota> s\<^sub>4)"
    using future\<^sub>\<iota>_P_current_holds current_speculating not_halting by auto  
  have next_speculating: \<open>\<not> speculating s\<^sub>3' \<and> \<not> speculating s\<^sub>4'\<close>
    using step4_\<iota>_not_branch_ns Step.hyps(1-2) current_speculating current_\<iota>_not_branch by blast
  then show ?case
    using current_speculates by (meson not_speculates_append not_speculates_append2)   
qed
  


lemma execute_future\<^sub>\<iota>_\<iota>_not_branch_low_equiv:
  assumes \<open>execute \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4\<close>
      and \<open>future\<^sub>\<iota>_P4 \<iota>_not_branch (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>forward_jumps4 (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>wfs4_triv (hd \<pi>\<^sub>1) (hd \<pi>\<^sub>2) (hd \<pi>\<^sub>3) (hd \<pi>\<^sub>4)\<close>
      and \<open>\<not>speculating (hd \<pi>\<^sub>3)\<close> and \<open> \<not>speculating (hd \<pi>\<^sub>4)\<close>
      and \<open>(hd \<pi>\<^sub>3) \<approx>\<^sub>\<L> (hd \<pi>\<^sub>4)\<close>
    shows \<open>\<pi>\<^sub>3 \<approx>\<^sub>\<L> \<pi>\<^sub>4\<close>
  apply (subgoal_tac \<open>Ball4 \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4 (future\<^sub>\<iota>_P4 \<iota>_not_branch)\<close>)
  apply (subgoal_tac \<open>\<not>speculates \<pi>\<^sub>3 \<and> \<not>speculates \<pi>\<^sub>4\<close>)
  defer
  using assms(1-3,5,6) execute_future\<^sub>\<iota>_\<iota>_not_branch_no_spec apply blast
  apply (simp add: assms(1) assms(2) assms(3) execute_future\<^sub>\<iota>_P)
using assms proof (induction rule: execute_inducts)
  case (Init s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4)
  then show ?case by (simp add: speculates_def)
next
  case (Step s\<^sub>1 s\<^sub>2 s\<^sub>3 s\<^sub>4 s\<^sub>1' s\<^sub>2' s\<^sub>3' s\<^sub>4' \<pi>\<^sub>1 \<pi>\<^sub>2 \<pi>\<^sub>3 \<pi>\<^sub>4)
  have ns: "\<not> speculates (\<pi>\<^sub>3 @ [s\<^sub>3]) \<and> \<not> speculates (\<pi>\<^sub>4 @ [s\<^sub>4])"
    using Step by (auto simp add: speculates_def)
  have execute_wfs4_ns: "Ball4 (\<pi>\<^sub>1 @ [s\<^sub>1]) (\<pi>\<^sub>2 @ [s\<^sub>2]) (\<pi>\<^sub>3 @ [s\<^sub>3]) (\<pi>\<^sub>4 @ [s\<^sub>4]) wfs4_triv"
    using Step by (simp add: execute_ns_wfs4_triv ns)
  have last_low_equiv: \<open>(\<pi>\<^sub>3 @ [s\<^sub>3]) \<approx>\<^sub>\<L> (\<pi>\<^sub>4 @ [s\<^sub>4])\<close>
    using Step.hyps(14) Step.hyps(15) Step.hyps(16) Step.hyps(17) Step.hyps(18) Step.prems(5) Step.prems(6) Step.prems(7) Step.prems(8) execute_ns_low_equiv ns by presburger
  then have \<open>s\<^sub>3 \<approx>\<^sub>\<L> s\<^sub>4\<close>
    using low_equiv_last_low_equiv by auto
  then have \<open>s\<^sub>3' \<approx>\<^sub>\<L> s\<^sub>4'\<close>
    using step4_ns_triv_and_triv_low_equiv Step
    by (metis Ball4_append_\<pi>_and_P_holds execute_wfs4_ns not_speculates_append2)
  then show ?case
    using last_low_equiv low_equiv_append2 by auto
qed

end