theory ProgramAnalysis
  imports StepSemantics
begin

section \<open>Program Analysis\<close>

text \<open>Analysis of the program \<Pi> to discharge trivial proof cases\<close>


subsection \<open>Future Instructions and Addresses\<close>

text \<open>Program analysis uses a naive approach for programs who's jumps always target program 
      addresses after the current. This is referred to as a "forward jump". These can be conditional
      in the case of the branch instruction, or unconditional in the case of the goto instruction. 
      (termination of program counter to the set of all future instructions). 

      TODO: In future this should be extended to the set of reachable program addresses where the 
      set of all program addresses is finite (finite (dom \<Pi> s)), though proving termination is non 
      trivial.\<close>

subsubsection \<open>Forward Jumps from specific address\<close>

text \<open>forward_jumps$_\<rho>$ is a predicate indicating that all jump instructions at an addresses greater
      than \<rho>' in \<Pi>' target an address greater their current address.\<close>

definition 
  forward_jumps\<^sub>\<rho> :: \<open>Addr\<^sub>\<Pi> \<Rightarrow> Prog \<Rightarrow> bool\<close>
where 
  \<open>forward_jumps\<^sub>\<rho> \<rho>' \<Pi>' = (\<forall>p. p \<in> (Set.filter (\<lambda>p. p \<ge> \<rho>') (dom \<Pi>')) \<longrightarrow> 
    (case the (\<Pi>' p) of
      Goto ca \<Rightarrow> ca > p
    | Branch _ ca \<Rightarrow> ca > p
    | _ \<Rightarrow> True))\<close>

text \<open>forward_jumps$_\<rho>$ implies that the current state is halting or if it performs a jump, the 
      target is greater than the current address (a forward jump).\<close>

lemma immediate:
  assumes "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)"
    shows "(\<rho> s) \<notin> (\<Pi>\<^sub>\<rho> s) \<or> (case (\<iota> s) of
              Goto ca \<Rightarrow> ca > \<rho> s
            | Branch _ ca \<Rightarrow> ca > \<rho> s
            | _ \<Rightarrow> True)"
  using assms by (simp add: forward_jumps\<^sub>\<rho>_def)

text \<open>Given we are not halting, forward_jumps$_\<rho>$ implies that if the current state  performs a 
      jump, the target is greater than the current address (a forward jump).\<close>

lemma immediate_no_halt:
  assumes "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)"
      and "\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"  
    shows "(case (\<iota> s) of
              Goto ca \<Rightarrow> ca > \<rho> s
            | Branch _ ca \<Rightarrow> ca > \<rho> s
            | _ \<Rightarrow> True)"
  using assms by (simp add: forward_jumps\<^sub>\<rho>_def halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def)

text \<open>Given we are not halting and the instruction is an unconditional jump, then the jump target 
      must be greater than the current address (a forward jump).\<close>

lemma forward_jumps\<^sub>\<rho>_Goto_forward:
  assumes "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)"
      and "\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
      and "\<iota> s = Goto c"
    shows "c > \<rho> s"
  using assms immediate_no_halt by fastforce

text \<open>Given we are not halting and the instruction is a conditional jump, then the jump target 
      must be greater than the current address (a forward jump).\<close>

lemma forward_jumps\<^sub>\<rho>_Branch_forward:
  assumes "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)"
      and "\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s"
      and "\<iota> s = Branch e c"
    shows "c > \<rho> s"
  using assms immediate_no_halt by fastforce

text \<open>Given the current state we are not halting and the instruction is a conditional jump, then the jump target 
      must be greater than the current address (a forward jump).\<close>

lemma forward_jumps\<^sub>\<rho>_gteq_\<rho>:
  assumes \<open>\<rho>' \<le> \<rho>''\<close>
      and \<open>forward_jumps\<^sub>\<rho> \<rho>' (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> \<rho>'' (\<Pi> s)\<close>
  using assms by (simp add: forward_jumps\<^sub>\<rho>_def)


lemma step_forward_jumps\<^sub>\<rho>_\<rho>_gteq:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<rho> s \<le> \<rho> s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms const_\<Pi> forward_jumps\<^sub>\<rho>_gteq_\<rho> by fastforce

text \<open>Given the speculation level remains constant between transitions and all jumps are forward, 
      then the next program address will be greater than or equal to the current program address.
      This proves that forward_jumps$_\<rho>$ will hold in the next state.\<close>

lemma step_\<rho>_advances_n_eq:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s = n s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
    shows \<open>\<rho> s \<le> \<rho> s'\<close>
  using assms(1) apply (induction rule: valid_step_cases)
  apply (metis assms(1) assms(2) dec_n_resolve diff_less dual_order.irrefl less_numeral_extra(1) speculating_def)
  apply (metis (mono_tags, hide_lams) Instr.distinct(33) assms(1) assms(2) const_pc\<^sub>0 dual_order.strict_implies_order less_add_one n_0_fence nat_le_linear next_\<rho>_def next_pc\<^sub>0)
  using assms(1) assms(2) inc_n_mispred apply fastforce
  apply (simp add: trans_tpred_def)
  using assms(3) forward_jumps\<^sub>\<rho>_Branch_forward less_imp_le_nat apply blast
  using assms(1) assms(2) inc_n_mispred apply force
  defer
  apply (simp add: trans_goto_def)
  using assms(3) forward_jumps\<^sub>\<rho>_Goto_forward less_imp_le_nat apply blast 
  by (simp_all add: trans_ntpred_def trans_register_update_def trans_noop_def trans_store_def 
                    trans_load_def Let_def next_\<rho>_gteq)

lemma step_forward_jumps\<^sub>\<rho>_n_eq:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s = n s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms step_forward_jumps\<^sub>\<rho>_\<rho>_gteq step_\<rho>_advances_n_eq by blast

text \<open>Given the speculation level increases between transitions and all jumps are forward, then the 
      next program address will be greater than or equal to the current program address. Proving 
      that forward_jumps$_\<rho>$ will hold in the next state.\<close>

lemma step_\<rho>_advances_n_gt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s < n s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
    shows \<open>\<rho> s \<le> \<rho> s'\<close>
  using assms(1) apply (induction rule: valid_step_cases)
  apply (metis Suc_n_not_le_n assms(1) assms(2) dec_n_resolve diff_Suc_1 dual_order.strict_implies_order gr0_implies_Suc speculating_def)
  using assms(1) assms(2) n_0_fence apply auto[1]
  apply (simp add: trans_tmispred_def next_\<rho>_gteq)
  apply (metis Instr.distinct(33) assms(1) assms(2) const_n order.strict_implies_not_eq)
  apply (simp add: trans_ntmispred_def next_\<rho>_gteq)
  using assms(3) forward_jumps\<^sub>\<rho>_Branch_forward less_imp_le_nat apply blast
  apply (metis Instr.distinct(33) assms(1) assms(2) const_n less_not_refl3)
  using assms(1) assms(2) const_n by force+

lemma step_forward_jumps\<^sub>\<rho>_n_gt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s < n s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms step_forward_jumps\<^sub>\<rho>_\<rho>_gteq step_\<rho>_advances_n_gt by blast

text \<open>Given the speculation level decreases, we're resolving, and all jumps at the previous 
      speculation level are forward, then the next program address will also be forward.\<close>

lemma step_forward_jumps\<^sub>\<rho>_resolve_n_lt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s > n s'\<close>
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s (n s - 1))) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
proof -
  have "(the (pc s (n s - 1))) = (\<rho> s')"
    using assms(1) apply (induction rule: valid_step_cases)
    apply (simp add: trans_resolve_def)
    using assms(1-2) apply (metis One_nat_def dec_n_resolve not_less_iff_gr_or_eq)
    by (simp_all add: assms(3))

  then show ?thesis
    using assms(1, 4) by (simp add: forward_jumps\<^sub>\<rho>_def const_\<Pi>)
qed

text \<open>Given the speculation level decreases, we're not resolving, and all jumps at the architectural
      speculation level are forward, then the next program address will also be forward.\<close>

lemma step_forward_jumps\<^sub>\<rho>_not_resolve_n_lt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s > n s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s 0)) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
proof -
  have "(the (pc s (0))) = (\<rho> s')"
    using assms(1) apply (induction rule: valid_step_cases)
    apply (simp add: assms(3))
    apply (metis \<rho>_spec_specfence_0 assms(1) assms(2) n_0_fence speculating_def)
    using assms(1) assms(2) inc_n_mispred const_n by auto

  then show ?thesis
    using assms(1, 4) by (simp add: forward_jumps\<^sub>\<rho>_def const_\<Pi>)
qed

text \<open>Given the speculation level decreases, and all jumps at the previous and architectural 
      speculation level are forward, then the next program address will also be forward.\<close>

lemma step_forward_jumps\<^sub>\<rho>_n_lt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s > n s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s 0)) (\<Pi> s)\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s (n s - 1))) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms step_forward_jumps\<^sub>\<rho>_not_resolve_n_lt step_forward_jumps\<^sub>\<rho>_resolve_n_lt by blast

text \<open>Given all jumps at the current, previous and architectural speculation level are forward, then
      the next program address will also be forward.\<close>

lemma step_forward_jumps\<^sub>\<rho>:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s 0)) (\<Pi> s)\<close>
      and \<open>forward_jumps\<^sub>\<rho> (the (pc s (n s - 1))) (\<Pi> s)\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms step_forward_jumps\<^sub>\<rho>_n_lt step_forward_jumps\<^sub>\<rho>_n_gt step_forward_jumps\<^sub>\<rho>_n_eq
  by (metis linorder_neqE_nat)


subsubsection \<open>Forward Jumps from a state\<close>

text \<open>forward_jumps is a predicate indicating that all jump instructions at an addresses greater
      than any less than or equal to the current speculation level target an address greater their 
      current address.\<close>

definition 
  forward_jumps :: \<open>MachineState \<Rightarrow> bool\<close>
where 
  \<open>forward_jumps s = (\<forall>n'. n' \<le> n s \<longrightarrow> forward_jumps\<^sub>\<rho> (the (pc s (n'))) (\<Pi> s))\<close>

text \<open>The forward_jumps predicate implies forward_jumps$_\<rho>$ at the current, previous and 
      architectural speculation levels.\<close>

lemma forward_jumps_implies_forward_jumps\<^sub>\<rho>_n_lt:
  assumes "forward_jumps s"
    shows "forward_jumps\<^sub>\<rho> (the (pc s (n s - 1))) (\<Pi> s) \<and> forward_jumps\<^sub>\<rho> (the (pc s 0)) (\<Pi> s) \<and> forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)"
  using assms forward_jumps_def by auto

text \<open>The forward_jumps predicate implies forward_jumps$_\<rho>$ at the next program address.\<close>

lemma step_forward_jumps\<^sub>\<rho>_forward_jumps:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')\<close>
  using assms forward_jumps_implies_forward_jumps\<^sub>\<rho>_n_lt step_forward_jumps\<^sub>\<rho> by auto

text \<open>The forward_jumps predicate implies forward_jumps at the next state given we are not resolving
      and the current instruction is a speculative fence.\<close>

lemma step_forward_jumps_specfence:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = SpecFence\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>forward_jumps s'\<close>
  using assms step_forward_jumps\<^sub>\<rho>_forward_jumps forward_jumps_def n_0_fence
  by (metis bot_nat_0.extremum_uniqueI)

text \<open>The forward_jumps predicate implies forward_jumps at the next state given we are not resolving
      and the current instruction is a speculative fence.\<close>

lemma step_pc_lt_n_constant:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s) \<longrightarrow> \<iota> s \<noteq> SpecFence\<close>
    shows \<open>(\<forall>n'. n' < n s \<longrightarrow> ( (the (pc s' n')) = (the (pc s n'))))\<close>
  using assms(1) apply (induction rule: valid_step_cases)
  using assms(2) by (auto simp add: trans_def)

lemma step_forward_jumps_mispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>forward_jumps s'\<close>
  using assms(1, 5) apply (simp_all only: forward_jumps_def const_\<Pi>)
  using assms(1) apply (induction rule: valid_step_cases)
  defer defer
  apply (simp add: trans_tmispred_def)
  using assms(5) apply (simp add: forward_jumps_def)
  apply auto[1]
  apply (subgoal_tac "(\<rho> s) < c")
  apply (subgoal_tac "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)")
  apply (simp add: forward_jumps\<^sub>\<rho>_def)
  apply blast  
  using forward_jumps\<^sub>\<rho>_Branch_forward assms(3) apply blast
  apply (subgoal_tac "forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')")
  apply (subgoal_tac "next_\<rho> (\<rho> s) = (\<rho> s')")
  using const_\<Pi> assms(1) apply metis
  apply force
  using assms(1) assms(5) step_forward_jumps\<^sub>\<rho>_forward_jumps apply blast
  defer
  using step_pc_lt_n_constant apply (simp add: trans_ntmispred_def)
  using assms(5) step_pc_lt_n_constant apply (simp)
  apply auto
  apply (subgoal_tac "next_\<rho> (\<rho> s) > (\<rho> s)")
  apply (subgoal_tac "forward_jumps\<^sub>\<rho> (\<rho> s) (\<Pi> s)")
  apply (simp add: forward_jumps\<^sub>\<rho>_def)
  apply blast
  using next_\<rho>_greater apply presburger
  apply (subgoal_tac "c = (\<rho> s')")
  apply (metis const_\<Pi> step_forward_jumps\<^sub>\<rho>_forward_jumps)
  by (simp_all add: assms(2-4))

lemma step_forward_jumps:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>forward_jumps s'\<close>
  apply (subgoal_tac "forward_jumps\<^sub>\<rho> (\<rho> s') (\<Pi> s')")
  defer 
  using assms(1) assms(2) step_forward_jumps\<^sub>\<rho>_forward_jumps apply blast
  using assms apply (simp add: forward_jumps_def const_\<Pi>)
  using assms(1) apply (induction rule: valid_step_cases)
  using step_pc_lt_n_constant dec_n_resolve apply (metis antisym_conv diff_le_self i_hate_words_helper leI)
  apply (simp add: n_0_fence)
  using step_forward_jumps_mispred const_\<Pi> forward_jumps_def apply auto[1]
  using step_pc_lt_n_constant apply (simp_all add: const_n le_less)
  using step_forward_jumps_mispred const_\<Pi> forward_jumps_def assms(2) by auto

lemma pc\<^sub>0_lowest_implies_forward_jumps:
  assumes \<open>\<forall>n'. n' \<le> (n s) \<longrightarrow> (the (pc s n')) \<ge> pc\<^sub>0 s\<close>
      and \<open>forward_jumps\<^sub>\<rho> (pc\<^sub>0 s) (\<Pi> s)\<close>
    shows \<open>forward_jumps s\<close>
  using assms apply (simp add: forward_jumps_def)
  using forward_jumps\<^sub>\<rho>_gteq_\<rho> by blast

definition 
  lowest\<^sub>\<rho> :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<Pi>\<close>
where
  \<open>lowest\<^sub>\<rho> s = pc\<^sub>0 s\<close>

text \<open>Return a set of all future program addresses\<close>

definition 
  future\<^sub>\<rho> :: \<open>MachineState \<Rightarrow> Addr\<^sub>\<Pi> set\<close>
where 
  \<open>future\<^sub>\<rho> s = Set.filter (\<lambda>\<rho>'. \<rho>' \<ge> lowest\<^sub>\<rho> s) (\<Pi>\<^sub>\<rho> s)\<close>

lemma future\<^sub>\<rho>_contains_current_\<rho>:
  assumes \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>\<not>speculating s\<close>
    shows \<open>\<rho> s \<in> future\<^sub>\<rho> s\<close>
  using assms by (auto simp add: future\<^sub>\<rho>_def lowest\<^sub>\<rho>_def halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def 
                                 speculating_def)

lemma step_future\<^sub>\<rho>_reduces: 
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>future\<^sub>\<rho> s' \<subseteq> future\<^sub>\<rho> s\<close>
  using assms(1) apply (simp add: future\<^sub>\<rho>_def lowest\<^sub>\<rho>_def const_\<Pi>)
  apply (subgoal_tac "pc\<^sub>0 s \<le> pc\<^sub>0 s'")
  apply fastforce
  apply (cases "n s = 0")
  apply (subgoal_tac "forward_jumps\<^sub>\<rho> (pc\<^sub>0 s) (\<Pi> s)")
  apply (cases "n s = n s'")
  using step_\<rho>_advances_n_eq const_pc\<^sub>0 apply metis
  apply (subgoal_tac "n s < n s'")
  using assms(1) apply (induction rule: valid_step_cases)
  apply (simp add: speculating_def)
  using n_0_fence apply fastforce
  apply (metis forward_jumps\<^sub>\<rho>_Branch_forward jmp_pc\<^sub>0 less_nat_zero_code less_or_eq_imp_le speculating_def)
  apply (metis less_nat_zero_code speculating_def speculating_mispred)
  apply (metis less_irrefl_nat next_\<rho>_gteq next_pc\<^sub>0 speculating_def)
  apply (metis Instr.distinct(33) const_n)
  apply (metis Instr.distinct(31) step_n_increases_branch_mispred)
  apply (metis Instr.distinct(5) step_n_increases_branch_mispred)
  apply (metis Instr.distinct(15) step_n_increases_branch_mispred)
  apply (metis Instr.distinct(23) step_n_increases_branch_mispred)
  apply (metis Instr.distinct(35) step_n_increases_branch_mispred)
  apply simp
  apply (metis assms(2) forward_jumps_implies_forward_jumps\<^sub>\<rho>_n_lt)
  by (simp add: const_pc\<^sub>0 speculating_def)


text \<open>Ensure that all future program addresses match a given predicate\<close>

definition 
  future\<^sub>\<rho>_P :: \<open>MachineState \<Rightarrow> (Addr\<^sub>\<Pi> \<Rightarrow> bool) \<Rightarrow>  bool\<close>
where 
  \<open>future\<^sub>\<rho>_P s P = (\<forall>p. p \<in> (future\<^sub>\<rho> s) \<longrightarrow> P p)\<close>

lemma future\<^sub>\<rho>_P_current_holds:
  assumes \<open>future\<^sub>\<rho>_P s P\<close>
      and \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>\<not>speculating s\<close> 
    shows \<open>P (\<rho> s)\<close>
  using assms future\<^sub>\<rho>_P_def future\<^sub>\<rho>_contains_current_\<rho> by auto

lemma step_future\<^sub>\<rho>_P_holds: 
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
      and \<open>future\<^sub>\<rho>_P s P\<close>
    shows \<open>future\<^sub>\<rho>_P s' P\<close>
  using assms by (metis future\<^sub>\<rho>_P_def step_future\<^sub>\<rho>_reduces subsetD)

text \<open>Return a set of all future instructions\<close>

definition 
  future\<^sub>\<iota> :: \<open>MachineState \<Rightarrow> Instr set\<close>
where 
  \<open>future\<^sub>\<iota> s = (\<lambda>\<rho>'. the (\<Pi> s \<rho>')) ` (future\<^sub>\<rho> s)\<close>

lemma future\<^sub>\<iota>_contains_current_\<iota>:
  assumes \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>\<not>speculating s\<close>
    shows \<open>\<iota> s \<in> future\<^sub>\<iota> s\<close>
  using assms by (simp add: future\<^sub>\<rho>_contains_current_\<rho> future\<^sub>\<iota>_def)

lemma step_future\<^sub>\<iota>_reduces: 
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
    shows \<open>future\<^sub>\<iota> s' \<subseteq> future\<^sub>\<iota> s\<close>
  apply (simp add: future\<^sub>\<iota>_def)
  using assms by (metis const_\<Pi> step_future\<^sub>\<rho>_reduces image_mono)

text \<open>Ensure that all future instructions match a given predicate\<close>

definition 
  future\<^sub>\<iota>_P :: \<open>MachineState \<Rightarrow> (Instr \<Rightarrow> bool) \<Rightarrow>  bool\<close>
where 
  \<open>future\<^sub>\<iota>_P s P = (\<forall>i. i \<in> (future\<^sub>\<iota> s) \<longrightarrow> P i)\<close>

lemma future\<^sub>\<iota>_P_current_holds:
  assumes \<open>future\<^sub>\<iota>_P s P\<close>
      and \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>\<not>speculating s\<close> 
    shows \<open>P (\<iota> s)\<close>
  using assms future\<^sub>\<iota>_P_def future\<^sub>\<iota>_contains_current_\<iota> by auto

lemma step_future\<^sub>\<iota>_P_holds: 
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>forward_jumps s\<close>
      and \<open>future\<^sub>\<iota>_P s P\<close>
    shows \<open>future\<^sub>\<iota>_P s' P\<close>
  using assms by (metis future\<^sub>\<iota>_P_def step_future\<^sub>\<iota>_reduces subsetD)

subsection \<open>Trivialities\<close>

text \<open>Intuitively speaking, \<iota>_not_branch ensures that \<Pi> contains no branch instruc-
tions. TPOD assumes that we start in a state without speculation and branching
introduces speculation. Therefore, if \<Pi> contains no branch instructions, there
can be no speculation and no violation of TPOD\<close>

definition "\<iota>_not_branch i \<longleftrightarrow> (\<forall>e c. i \<noteq> Branch e c)"

text \<open>\<iota>_not_load_store ensures that \<Pi> contains no load or store instructions. This is given the 
assumption that the speculative traces are initially architecturally equivalent, thus the absence of
memory operations prevents these traces from diverging.\<close>

definition "\<iota>_not_load_store i \<longleftrightarrow> (\<forall>e\<^sub>1 e\<^sub>2 r. i \<noteq> Load r e\<^sub>1 \<or> i \<noteq> Store e\<^sub>1 e\<^sub>2)"

text \<open>\<rho>_trusted ensure that all instructions are trusted. Any two high states are trivially low 
equivalent to the adversary even if their states differ and would otherwise violate TPOD.\<close>

definition "\<rho>_trusted s p \<longleftrightarrow> (p \<notin> \<T>\<^sub>\<rho> s)"

end