theory StepSemantics
    imports StatementSemantics
begin


section \<open>System Model for Speculative AIR\<close>

abbreviation \<open>t_pred     s e \<equiv>  (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> \<not>mispred (n s) (\<beta> s) (pc s)\<close>
abbreviation \<open>nt_pred    s e \<equiv> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> \<not>mispred (n s) (\<beta> s) (pc s)\<close>
abbreviation \<open>t_mispred  s e \<equiv>  (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and>  mispred (n s) (\<beta> s) (pc s)\<close>
abbreviation \<open>nt_mispred s e \<equiv> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and>  mispred (n s) (\<beta> s) (pc s)\<close>


subsection \<open>Step function\<close>

definition step :: \<open>MachineStateTransformer\<close> where
\<open>step s \<equiv> 
  if resolve (n s) (\<beta> s) (pc s) then trans_resolve (\<rho> s) s
  else if halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s then s
  else eval_instr (\<rho> s) (\<iota> s) s
\<close>

subsection \<open>TPOD step constraints\<close>

definition \<open>tpod_step s \<longleftrightarrow> 
  (\<not>speculating s \<longrightarrow> \<not>resolve (n s) (\<beta> s) (pc s)) \<and>
  (speculating s \<and> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s \<longrightarrow> resolve (n s) (\<beta> s) (pc s)) \<and>
  conformantStoreAddr s
\<close>

subsection \<open>Valid Transition\<close>

definition valid_step 
  :: \<open>MachineState \<Rightarrow> MachineState \<Rightarrow> bool\<close> (infixl \<open>\<leadsto>\<close> 101) 
where
  \<open>s \<leadsto> s' \<equiv> tpod_step s \<and> tpod_step s' \<and> conformantEntrypoint s s' \<and> \<not>terminating s \<and> s' = step s\<close>

lemmas valid_step_defs = valid_step_def step_def tpod_step_def terminating_def

text \<open>Case simplification for the valid transition\<close>

lemma valid_step_cases [consumes 1]:
  assumes \<open>s \<leadsto> s'\<close>
  obtains 
    (Resolve) 
          \<open>resolve (n s) (\<beta> s) (pc s)\<close> 
          \<open>speculating s\<close>
          \<open>s' = trans_resolve (\<rho> s) s\<close>
  | (SpecFence) 
          \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>  
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close> 
          \<open>\<iota> s = SpecFence\<close> 
          \<open>s' = trans_specfence (\<rho> s) s\<close>
  | (TMispred) e c where \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close> \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close> \<open>\<iota> s = Branch e c\<close>
          \<open>s' = trans_tmispred (\<rho> s) c s\<close>
          \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
          \<open>mispred (n s) (\<beta> s) (pc s)\<close>
  | (TPred) e c where \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close> 
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Branch e c\<close>
          \<open>s' = trans_tpred (\<rho> s) c s\<close>
          \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
          \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
  | (NTMispred) e c where \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Branch e c\<close>
          \<open>s' = trans_ntmispred (\<rho> s) c s\<close>
          \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
          \<open>mispred (n s) (\<beta> s) (pc s)\<close>
  | (NTPred) e c where \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Branch e c\<close>
          \<open>s' = trans_ntpred (\<rho> s) s\<close>
          \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
          \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
  | (Goto) c where \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Goto c\<close>
          \<open>s' = trans_goto (\<rho> s) c s\<close>
  | (RegisterUpdate) r e where 
          \<open>\<not> resolve (n s) (\<beta> s) (pc s) \<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = RegisterUpdate r e\<close>
          \<open>s' = trans_register_update (\<rho> s) r e s\<close>
  | (Load) r e where 
          \<open>\<not> resolve (n s) (\<beta> s) (pc s) \<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Load r e\<close>
          \<open>s' = trans_load (\<rho> s) r e s\<close>
  | (Store) e\<^sub>1 e\<^sub>2 where 
          \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
          \<open>s' = trans_store (\<rho> s) e\<^sub>1 e\<^sub>2 s\<close>
  | (NoOp) 
          \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
          \<open>\<not>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
          \<open>\<iota> s = NoOp\<close>
          \<open>s' = trans_noop (\<rho> s) s\<close>
  using assms
  apply (simp add: valid_step_defs)
  apply (elim conjE)
  apply (case_tac \<open>resolve (n s) (\<beta> s) (pc s)\<close>)
  apply (case_tac \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>)
  apply fastforce
  apply fastforce
  by (metis \<iota>_some_Instr)

subsubsection \<open>Instructions (\<iota>)\<close>

lemma Halt_step:
  assumes \<open>s' = trans_resolve (\<rho> s) s\<close>
      and \<open>speculating s\<close>
      and \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma step_Resolve:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
    shows \<open>s' = trans_resolve (\<rho> s) s\<close>
  using assms valid_step_defs by auto

lemma Resolve_step:
  assumes \<open>s' = trans_resolve (\<rho> s) s\<close>
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma step_SpecFence:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>\<iota> s = SpecFence\<close>
    shows \<open>s' = trans_specfence (\<rho> s) s\<close>
  using assms valid_step_defs by auto

lemma SpecFence_step:
  assumes \<open>s' = trans_specfence (\<rho> s) s\<close>
      and \<open>\<iota> s = SpecFence\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma SpecFence_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_specfence (\<rho> s) s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = SpecFence\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: SpecFence_step Resolve_step)

lemma step_Load:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Load r e\<close>
    shows \<open>s' = trans_load (\<rho> s) r e s\<close>
  using assms valid_step_defs by auto

lemma Load_step:
  assumes \<open>s' = trans_load (\<rho> s) r e s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Load r e\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma step_Store:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
    shows \<open>s' = trans_store (\<rho> s) e\<^sub>1 e\<^sub>2 s\<close>
  using assms valid_step_defs by auto

lemma Store_step:
  assumes \<open>s' = trans_store (\<rho> s) e\<^sub>1 e\<^sub>2 s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto


lemma step_RegisterUpdate:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = RegisterUpdate r e\<close>
    shows \<open>s' = trans_register_update (\<rho> s) r e s\<close>
  using assms valid_step_defs by auto

lemma RegisterUpdate_step:
  assumes \<open>s' = trans_register_update (\<rho> s) r e s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = RegisterUpdate r e\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto


lemma step_BranchTrueMispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
    shows \<open>s' = trans_tmispred (\<rho> s) c s\<close>
  using assms valid_step_defs by auto

lemma BranchTrueMispred_step:
  assumes \<open>s' = trans_tmispred (\<rho> s) c s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma step_BranchFalseMispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
    shows \<open>s' = trans_ntmispred (\<rho> s) c s\<close>
  using assms valid_step_defs by auto

lemma BranchFalseMispred_step:
  assumes \<open>s' = trans_ntmispred (\<rho> s) c s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto


lemma step_BranchTruePred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
    shows \<open>s' = trans_tpred (\<rho> s) c s\<close>
  using assms valid_step_defs by auto

lemma BranchTruePred_step:
  assumes \<open>s' = trans_tpred (\<rho> s) c s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma step_BranchFalsePred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
    shows \<open>s' = trans_ntpred (\<rho> s) s\<close>
  using assms valid_step_defs by auto

lemma BranchFalsePred_step:
  assumes \<open>s' = trans_ntpred (\<rho> s) s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto


lemma step_BranchPred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
    shows \<open>s' = trans_tpred (\<rho> s) c s \<or> s' = trans_ntpred (\<rho> s) s\<close>
  using assms step_BranchFalsePred step_BranchTruePred by blast

lemma BranchPred_step:
  assumes \<open>((\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> s' = trans_tpred (\<rho> s) c s) \<or> (\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> s' = trans_ntpred (\<rho> s) s)\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms BranchTruePred_step BranchFalsePred_step by auto 

lemma step_BranchMispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
    shows \<open>s' = trans_tmispred (\<rho> s) c s \<or> s' = trans_ntmispred (\<rho> s) c s\<close>
  using assms step_BranchFalseMispred step_BranchTrueMispred by blast

lemma BranchMispred_step:
  assumes \<open>((\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> s' = trans_tmispred (\<rho> s) c s) \<or> (\<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)) \<and> s' = trans_ntmispred (\<rho> s) c s)\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms BranchFalseMispred_step BranchTrueMispred_step by auto

lemma step_Branch:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
    shows \<open>s' = trans_tpred (\<rho> s) c s \<or> s' = trans_ntpred (\<rho> s) s \<or> s' = trans_tmispred (\<rho> s) c s \<or> s' = trans_ntmispred (\<rho> s) c s\<close>
  using assms step_BranchMispred step_BranchPred by blast

lemma Branch_step:
  assumes \<open>(s' = trans_tpred (\<rho> s) c s \<and> \<not>mispred (n s) (\<beta> s) (pc s) \<and> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
           (s' = trans_ntpred (\<rho> s) s \<and> \<not>mispred (n s) (\<beta> s) (pc s) \<and> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
           (s' = trans_tmispred (\<rho> s) c s \<and> mispred (n s) (\<beta> s) (pc s) \<and> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
           (s' = trans_ntmispred (\<rho> s) c s \<and> mispred (n s) (\<beta> s) (pc s) \<and> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms BranchMispred_step BranchPred_step by blast

lemma step_Goto:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Goto c\<close>
    shows \<open>s' = trans_goto (\<rho> s) c s\<close>
  using assms valid_step_defs by auto

lemma Goto_step:
  assumes \<open>s' = trans_goto (\<rho> s) c s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Goto c\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma Goto_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_goto (\<rho> s) c s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = Goto c\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: Goto_step Resolve_step)

lemma RegisterUpdate_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_register_update (\<rho> s) r e s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = RegisterUpdate r e\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: RegisterUpdate_step Resolve_step)

lemma Load_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_load (\<rho> s) r e s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = Load r e\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: Load_step Resolve_step)

lemma Store_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_store (\<rho> s) e\<^sub>1 e\<^sub>2 s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: Store_step Resolve_step)


lemma step_NoOp:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = NoOp\<close>
    shows \<open>s' = trans_noop (\<rho> s) s\<close>
  using assms valid_step_defs by auto

lemma NoOp_step:
  assumes \<open>s' = trans_noop (\<rho> s) s\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = NoOp\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms valid_step_defs by auto

lemma NoOp_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_noop (\<rho> s) s) \<or>  
           ( resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = NoOp\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms by (auto simp add: NoOp_step Resolve_step)

lemma Branch_s_step:
  assumes \<open>(\<not>resolve (n s) (\<beta> s) (pc s) \<and> (
            (s' = trans_tpred (\<rho> s) c s \<and> \<not>mispred (n s) (\<beta> s) (pc s) \<and> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
            (s' = trans_ntpred (\<rho> s) s \<and> \<not>mispred (n s) (\<beta> s) (pc s) \<and> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
            (s' = trans_tmispred (\<rho> s) c s \<and> mispred (n s) (\<beta> s) (pc s) \<and> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))) \<or> 
            (s' = trans_ntmispred (\<rho> s) c s \<and> mispred (n s) (\<beta> s) (pc s) \<and> \<not>(\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))))
           ) \<or> (resolve (n s) (\<beta> s) (pc s) \<and> s' = trans_resolve (\<rho> s) s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
      and \<open>conformantEntrypoint s s'\<close>
      and \<open>tpod_step s\<close> and \<open>tpod_step s'\<close>
    shows \<open>s \<leadsto> s'\<close>
  using assms apply (auto simp add: Branch_step Resolve_step)
  using Branch_step assms(1) by blast+


subsubsection \<open>Speculation level (n)\<close>

text \<open>Correct prediction and non spec fence instructions do not change the spec level.\<close>

lemma const_n:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s \<noteq> SpecFence\<close>
  and \<open>(\<And>e c . \<iota> s = Branch e c \<longrightarrow> \<not>mispred (n s) (\<beta> s) (pc s))\<close>
shows \<open>n s' = n s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                trans_goto_def trans_tpred_def trans_ntpred_def trans_noop_def)

text \<open>Only speculate on mispredictions.\<close>

lemma inc_n_mispred:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
shows \<open>n s' = n s + 1\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tmispred_def trans_ntmispred_def)

text \<open>Specfence resets all speculation.\<close>

lemma n_0_fence: 
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = SpecFence\<close>
shows \<open>n s' = 0\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_specfence_def Let_def)

text \<open>Resolve will resolve a miss-predicted branch by reducing the speculation level.\<close>

lemma dec_n_resolve: 
  assumes \<open>s \<leadsto> s'\<close> 
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
    shows \<open>n s' = n s - 1\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms(2) by (simp_all add: trans_resolve_def Let_def)

lemmas n_trans = const_n n_0_fence dec_n_resolve inc_n_mispred

subsubsection \<open>Program memory (\<Pi>)\<close>

text \<open>The program memory is always invariant.\<close>

lemma const_\<Pi>: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<Pi> s' = \<Pi> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Register state (\<Delta>)\<close>

text \<open>Register state is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma const_\<Delta>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = NoOp\<close>
  and \<open>\<iota> s = Branch e c \<longrightarrow> \<not> mispred (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<Delta> s' = \<Delta> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def trans_goto_def trans_tpred_def 
                                trans_ntpred_def trans_noop_def)

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma \<Delta>_resolve:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<Delta> s' = \<Delta>\<^sub>r\<^sub>e\<^sub>s s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_resolve_def)

text \<open>Updating a register (r) with the RegisterUpdate instruction sets its value to v.\<close>

lemma \<Delta>_specfence:
  assumes \<open>s \<leadsto> s'\<close> 
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = SpecFence\<close>
shows \<open>\<Delta> s' = \<Delta> s |` {0}\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)

text \<open>Updating a register (r) with the RegisterUpdate instruction sets its value to v.\<close>

lemma \<Delta>_n_r_v:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = RegisterUpdate r e\<close>
shows \<open>\<Delta> s' = \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_register_update_def Let_def)

text \<open>Updating a register (r) with the Load instruction sets its value to that of the data located 
      at address a in \<mu>.\<close>

lemma \<Delta>_n_r_v_at_a:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Load r e\<close>
shows \<open>\<Delta> s' = \<Delta>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s r (\<mu>\<^sub>r\<^sub>e\<^sub>a\<^sub>d s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_load_def Let_def)

text \<open>If miss prediction occurs on a Branch instruction then the registers at the new spec level 
      equal the registers at the old spec level.\<close>

lemma \<Delta>_n_equals_\<Delta>'_n':
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
  and \<open>wf\<Delta> s\<close>
shows \<open>\<Delta> s' = \<Delta>\<^sub>s\<^sub>p\<^sub>e\<^sub>c s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def wf\<Delta>_def trans_tmispred_def trans_ntmispred_def)

lemmas \<Delta>_trans = const_\<Delta> \<Delta>_n_r_v \<Delta>_n_r_v_at_a \<Delta>_n_equals_\<Delta>'_n' \<Delta>_resolve \<Delta>_specfence



subsubsection \<open>Data memory (\<mu>)\<close>

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma const_\<mu>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>(\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = NoOp)\<close>
  and \<open>\<iota> s = Branch e c \<longrightarrow> \<not> mispred (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<mu> s' = \<mu> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tpred_def trans_ntpred_def trans_load_def 
                                trans_goto_def trans_register_update_def trans_noop_def)

text \<open>Data memory is invariant on Resolve and during the SpecFence, Goto and Store instructions.
      Also invariant on the Branch instruction given we do not miss predict.\<close>

lemma \<mu>_resolve:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<mu> s' = (\<mu> s)(n s := None)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_resolve_def)

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma \<mu>_specfence:
  assumes \<open>s \<leadsto> s'\<close> 
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = SpecFence\<close>
shows \<open>\<mu> s' = \<mu> s |` {0}\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)


text \<open>Updating a memory address (a) with the Store instruction sets its value to that of the value 
      v\<close>

lemma \<mu>_n_a_v:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
shows \<open>\<mu> s' = \<mu>\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e s (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>n s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>n s))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def)

text \<open>If miss prediction occurs on a Branch instruction then the memory at the new spec level equal 
      the memory at the old spec level.\<close>

lemma \<mu>_n_equals_\<mu>'_n':
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
  and \<open>wf\<mu> s\<close>
shows \<open>\<mu> s' = ((\<mu> s)(n s' \<mapsto> the ((\<mu> s)(n s))))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def wf\<mu>_def trans_tmispred_def trans_ntmispred_def)

lemmas \<mu>_trans = const_\<mu> \<mu>_n_a_v \<mu>_n_equals_\<mu>'_n' \<mu>_resolve \<mu>_specfence



subsubsection \<open>Program Counter (pc)\<close>

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma pc_resolve:
  assumes \<open>s \<leadsto> s'\<close> 
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
    shows \<open>pc s' = (pc s)(n s := None)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def trans_resolve_def)

lemma \<rho>_resolve_less:
  assumes \<open>s \<leadsto> s'\<close> 
      and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
    shows \<open>\<rho> s' = (the (pc s (n s - 1)))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms pc_resolve apply (simp add: halting_def)
  apply (metis One_nat_def dec_n_resolve less_not_refl minus_eq n_not_Suc_n speculating_def)
  using assms(2) by simp_all


text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c\<close>

lemma pc_specfence:
  assumes \<open>s \<leadsto> s'\<close> 
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = SpecFence\<close>
shows \<open>pc s' = pc s ((n s) \<mapsto> next_\<rho> (pc\<^sub>0 s)) |` {0}\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_specfence_def)

lemma \<rho>_spec_specfence_0:
  assumes \<open>s \<leadsto> s'\<close> 
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = SpecFence\<close>
      and \<open>speculating s\<close>
      and \<open>\<not> halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>      
    shows \<open>\<rho> s' = pc\<^sub>0 s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  prefer 2 using assms pc_specfence n_0_fence apply (simp add: speculating_def)
  using assms(2,3) by auto

text \<open>Predicted false evaluation sets the program counter at the current speculation level to \<rho> + 1\<close>

lemma pc_n'_next_\<rho>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Branch e c \<or> \<iota> s = SpecFence\<close>
  and \<open>\<iota> s = SpecFence \<longrightarrow> \<not> speculating s\<close>
  and \<open>\<iota> s = Branch e c \<longrightarrow> nt_pred s e\<close>
  and \<open>wfpc s\<close>
shows \<open>pc s' = (pc s)(n s' \<mapsto> next_\<rho> (\<rho> s))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                   trans_ntpred_def speculating_def trans_specfence_def)
  by (auto simp add: wfpc_def Let_def)

text \<open>Predicted true evaluation sets the program counter at the current speculation level to c\<close>

lemma pc_n'_c:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Goto c \<or> \<iota> s = Branch e c\<close>
  and \<open>\<iota> s = Branch e c \<longrightarrow> t_pred s e\<close>
shows \<open>pc s' = (pc s)(n s' \<mapsto> c)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tpred_def trans_goto_def trans_noop_def)

text \<open>Miss-predicted true evaluation sets the program counter at the new speculation level to \<rho> + 1
      and to c at the current speculation level.\<close>

lemma pc_n_c_n'_next_\<rho>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>t_mispred s e\<close>
shows \<open>pc s' = (pc s)([n s, n s'] [\<mapsto>] [c, next_\<rho> (\<rho> s)])\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_tmispred_def)

text \<open>Miss-predicted false evaluation sets the program counter at the new speculation level to c and
      to \<rho> + 1 at the current speculation level.\<close>

lemma pc_n_next_\<rho>_n'_c:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>nt_mispred s e\<close>
shows \<open>pc s' = (pc s)([n s, n s'] [\<mapsto>] [next_\<rho> (\<rho> s), c])\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: Let_def trans_ntmispred_def)

lemmas pc_trans = pc_resolve pc_specfence pc_n'_next_\<rho> pc_n_c_n'_next_\<rho> pc_n'_c pc_n_next_\<rho>_n'_c



subsubsection \<open>Trace of instructions and memory accessed (\<omega>)\<close>

text \<open>Resolve and specfence instructions do not leave a trace.\<close>

lemma const_\<omega>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<omega> s' = \<omega> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_resolve_def)

text \<open>Instructions which do not access memory only add their program address to the trace.\<close>

lemma \<omega>_\<rho>_None:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = RegisterUpdate r e \<or> \<iota> s = Branch e c \<or> \<iota> s = Goto c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp\<close>
shows \<open>\<omega> s' = \<omega> s @ [(\<rho> s, None)]\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_def)

text \<open>Instructions that access memory add this to the trace along with their program address.\<close>

lemma \<omega>_\<rho>_a:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<iota> s = Load r e \<or> \<iota> s = Store e e\<^sub>2\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>a = (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s))\<close>
shows \<open>\<omega> s' = \<omega> s @ [(\<rho> s, Some a)]\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_store_def trans_load_def)

text \<open>\<omega> is always a prefix of \<omega>'\<close>

lemma \<omega>_always_prefix:
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>prefix (\<omega> s) (\<omega> s')\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)


text \<open>Program address is always logged and the memory address can be logged\<close>

lemma \<omega>_always_appends_\<rho>:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
    shows \<open>\<exists>b. \<omega> s' = \<omega> s @ [(\<rho> s, b)]\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms(1,2) by (simp_all add: \<omega>_\<rho>_None \<omega>_\<rho>_a)



subsubsection \<open>Branch predictor (\<beta>)\<close>

text \<open>Non branching instructions do not change the branch predictor.\<close>

lemma const_\<beta>:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<iota> s = RegisterUpdate r e \<or> \<iota> s = Load r e \<or> \<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<beta> s' = \<beta> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_register_update_def trans_load_def trans_store_def 
                                trans_specfence_def trans_noop_def)

text \<open>Branching instructions add accesses to the branch predictor.\<close>

lemma update_\<beta>_True:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
  and \<open>\<iota> s = Branch e c\<close>
  and \<open>t_mispred s e \<or> nt_mispred s e\<close>
shows \<open>\<beta> s' = update (n s) (\<rho> s) True (\<beta> s)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_ntpred_def 
                  trans_tmispred_def trans_ntmispred_def halting_def)

lemma update_\<beta>\<^sub>\<omega>_True:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
      and \<open>t_mispred s e \<or> nt_mispred s e\<close>
    shows \<open>\<beta> s' = \<beta> s @ [(n s, \<rho> s, True)]\<close>
proof -
  have \<open>\<beta> s' = update (n s) (\<rho> s) True (\<beta> s)\<close>
    using assms update_\<beta>_True by simp
  then show ?thesis by (simp add: update_def)
qed

lemma update_\<beta>_False:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not> resolve (n s) (\<beta> s) (pc s) \<longrightarrow> \<iota> s = Branch e c \<or> \<iota> s = Goto c\<close>
  and \<open>(\<not> resolve (n s) (\<beta> s) (pc s) \<and> \<iota> s = Branch e c) \<longrightarrow> t_pred s e \<or> nt_pred s e\<close>
shows \<open>\<beta> s' = update (n s) (\<rho> s) False (\<beta> s)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: Let_def trans_resolve_def trans_goto_def update_def trans_ntpred_def
                  trans_tpred_def halting_def)

lemma update_\<beta>\<^sub>\<omega>_False:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> resolve (n s) (\<beta> s) (pc s) \<longrightarrow> \<iota> s = Branch e c \<or> \<iota> s = Goto c\<close>
      and \<open>(\<not> resolve (n s) (\<beta> s) (pc s) \<and> \<iota> s = Branch e c) \<longrightarrow> t_pred s e \<or> nt_pred s e\<close>
    shows \<open>\<beta> s' = \<beta> s @ [(n s, \<rho> s, False)]\<close>
proof -
  have \<open>\<beta> s' = update (n s) (\<rho> s) False (\<beta> s)\<close>
    using assms update_\<beta>_False by simp
  then show ?thesis by (simp add: update_def)
qed

lemmas \<beta>_trans = const_\<beta> update_\<beta>_True update_\<beta>_False

subsection \<open>Adversary Observables\<close>

subsubsection \<open>Trusted program instructions (\<T>\<rho>)\<close>

text \<open>The trusted set is always invariant.\<close>

lemma const_\<T>\<^sub>\<rho>: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<T>\<^sub>\<rho> s' = \<T>\<^sub>\<rho> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Adversary writable addresses (\<U>wr)\<close>

text \<open>The adversary writable set is always invariant.\<close>

lemma const_\<U>wr: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<U>wr s' = \<U>wr s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Adversary readable addresses (\<U>rd)\<close>

text \<open>The adversary readable set is always invariant.\<close>

lemma const_\<U>rd: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<U>rd s' = \<U>rd s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Trusted program instructions (\<S>\<T>)\<close>

text \<open>The trusted set is always invariant.\<close>

lemma const_\<S>\<^sub>\<T>: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<S>\<^sub>\<T> s' = \<S>\<^sub>\<T> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)



subsubsection \<open>Trusted program entry point (\<E>\<P>)\<close>

text \<open>The entry point is always invariant.\<close>

lemma const_\<E>\<P>: 
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>\<E>\<P> s' = \<E>\<P> s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  by (simp_all add: trans_def)

subsubsection \<open>Adversary observable program counter (pc\<A>)\<close>

lemma const_pc\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>speculating s\<close>
    shows \<open>pc\<^sub>0 s' = pc\<^sub>0 s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: pc_resolve pc_specfence speculating_def)
  by (auto simp add: trans_def)

lemma next_pc\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Load r e\<^sub>1 \<or> \<iota> s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s = Branch e\<^sub>1 c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp\<close>
      and \<open>\<iota> s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s e\<^sub>1 \<or> nt_mispred s e\<^sub>1)\<close>
    shows \<open>pc\<^sub>0 s' = next_\<rho> (\<rho> s)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_ntmispred_def trans_load_def Let_def trans_store_def 
                                trans_ntpred_def trans_register_update_def trans_goto_def 
                                trans_specfence_def speculating_def trans_noop_def)

lemma pc\<^sub>0_greater:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2 \<or> \<iota> s = Load r e\<^sub>1 \<or> \<iota> s = RegisterUpdate r e\<^sub>1 \<or> \<iota> s = Branch e\<^sub>1 c \<or> \<iota> s = SpecFence \<or> \<iota> s = NoOp\<close>
      and \<open>\<iota> s = Branch e\<^sub>1 c \<longrightarrow> (nt_pred s e\<^sub>1 \<or> nt_mispred s e\<^sub>1)\<close>
    shows \<open>pc\<^sub>0 s' > pc\<^sub>0 s\<close>
  using assms next_pc\<^sub>0 next_\<rho>_greater pc\<^sub>0_pc_no_spec by auto


lemma jmp_pc\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
      and \<open>\<iota> s = Branch e c \<or> \<iota> s =  Goto c\<close>
      and \<open>\<iota> s = Branch e c \<longrightarrow> (t_pred s e \<or> t_mispred s e)\<close>
    shows \<open>pc\<^sub>0 s' = c\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: trans_tmispred_def Let_def trans_tpred_def trans_goto_def 
                                speculating_def)  

subsubsection \<open>Adversary observable registers (\<Delta>\<A>)\<close>


lemma const_\<Delta>\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s \<longrightarrow> (\<forall>e r. \<iota> s \<noteq> Load r e \<and> \<iota> s \<noteq> RegisterUpdate r e)\<close>
    shows \<open>\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s\<close>
  using assms(1) apply (induct rule: valid_step_cases)  
  using assms by (auto simp add: const_\<Delta> trans_ntmispred_def trans_tmispred_def
                                 trans_register_update_def trans_load_def Let_def speculating_def 
                                 trans_specfence_def trans_resolve_def)


lemma nonspec_load_\<Delta>\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not>speculating s\<close>
  and \<open>\<iota> s = Load r e\<close>
  and \<open>wfs s\<close>
shows \<open>\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s (r \<mapsto> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_load_def speculating_def Let_def 
                                 wfs_def wf\<Delta>_def wf\<mu>_def)

lemma nonspec_reg_update_\<Delta>\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not>speculating s\<close>
  and \<open>\<iota> s = RegisterUpdate r e\<close>
  and \<open>wfs s\<close>
shows \<open>\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s (r \<mapsto> (\<lbrakk>e\<rbrakk>(\<Delta>\<^sub>n s)))\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_register_update_def speculating_def 
                                 Let_def wfs_def wf\<Delta>_def)


lemma \<Delta>\<^sub>0_always_defined:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<Delta> s 0 \<noteq> None\<close>
shows \<open>\<Delta> s' 0 \<noteq> None\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using \<Delta>_resolve assms(1) assms(2) speculating_def apply auto[1]
  using assms by (auto simp add: \<Delta>_trans trans_ntmispred_def trans_tmispred_def Let_def)

subsubsection \<open>Adversary observable memory (\<mu>\<A>)\<close>

lemma const_\<mu>\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not>speculating s \<longrightarrow> (\<forall>e\<^sub>1 e\<^sub>2. \<iota> s \<noteq> Store e\<^sub>1 e\<^sub>2)\<close>
  shows \<open>\<mu>\<^sub>0 s' = \<mu>\<^sub>0 s\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: const_\<mu> trans_ntmispred_def trans_tmispred_def trans_resolve_def
                                 trans_store_def speculating_def Let_def trans_specfence_def)

lemma nonspec_store_\<mu>\<^sub>0:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
      and \<open>\<iota> s = Store e\<^sub>1 e\<^sub>2\<close>
    shows \<open>\<mu>\<^sub>0 s' = \<mu>\<^sub>n\<^sub>w\<^sub>r\<^sub>i\<^sub>t\<^sub>e (\<lbrakk>e\<^sub>1\<rbrakk>(\<Delta>\<^sub>0 s)) (\<lbrakk>e\<^sub>2\<rbrakk>(\<Delta>\<^sub>0 s)) (\<mu>\<^sub>0 s)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: trans_store_def speculating_def Let_def 
                                 wf\<mu>_def)

lemma \<mu>\<^sub>0_always_defined:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<mu> s 0 \<noteq> None\<close>
shows \<open>\<mu> s' 0 \<noteq> None\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using \<mu>_resolve assms(1) assms(2) speculating_def apply auto[1]
  using assms by (auto simp add: \<mu>_trans trans_ntmispred_def trans_tmispred_def Let_def)

subsection \<open>Lemmas for the System Model\<close>

text \<open>Next states store addresses are conformant.\<close>

lemma s'_conformantStoreAddr:
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>conformantStoreAddr s'\<close>
  using assms apply (simp add: valid_step_defs)
  by metis

text \<open>Next states entry points are conformant.\<close>

lemma s'_conformantEntrypoint:
  assumes \<open>s \<leadsto> s'\<close>
    shows \<open>conformantEntrypoint s s'\<close>
  using assms apply (simp add: valid_step_defs)
  by metis

text \<open>Never resolve if executing speculatively.\<close>

lemma never_resolve_not_speculating:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not>speculating s\<close>
shows \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
  using assms by (simp add: valid_step_defs)

text \<open>Only resolve if speculating.\<close>

lemma resolve_must_be_speculating:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>speculating s\<close>
  using assms(1) assms(2) never_resolve_not_speculating by blast

text \<open>If halting but still speculating always resolve.\<close>

lemma always_resolve_on_halt:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>speculating s\<close>
  and \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
shows \<open>resolve (n s) (\<beta> s) (pc s)\<close>
  using assms by (simp add:  valid_step_defs)

text \<open>If halting but still speculating always resolve.\<close>

lemma n_1_resolve_nospec:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>speculating s \<and> \<not> speculating s'\<close>
  and \<open>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>n s = 1\<close>
  using assms by (simp add: speculating_def dec_n_resolve)


text \<open>If halting then no valid transition.\<close>

lemma no_valid_transition_halt:
  assumes \<open>\<not> speculating s\<close>
  and \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s\<close>
shows \<open>\<not> s \<leadsto> s'\<close>
  apply (rule notI)  
  using assms by (meson valid_step_defs)

text \<open>Performing a jump to HALT will halt the program counter in the next state\<close>

lemma jmp_halt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s \<in> {Goto c, Branch e c}\<close>
      and \<open>\<iota> s = Branch e c \<longrightarrow> (t_pred s e \<or> nt_mispred s e)\<close>
      and \<open>halting c (\<Pi> s)\<close>
    shows \<open>halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def trans_tpred_def trans_ntmispred_def trans_goto_def)

text \<open>If n increases then the current instruction must be a branch\<close>

lemma step_n_increases_branch_mispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>n s < n s'\<close>
    shows \<open>\<exists>e c. \<iota> s = Branch e c\<close>
  using assms apply (induction rule: valid_step_cases)
  apply (metis assms(1) dec_n_resolve diff_le_self eq_iff nat_less_le)
  apply (metis assms(1) less_zeroE n_0_fence)
  apply blast+
  using assms(1) const_n by auto


text \<open>If we mispredict then in the next state we speculate\<close>

lemma mispred_speculating:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<iota> s = Branch e c\<close>
    shows \<open>speculating s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: speculating_def inc_n_mispred)

lemma ns_no_resolve:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not> speculating s\<close>
    shows \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
  using assms(1) assms(2) resolve_must_be_speculating by auto

lemma mispred_ns:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>(\<forall>e c.  \<iota> s \<noteq> Branch e c)\<close>
      and \<open>\<not>speculating s\<close>
    shows \<open>\<not>speculating s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: const_n n_0_fence speculating_def inc_n_mispred)

lemma step_no_mispred_ns:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>speculating s\<close>
    shows \<open>\<not>speculating s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (simp_all add: const_n n_0_fence speculating_def inc_n_mispred)


text \<open>If we are not speculating in the next state then we cannot mispredict in this state\<close>

lemma not_speculating_no_mispred:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
      and \<open>\<not>speculating s'\<close>
      and \<open>\<iota> s = Branch e c\<close>
    shows \<open>\<not>mispred (n s) (\<beta> s) (pc s)\<close>
  using assms mispred_speculating by auto

text \<open>If we are not speculating in the next state then we cannot mispredict in this state\<close>

lemma speculating_mispred:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>\<not>speculating s\<close>
  and \<open>speculating s'\<close>
shows \<open>mispred (n s) (\<beta> s) (pc s)\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: speculating_def n_trans)

lemma speculating_branch:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
      and \<open>speculating s'\<close>
    shows \<open>(\<exists> e c . \<iota> s = Branch e c)\<close>
  using assms(1-3) mispred_ns speculating_mispred resolve_must_be_speculating by blast

lemma speculating_no_specfence:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>speculating s\<close>
  and \<open>speculating s'\<close>
  and \<open>\<not>resolve (n s) (\<beta> s) (pc s)\<close>
shows \<open>\<iota> s \<noteq> SpecFence\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms by (auto simp add: speculating_def n_trans) 

lemma no_valid_transition_nospec_adversary_halt:
  assumes \<open>s \<leadsto> s'\<close>
      and \<open>\<not>speculating s\<close>
    shows \<open>\<not>halting (pc\<^sub>0 s) (\<Pi> s)\<close>
  using assms(1) assms(2) halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def no_valid_transition_halt pc\<^sub>0_pc_no_spec by auto


lemma low_equiv_props_no_step:
  assumes \<open>\<Delta>\<^sub>0 s' = \<Delta>\<^sub>0 s\<close>
      and \<open>(\<forall>a . a \<in> \<U>rd s \<longrightarrow> \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s' a = \<mu>\<^sub>0\<^sub>r\<^sub>e\<^sub>a\<^sub>d s a)\<close>
      and \<open>\<omega> s' = \<omega> s\<close>
      and \<open>\<beta> s' = \<beta> s\<close>
    shows \<open>\<not>(s \<leadsto> s')\<close>
proof - 
  have \<open>\<beta> s @ [(n s, \<rho> s, False)] \<noteq> \<beta> s\<close>
    by auto
  then show ?thesis
  apply auto
  apply (erule valid_step_cases)
  using assms apply auto
  by (simp_all add: trans_def update_def)
qed

subsection \<open>Well formed transitions\<close>

text \<open>These lemmas prove that s' is always well formed given a valid transition from s to s' where s
      is a well formed state.\<close>
       
lemma wf\<Delta>_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<Delta> s\<close>
shows \<open>wf\<Delta> s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<Delta>_def n_trans \<Delta>_trans)
  using never_resolve_not_speculating speculating_def by fastforce

lemma wf\<mu>_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<mu> s\<close>
shows \<open>wf\<mu> s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<mu>_def n_trans \<mu>_trans)
  using never_resolve_not_speculating speculating_def by fastforce

lemma wfpc_trans: 
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wfpc s\<close>
shows \<open>wfpc s'\<close>
  using assms apply (simp_all add: wfpc_def Let_def)
  apply (intro allI impI conjI)
  using assms(1) apply (induct rule: valid_step_cases)
  prefer 12 using assms(1) apply (induct rule: valid_step_cases)
  apply (simp_all add: halting\<^sub>p\<^sub>r\<^sub>e\<^sub>d_def halting_def n_trans trans_store_def trans_resolve_def Let_def
                       trans_specfence_def trans_tmispred_def trans_tpred_def trans_ntmispred_def
                       trans_ntpred_def trans_goto_def trans_register_update_def trans_load_def
                       trans_noop_def)
  by (meson nat_le_Suc_less nat_less_le speculating_def)
    

lemma wf\<omega>_trans: 
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<omega> s\<close>
  and \<open>wfpc s\<close>
shows \<open>wf\<omega> s'\<close>
  using assms(1) apply (induct rule: valid_step_cases)
  using assms apply (simp_all add: wf\<omega>_def wfpc_def const_\<Pi> Let_def const_\<omega> \<omega>_\<rho>_None \<omega>_\<rho>_a n_trans halting_def)
  using assms(3) not_halting_\<rho>_in_\<Pi> by presburger+

lemma wf\<T>\<^sub>\<rho>_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<T>\<^sub>\<rho> s\<close>
shows \<open>wf\<T>\<^sub>\<rho> s'\<close>
  using assms by (simp add: wf\<T>\<^sub>\<rho>_def const_\<T>\<^sub>\<rho> const_\<Pi>)

lemma wf\<E>\<P>_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<E>\<P> s\<close>
shows \<open>wf\<E>\<P> s'\<close>
  using assms by (simp add: wf\<E>\<P>_def const_\<T>\<^sub>\<rho> const_\<E>\<P>)

lemma wf\<U>rd_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<U>rd s\<close>
  and \<open>wf\<mu> s\<close>
shows \<open>wf\<U>rd s'\<close>
  using assms by (simp add: wf\<U>rd_def const_\<S>\<^sub>\<T> const_\<U>rd)


lemma wf\<U>wr_trans:
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wf\<U>wr s\<close>
  and \<open>wf\<mu> s\<close>
shows \<open>wf\<U>wr s'\<close>
  using assms by (simp add: wf\<U>wr_def const_\<S>\<^sub>\<T> const_\<U>wr)


lemma wfs_trans: 
  assumes \<open>s \<leadsto> s'\<close>
  and \<open>wfs s\<close>
shows \<open>wfs s'\<close>
  using assms by (simp add: wfs_def wf\<Delta>_trans wf\<mu>_trans wfpc_trans wf\<omega>_trans wf\<T>\<^sub>\<rho>_trans wf\<E>\<P>_trans 
                            wf\<U>rd_trans wf\<U>wr_trans)

end
