theory Step4Semantics
  imports Step2Semantics
begin

type_synonym MachineState4 = "(MachineState × MachineState × MachineState × MachineState)"
type_synonym MachineState4Predicate = "MachineState \<Rightarrow> MachineState \<Rightarrow> MachineState \<Rightarrow> MachineState ⇒ bool"
type_synonym MachineState4Transformer = "MachineState4 ⇒ MachineState4"

subsection ‹Well formed 4 state step›

definition "wfs4 s⇩1 s⇩2 s⇩3 s⇩4 ⟷ 
  wfs2 s⇩1 s⇩3 ∧ wfs2 s⇩2 s⇩4 ∧
  Π s⇩1 = Π s⇩2 ∧ 
  𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩2 ∧
  ℰ𝒫 s⇩1 = ℰ𝒫 s⇩2 ∧
  𝒮⇩𝒯 s⇩1 = 𝒮⇩𝒯 s⇩2 ∧
  𝒰rd s⇩1 = 𝒰rd s⇩2 ∧ 
  𝒰wr s⇩1 = 𝒰wr s⇩2
"


subsection ‹Well formed 4 state step under the assumption of no speculation from any point›

definition "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4 ⟷ 
  β s⇩1 = β s⇩2 ∧ ω s⇩1 = ω s⇩2 ∧
  β s⇩3 = β s⇩4 ∧ ω s⇩3 = ω s⇩4
"

subsection ‹Well formed 4 state step under the assumption of no speculation from the start of the 
            program›

definition "wfs4_ns s⇩1 s⇩2 s⇩3 s⇩4 ⟷ 
  wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4 ∧
  wfs2_ns s⇩1 s⇩3 ∧ wfs2_ns s⇩2 s⇩4
"


lemma wfs4_s⇩1_s⇩2_ns:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "¬speculating s⇩1 ∧ ¬speculating s⇩2"
  using assms apply (simp add: wfs4_def)
  using wfs2_s⇩n⇩s_ns by blast

lemma wfs4_s⇩1_s⇩2_never_resolves:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
  using assms wfs4_def wfs2_no_resolve_s⇩n⇩s by blast





subsection ‹Step4 function›

fun step⇩4 :: ‹MachineState4Transformer› where
‹step⇩4 (s⇩1, s⇩2, s⇩3, s⇩4) = (
  let (s⇩1', s⇩3') = step⇩2 (s⇩1, s⇩3) in 
  let (s⇩2', s⇩4') = step⇩2 (s⇩2, s⇩4) in 
    (s⇩1', s⇩2', s⇩3', s⇩4')
)›

subsection ‹Constraints for a valid step$_4$ under TPOD›

definition 
  tpod_step⇩4 :: ‹MachineState ⇒ MachineState ⇒ MachineState ⇒ MachineState ⇒ bool›
where
  ‹tpod_step⇩4 s⇩1 s⇩2 s⇩3 s⇩4 ⟷ tpod_step⇩2 s⇩1 s⇩3 ∧ tpod_step⇩2 s⇩2 s⇩4 ∧ s⇩1 ≈⇩ℒ s⇩2 ∧
     ((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩2›

subsection ‹Predicate identifying early violations that will eventually cause an invalid trace. This 
            minimises the state space›

definition "violation\<^sub>\<L> s⇩1 s⇩2 s⇩3 s⇩4 \<longleftrightarrow>
 (length (\<omega> s⇩1) = length (\<omega> s⇩2) \<longrightarrow> (\<omega> s⇩1) = (\<omega> s⇩2)) ∧
 (length (\<beta> s⇩1) = length (\<beta> s⇩2) \<longrightarrow> (\<beta> s⇩1) = (\<beta> s⇩2))
"

subsection ‹Predicate identifying a valid step4 under tpod between the states s to s'›

fun 
  valid_step4 :: "MachineState4 ⇒ MachineState4 ⇒ bool" (infixl "↝⇩4" 101) 
where
"(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') = (
  tpod_step⇩4 s⇩1 s⇩2 s⇩3 s⇩4 ∧ tpod_step⇩4 s⇩1' s⇩2' s⇩3' s⇩4' ∧
  violation\<^sub>\<L> s⇩1 s⇩2 s⇩3 s⇩4 \<and> violation\<^sub>\<L> s⇩1' s⇩2' s⇩3' s⇩4' ∧
  (s⇩1', s⇩2', s⇩3', s⇩4') = step⇩4 (s⇩1, s⇩2, s⇩3, s⇩4) ∧ 
  conformantEntrypoint s⇩1 s⇩1' ∧ conformantEntrypoint s⇩2 s⇩2' ∧ 
  conformantEntrypoint s⇩3 s⇩3' ∧ conformantEntrypoint s⇩4 s⇩4' \<and> 
  (¬speculating s⇩3 \<longrightarrow>  ¬terminating s⇩1) ∧ (¬speculating s⇩4 \<longrightarrow> ¬terminating s⇩2) \<and> 
  ¬terminating s⇩3 ∧ ¬terminating s⇩4
)"

text ‹Rule of equivalence between step2 and step4›

lemma valid_step4_equiv_valid_step2:
  ‹(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<longleftrightarrow> 
    (s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') ∧ (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4') ∧ s⇩1 ≈⇩ℒ s⇩2 \<and> s⇩1' ≈⇩ℒ s⇩2' ∧ 
    ((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩2 ∧ ((op⇩ℒ s⇩1')::low_operation) = op⇩ℒ s⇩2' ∧
    violation\<^sub>\<L> s⇩1 s⇩2 s⇩3 s⇩4 \<and> violation\<^sub>\<L> s⇩1' s⇩2' s⇩3' s⇩4'
  ›
  by (auto simp add: tpod_step⇩4_def)

text ‹Step4 implies two calls of step2›

lemma valid_step4_implies_valid_step2:
  assumes ‹(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')›
    shows ‹(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') ∧ (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')›
  using assms valid_step4_equiv_valid_step2 by blast

lemma step4_s⇩1_s⇩2_no_mispred: 
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"  
  using assms step2_s⇩n⇩s_no_mispred valid_step4_implies_valid_step2 by blast

lemma step4_s⇩1_s⇩2_never_speculates:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4" 
    shows "¬speculating s⇩1' ∧ ¬speculating s⇩2'"
  using assms valid_step4_implies_valid_step2 apply (auto simp only: wfs4_def)
  using step2_wfs2 wfs2_s⇩n⇩s_ns by blast+

lemma step4_pair_eval_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "((\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3)) ∧ 
           ((\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4))"  
  using assms apply (auto simp add: wfs4_def)
  using assms(1) wfs2_eval_equiv apply (meson assms(1) eval_exp_bool.simps valid_step4_implies_valid_step2)
  using step2_ns_eval_equiv_F apply (metis (no_types, hide_lams) Δ⇩0_Δ_no_spec wfs2_def) 
  using wfs2_eval_equiv apply (meson assms(1) eval_exp_bool.simps valid_step4_implies_valid_step2)
  using step2_ns_eval_equiv_F by (meson assms(1) eval_exp_bool.simps valid_step4_implies_valid_step2)

text \<open>Any predicate holds for a non speculative state given its speculative state pair is speculating\<close>

lemma step4_spec_s⇩3_s⇩1_holds: 
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); speculating s⇩3; P s⇩1⟧ ⟹ P s⇩1'"
  by simp

lemma step4_spec_s⇩4_s⇩2_holds: 
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); speculating s⇩4; P s⇩2⟧ ⟹ P s⇩2'"
  by simp

lemma step4_conformantEntrypoint:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')⟧ 
    ⟹ conformantEntrypoint s⇩1 s⇩1' ∧ conformantEntrypoint s⇩2 s⇩2' ∧ conformantEntrypoint s⇩3 s⇩3' 
      ∧ conformantEntrypoint s⇩4 s⇩4'"
  by (auto simp only: valid_step4.simps step2_conformantEntrypoint)

lemma step4_conformantStutter:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')⟧ 
      ⟹ conformantStutter s⇩3 s⇩1 s⇩1' ∧ conformantStutter s⇩4 s⇩2 s⇩2'"
  using valid_step4_implies_valid_step2 step2_conformantStutter by blast

lemma step4_conformantStoreAddr:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); conformantStoreAddr s⇩1; conformantStoreAddr s⇩2⟧
      ⟹ conformantStoreAddr s⇩1' ∧ conformantStoreAddr s⇩2' ∧ conformantStoreAddr s⇩3' ∧ conformantStoreAddr s⇩4'"
  using valid_step4_implies_valid_step2 step2_conformantStoreAddr by blast



subsubsection \<open>Lemmas for resolve\<close>

lemma step4_resolve_s\<^sub>4_spec_n_gt_1:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩4'"
      and "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
    shows "1 < n s⇩4"
  using assms valid_step4_implies_valid_step2
  by (metis Suc_diff_1 Zero_not_Suc dec_n_resolve diff_is_0_eq' speculating_def valid_step2_step verit_comp_simplify1(3))

lemma step4_resolve_s\<^sub>4_ns_n_eq_1:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩4'"
      and "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
    shows "1 = n s⇩4"
  using assms valid_step4_implies_valid_step2
  by (metis n_1_resolve_nospec never_resolve_not_speculating valid_step2_s⇩s_always_step)


lemma step4_resolve_s\<^sub>3_spec_n_gt_1:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩3'"
      and "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
    shows "1 < n s⇩3"
  using assms valid_step4_implies_valid_step2
  by (metis Suc_le_lessD dec_n_resolve diff_is_0_eq' gr_implies_not0 not_less_eq_eq speculating_def valid_step2_s⇩s_always_step)

lemma step4_resolve_s\<^sub>3_ns_n_eq_1:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3'"
      and "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
    shows "1 = n s⇩3"
  using assms valid_step4_implies_valid_step2
  by (metis n_1_resolve_nospec resolve_must_be_speculating valid_step2_s⇩s_always_step)
 
lemma step4_not_resolve_s\<^sub>3_ns_halting_specfence:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩3 ∧ ¬ speculating s⇩3'"
      and "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
    shows "¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 = SpecFence"
  using assms valid_step4_implies_valid_step2 
  by (metis always_resolve_on_halt const_n speculating_def step2_ns_s⇩s'_no_mispred valid_step2_step)


lemma step4_not_resolve_s\<^sub>4_ns_halting_specfence:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩4 ∧ ¬ speculating s⇩4'"
      and "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
    shows "¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 = SpecFence"
  using assms valid_step4_implies_valid_step2
  by (metis always_resolve_on_halt const_n speculating_def step2_ns_s⇩s'_no_mispred valid_step2_step)

lemma step4_not_resolve_s\<^sub>3_spec_halting_not_specfence:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩3'"
      and "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
    shows "¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 \<noteq> SpecFence"
  using assms valid_step4_implies_valid_step2 
  by (metis always_resolve_on_halt less_nat_zero_code n_0_fence no_valid_transition_halt speculating_def valid_step2_s⇩s_always_step)

lemma step4_not_resolve_s\<^sub>4_spec_halting_not_specfence:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩4'"
      and "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
    shows "¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 \<noteq> SpecFence"
  using assms valid_step4_implies_valid_step2 
  by (metis always_resolve_on_halt less_nat_zero_code n_0_fence no_valid_transition_halt speculating_def valid_step2_s⇩s_always_step)

lemma wfs4_no_resolve_s⇩1_s⇩2:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "¬resolve (n s⇩1) (β s⇩1) (pc s⇩1) \<and> ¬resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
  using wfs2_no_resolve_s⇩n⇩s assms wfs4_def by blast

lemma step4_ns_no_resolve_s⇩3:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3"
    shows "¬resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
  using assms resolve_must_be_speculating valid_step2_s⇩s_always_step valid_step4_implies_valid_step2 by blast

lemma step4_ns_no_resolve_s⇩4:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩4"
    shows "¬resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
  using assms resolve_must_be_speculating valid_step2_s⇩s_always_step valid_step4_implies_valid_step2 by blast

lemma step4_ns_no_resolve:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
    shows "¬resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬resolve (n s⇩2) (β s⇩2) (pc s⇩2) \<and> 
           ¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
  using assms step4_ns_no_resolve_s⇩3 step4_ns_no_resolve_s⇩4 by (meson wfs4_no_resolve_s⇩1_s⇩2)



















lemma step4_ns_\<iota>_s12_same:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
    shows "\<iota> s⇩1 = \<iota> s⇩2"
  by (metis assms(3) assms(4) ns_pc⇩0_ρ_equal wfs4_def wfs4_s⇩1_s⇩2_ns)

lemma step4_s12_eval_e_same:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
    shows "(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1)) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
  using assms by (metis Δ⇩0_Δ_no_spec wfs4_s⇩1_s⇩2_ns)

lemma step4_s\<^sub>3_ns_s_branch:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3 ∧ speculating s⇩3'"
    shows "\<exists> e c. \<iota> s⇩3 = Branch e c"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step by blast  

lemma step4_s\<^sub>4_ns_s_branch:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩4 ∧ speculating s⇩4'"
    shows "\<exists> e c. \<iota> s⇩4 = Branch e c"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step by blast

lemma step4_s\<^sub>3_ns_s_mispred:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3 ∧ speculating s⇩3'"
    shows "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step 
        speculating_mispred by blast

lemma step4_s\<^sub>4_ns_s_mispred:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩4 ∧ speculating s⇩4'"
    shows "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step 
        speculating_mispred by blast


lemma step4_s\<^sub>3_ns_ns_no_mispred:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩3'"
      and "\<iota> s⇩3 = Branch e c"
    shows "¬mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step 
  by (meson not_speculating_no_mispred resolve_must_be_speculating)


lemma step4_s\<^sub>4_ns_ns_no_mispred:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩4 ∧ ¬speculating s⇩4'"
      and "\<iota> s⇩4 = Branch e c"
    shows "¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
  using assms valid_step4_implies_valid_step2 speculating_branch valid_step2_s⇩s_always_step 
  by (meson not_speculating_no_mispred resolve_must_be_speculating)


lemma step4_spec_retains_lo: 
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∨ speculating s⇩4"
    shows "(((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1' ∧ ((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2' ∧ ((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3' ∧ ((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4')"
  using assms valid_step4_implies_valid_step2 by (metis (no_types, lifting) step2_spec_retains_lo_hi tpod_step⇩2_def tpod_step⇩4_def valid_step4.simps wfs4_def)


lemma step4_spec_retains_hi_pairs: 
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "(speculating s⇩3 ⟶ (((op⇩ℋ s⇩1)::low_operation) = op⇩ℋ s⇩1' ∧ ((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3')) ∧
           (speculating s⇩4 ⟶ (((op⇩ℋ s⇩2)::low_operation) = op⇩ℋ s⇩2' ∧ ((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'))"
  using assms
  by (metis (mono_tags, lifting) conformantStutter_def step2_spec_retains_lo_hi step4_conformantStutter valid_step4_implies_valid_step2 wfs4_def)


lemma step4_spec_retains_lo_s_s': 
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∨ speculating s⇩4"
    shows "(op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩2 ≠ \<bottom> ∧ op⇩ℒ s⇩3 ≠ \<bottom> ∧ op⇩ℒ s⇩4 ≠ \<bottom>) ⟷ (op⇩ℒ s⇩1' ≠ \<bottom> ∧ op⇩ℒ s⇩2' ≠ \<bottom> ∧ op⇩ℒ s⇩3' ≠ \<bottom> ∧ op⇩ℒ s⇩4' ≠ \<bottom>)"
  using assms step4_spec_retains_lo by presburger







subsubsection \<open>Lemmas for low equivalence TODO RENAME\<close>

lemma wfs4_ι⇩0_same:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)" and "¬halting (pc⇩0 s⇩2) (\<Pi> s⇩2) ∧ ¬halting (pc⇩0 s⇩3) (\<Pi> s⇩3) ∧ ¬halting (pc⇩0 s⇩4) (\<Pi> s⇩4)"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "ι⇩0 s⇩1 = ι⇩0 s⇩2 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩3 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩4"
  using assms apply (simp add: wfs4_def tpod_step\<^sub>4_def)
  apply (elim conjE)
  using wfs2_no_halt_ι⇩0_same by (smt (z3) op⇩ℒ_sub_def option.distinct(1))

lemma step4_ι⇩0_same:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)" and "¬halting (pc⇩0 s⇩2) (\<Pi> s⇩2) ∧ ¬halting (pc⇩0 s⇩3) (\<Pi> s⇩3) ∧ ¬halting (pc⇩0 s⇩4) (\<Pi> s⇩4)"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "ι⇩0 s⇩1 = ι⇩0 s⇩2 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩3 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩4"
  using assms wfs4_ι⇩0_same apply (simp only: valid_step4.simps)
  by blast

lemma wfs4_no_spec_ι_same:
  assumes "¬speculating s⇩3"
      and "¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1) ∧ ¬halting (pc⇩0 s⇩2) (\<Pi> s⇩2) ∧ ¬halting (pc⇩0 s⇩3) (\<Pi> s⇩3) ∧ ¬halting (pc⇩0 s⇩4) (\<Pi> s⇩4)"
      and "op⇩ℒ s⇩1 ≠ \<bottom>" (*showing that we are in the low state*)
    shows "ι s⇩1 = ι s⇩2 ∧ ι s⇩1 = ι s⇩3 ∧ ι s⇩1 = ι s⇩4"
proof -
  have ι_pair_same: "ι s⇩1 = ι s⇩3 ∧ ι s⇩2 = ι s⇩4"
    using assms apply (auto simp only: wfs4_def)
     apply (metis ι⇩0_ι_no_spec wfs2_no_halt_ι⇩0_same wfs2_s⇩n⇩s_ns)
    using ι⇩0_ι_no_spec wfs2_no_halt_ι⇩0_same wfs2_s⇩n⇩s_ns by auto
  have ι⇩0_same: "ι⇩0 s⇩1 = ι⇩0 s⇩2"
    using assms wfs4_ι⇩0_same by blast
  have ι_ns_same: "ι s⇩1 = ι s⇩2"
    using assms apply (auto simp add: wfs4_def wfs2_def speculating_def)
    by (metis ι⇩0_same)
  then show ?thesis
    using ι_ns_same ι_pair_same by auto
qed

lemma step4_no_spec_ι_same:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3"
      and "¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1) ∧ ¬halting (pc⇩0 s⇩2) (\<Pi> s⇩2) ∧ ¬halting (pc⇩0 s⇩3) (\<Pi> s⇩3) ∧ ¬halting (pc⇩0 s⇩4) (\<Pi> s⇩4)"
      and "op⇩ℒ s⇩1 ≠ \<bottom>" (*showing that we are in the low state*)
    shows "ι s⇩1 = ι s⇩2 ∧ ι s⇩1 = ι s⇩3 ∧ ι s⇩1 = ι s⇩4"
  using assms wfs4_no_spec_ι_same apply (simp only: valid_step4.simps)
  by blast

lemma wfs4_ns_opL_Δ⇩0_equal:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
proof (cases "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)")
  case True
  then show ?thesis
    using assms apply (auto simp add: wfs4_def wfs2_def tpod_step\<^sub>4_def)
    using assms(1) assms(2) low_equiv_sub_def op⇩ℒ_not_inst⇩𝒯_s wfs4_s⇩1_s⇩2_ns by auto
next
  case False
  then show ?thesis
    using assms(2) assms(3) op⇩ℋ⇩ℒ_Halt_None tpod_step⇩4_def by force
qed

lemma wfs4_ns_opL_\<mu>⇩0_equal:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "∀a. a ∈ 𝒰rd s⇩1 ⟶
             μ⇩0⇩r⇩e⇩a⇩d s⇩1 a =
             μ⇩0⇩r⇩e⇩a⇩d s⇩2 a"
proof (cases "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)")
  case True
  then show ?thesis
    using assms apply (auto simp add: wfs4_def wfs2_def tpod_step\<^sub>4_def)
    using assms(1) assms(2) low_equiv_sub_def op⇩ℒ_not_inst⇩𝒯_s wfs4_s⇩1_s⇩2_ns by auto
next
  case False
  then show ?thesis
    using assms(2) assms(3) op⇩ℋ⇩ℒ_Halt_None tpod_step⇩4_def by force
qed

lemma wfs4_ns_opL_\<beta>_equal:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "\<beta> s⇩1 = \<beta> s⇩2"
proof (cases "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)")
  case True
  then show ?thesis
    using assms apply (auto simp add: wfs4_def wfs2_def tpod_step\<^sub>4_def)
    using assms(1) assms(2) low_equiv_sub_def op⇩ℒ_not_inst⇩𝒯_s wfs4_s⇩1_s⇩2_ns by auto
next
  case False
  then show ?thesis
    using assms(2) assms(3) op⇩ℋ⇩ℒ_Halt_None tpod_step⇩4_def by force
qed

lemma wfs4_ns_opL_\<omega>_equal:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "\<omega> s⇩1 = \<omega> s⇩2"
proof (cases "¬halting (pc⇩0 s⇩1) (\<Pi> s⇩1)")
  case True
  then show ?thesis
    using assms apply (auto simp add: wfs4_def wfs2_def tpod_step\<^sub>4_def)
    using assms(1) assms(2) low_equiv_sub_def op⇩ℒ_not_inst⇩𝒯_s wfs4_s⇩1_s⇩2_ns by auto
next
  case False
  then show ?thesis
    using assms(2) assms(3) op⇩ℋ⇩ℒ_Halt_None tpod_step⇩4_def by force
qed

lemma step4_ns_opL_low_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> \<or> op⇩ℒ s⇩2 ≠ \<bottom>"
    shows "\<omega> s⇩1 = \<omega> s⇩2 \<and> 
          \<beta> s⇩1 = \<beta> s⇩2 \<and>
          (∀a. a ∈ 𝒰rd s⇩1 ⟶
             μ⇩0⇩r⇩e⇩a⇩d s⇩1 a =
             μ⇩0⇩r⇩e⇩a⇩d s⇩2 a) \<and>
          Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
  using assms low_equiv_s_not_low valid_step4_equiv_valid_step2 wfs4_s⇩1_s⇩2_ns by blast

lemma wfs4_ns_pair_eval_equiv:
  assumes "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom>"
    shows "(\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2)"
  using assms by (metis Δ⇩0_Δ_no_spec wfs4_ns_opL_Δ⇩0_equal wfs4_s⇩1_s⇩2_ns)

lemma step4_ns_pair_eval_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom>"
    shows "(\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2)"
  using assms wfs4_ns_pair_eval_equiv by simp

lemma wfs4_op⇩ℒ_eval_equiv:
  assumes "¬speculating s⇩3 ∧ ¬speculating s⇩4"
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 \<noteq> \<bottom>"
    shows "(\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2) ∧
           (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3) ∧
           (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4)"
  using assms wfs4_ns_pair_eval_equiv step4_pair_eval_equiv by blast

lemma step4_op⇩ℒ_eval_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 \<noteq> \<bottom>"
    shows "(\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2) ∧
           (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3) ∧
           (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4)"
  using assms wfs4_op⇩ℒ_eval_equiv by simp

















lemma step4_spec_retains_not_lo_s_s': 
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∨ speculating s⇩4"
    shows "(op⇩ℒ s⇩1 = \<bottom> ∧ op⇩ℒ s⇩2 = \<bottom> ∧ op⇩ℒ s⇩3 = \<bottom> ∧ op⇩ℒ s⇩4 = \<bottom>) ⟷ (op⇩ℒ s⇩1' = \<bottom> ∧ op⇩ℒ s⇩2' = \<bottom> ∧ op⇩ℒ s⇩3' = \<bottom> ∧ op⇩ℒ s⇩4' = \<bottom>)"  
  using assms step4_spec_retains_lo by presburger

lemma step4_op⇩ℒ_spec_diff_invalid:
  assumes "speculating s⇩3 \<noteq> speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "¬(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
proof -
  have s3_eq: "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow> speculating s⇩3 \<Longrightarrow> 
          Δ⇩0 s⇩1' = Δ⇩0 s⇩1 ∧ 
          (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1' a = μ⇩0⇩r⇩e⇩a⇩d s⇩1 a) ∧ 
          ω s⇩1' = ω s⇩1 ∧ 
          β s⇩1' = β s⇩1"
    by auto
  have s4_eq: "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow> speculating s⇩4 \<Longrightarrow>
          Δ⇩0 s⇩2' = Δ⇩0 s⇩2 ∧ 
          (∀a . a ∈ 𝒰rd s⇩2 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩2' a = μ⇩0⇩r⇩e⇩a⇩d s⇩2 a) ∧ 
          ω s⇩2' = ω s⇩2 ∧ 
          β s⇩2' = β s⇩2"
    by auto
  have s1s2: "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow>
           Δ⇩0 s⇩2 = Δ⇩0 s⇩1 ∧ 
           (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩2 a = μ⇩0⇩r⇩e⇩a⇩d s⇩1 a) ∧ 
           ω s⇩2 = ω s⇩1 ∧ 
           β s⇩2 = β s⇩1"
    using assms(2) apply (simp add: wfs4_def)     
    using assms(2) assms(3) wfs4_ns_opL_Δ⇩0_equal wfs4_ns_opL_β_equal wfs4_ns_opL_μ⇩0_equal wfs4_ns_opL_ω_equal by auto
  have s1's2': "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow>
           Δ⇩0 s⇩2' = Δ⇩0 s⇩1' ∧ 
           (∀a . a ∈ 𝒰rd s⇩1' ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩2' a = μ⇩0⇩r⇩e⇩a⇩d s⇩1' a) ∧ 
           ω s⇩2' = ω s⇩1' ∧ 
           β s⇩2' = β s⇩1'"
    apply (subgoal_tac "s⇩1 ≈⇩ℒ s⇩2")
    using low_equiv_s_not_low
    apply (metis assms(2) assms(3) step4_s⇩1_s⇩2_never_speculates valid_step4_equiv_valid_step2)
    using valid_step4_equiv_valid_step2 by blast
  have s3_s1s2s1's2': "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow> speculating s⇩3 \<Longrightarrow> 
          Δ⇩0 s⇩1 = Δ⇩0 s⇩1' ∧ Δ⇩0 s⇩1 = Δ⇩0 s⇩2 ∧ Δ⇩0 s⇩1 = Δ⇩0 s⇩2' ∧ 
          (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩1' a) ∧ (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩2 a) ∧ (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩2' a) ∧ 
          ω s⇩1 = ω s⇩1' ∧ ω s⇩1 = ω s⇩2' ∧ ω s⇩1 = ω s⇩2 ∧ 
          β s⇩1 = β s⇩1' ∧ β s⇩1 = β s⇩2' ∧ β s⇩1 = β s⇩2"
    using s3_eq s1s2 s1's2' by auto
  have s4_s1s2s1's2': "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4') \<Longrightarrow> speculating s⇩4 \<Longrightarrow>
          Δ⇩0 s⇩1 = Δ⇩0 s⇩1' ∧ Δ⇩0 s⇩1 = Δ⇩0 s⇩2 ∧ Δ⇩0 s⇩1 = Δ⇩0 s⇩2' ∧ 
          (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩1' a) ∧ (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩2 a) ∧ (∀a . a ∈ 𝒰rd s⇩1 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1 a = μ⇩0⇩r⇩e⇩a⇩d s⇩2' a) ∧ 
          ω s⇩1 = ω s⇩1' ∧ ω s⇩1 = ω s⇩2' ∧ ω s⇩1 = ω s⇩2 ∧ 
          β s⇩1 = β s⇩1' ∧ β s⇩1 = β s⇩2' ∧ β s⇩1 = β s⇩2"
    using s4_eq s1s2 s1's2' by (metis conformantStutter_def step2_𝒰rd_const step4_conformantStutter valid_step4_implies_valid_step2)
  show ?thesis
    apply (rule notI)
    apply (cases "speculating s⇩3")
    apply (cases "speculating s⇩4")
    using assms(1) apply blast
    defer
    apply (cases "speculating s⇩4")
    using s4_s1s2s1's2' low_equiv_props_no_step apply (metis valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)
    using assms(1) apply blast
    apply (subgoal_tac "(∀a. a ∈ 𝒰rd s⇩2 ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩2' a = μ⇩0⇩r⇩e⇩a⇩d s⇩2 a)")
    apply (subgoal_tac "Δ⇩0 s⇩2' = Δ⇩0 s⇩2 \<and> ω s⇩2' = ω s⇩2 \<and> β s⇩2' = β s⇩2")
    using s3_s1s2s1's2' low_equiv_props_no_step valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2 
    apply meson 
    apply (metis s3_s1s2s1's2')
    using assms(2) s3_s1s2s1's2' wfs4_def by fastforce
qed


lemma step4_op⇩ℒ_spec_implied:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "speculating s⇩3 \<longleftrightarrow> speculating s⇩4"
  by (meson assms(1) assms(2) assms(3) step4_op⇩ℒ_spec_diff_invalid)

lemma step4_op⇩ℒ_to_op⇩ℋ⇩s_equal_pc⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬ speculating s⇩3 ∧ ¬ speculating s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom>"
      and "inst⇩𝒯 s⇩1' ≠ None ∧ inst⇩𝒯 s⇩2' ≠ None ∧ inst⇩𝒯 s⇩3' ≠ None ∧ inst⇩𝒯 s⇩4' ≠ None"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩2' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩4'"
proof - 
  have inst⇩𝒯s: "inst⇩𝒯 s⇩1 = None ∧ inst⇩𝒯 s⇩2 = None ∧ inst⇩𝒯 s⇩3 = \<bottom> ∧ inst⇩𝒯 s⇩4 = None"
    using assms by (metis (no_types, lifting) inst⇩𝒯_not_op⇩ℒ_s tpod_step⇩2_def tpod_step⇩4_def valid_step4.simps)
  show ?thesis
    using assms apply (simp add: valid_step_def conformantEntrypoint_def wfs4_def wfs2_def)
    by (metis inst⇩𝒯s)
qed




lemma valid_step4_nospec_cases [consumes 4]:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
obtains
  (* 7 non speculative cases *)
  (NS_SpecFence) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = SpecFence"
          "ι s⇩2 = SpecFence"
          "ι s⇩3 = SpecFence"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_TPred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_NTPred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_Goto) c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Goto c"
          "ι s⇩2 = Goto c"
          "ι s⇩3 = Goto c"
          "ι s⇩4 = Goto c"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_RegisterUpdate) r e where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = RegisterUpdate r e"
          "ι s⇩2 = RegisterUpdate r e"
          "ι s⇩3 = RegisterUpdate r e"
          "ι s⇩4 = RegisterUpdate r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_Load) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Load r e"
          "ι s⇩2 = Load r e"
          "ι s⇩3 = Load r e"
          "ι s⇩4 = Load r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_Store) e⇩1 e⇩2 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Store e⇩1 e⇩2"
          "ι s⇩2 = Store e⇩1 e⇩2"
          "ι s⇩3 = Store e⇩1 e⇩2"
          "ι s⇩4 = Store e⇩1 e⇩2"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  | (NS_NoOp)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = NoOp"
          "ι s⇩2 = NoOp"
          "ι s⇩3 = NoOp"
          "ι s⇩4 = NoOp"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
  (* 2 additional cases leading to violations of secure spec *)
  | (NS_High_Trans)
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = ι s⇩3"
          "ι s⇩2 = ι s⇩4"
          "ι s⇩1 ≠ ι s⇩2"
  | (NS_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False) e c where
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
  | (NS_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True) e c where
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
proof (cases "op⇩ℒ s⇩1 = \<bottom> ⟶ (ι s⇩1 = ι s⇩2 ∧ Δ⇩0 s⇩1 = Δ⇩0 s⇩2)")
  have precons: "¬ halting⇩p⇩r⇩e⇩d s⇩1 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩2 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧
                 ¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧ ¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ∧
                 valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4') ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧
                 wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3"
    using assms apply auto
    apply (meson terminating_def wfs4_s⇩1_s⇩2_ns)+
    apply (meson wfs4_s⇩1_s⇩2_never_resolves)+
    using assms(1) step4_ns_no_resolve tpod_step⇩4_def wfs4_def by blast+
  case True
  then show ?thesis
    using assms(1) proof (simp only: valid_step4.simps)
    have matching_delta: "Δ⇩0 s⇩1 = Δ⇩0 s⇩3 ∧ Δ⇩0 s⇩2 = Δ⇩0 s⇩4"
      using assms(4) by (simp add: wfs4_def wfs2_def)
    have well_formed: "wfs s⇩1 ∧ wfs s⇩2 ∧ wfs s⇩3 ∧ wfs s⇩4"
      using assms(4) by (simp add: wfs4_def wfs2_def)
    have same_ι: "ι s⇩1 = ι s⇩3 ∧ ι s⇩1 = ι s⇩2 ∧ ι s⇩1 = ι s⇩4"
      using True assms(1,3,4) step2_nospec_pc⇩0_NeverHalts step2_ns_ι_same wfs4_def 
            step4_no_spec_ι_same valid_step4_implies_valid_step2 by (smt (z3))
    have nospec: "¬speculating s⇩1 ∧ ¬speculating s⇩2"
      using precons wfs2_s⇩n⇩s_ns by blast
    have nomispred: "⋀e c. (ι s⇩3 = Branch e c ∧ ι s⇩4 = Branch e c) \<longrightarrow> ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
      using assms(1-2) valid_step4_implies_valid_step2 precons step2_ns_s⇩s'_no_mispred by blast
    have wfs2: "wfs2 s⇩1 s⇩3 ∧ wfs2 s⇩2 s⇩4"
      using assms(4) wfs4_def by blast
    have "⋀e. (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1) ⟷ (\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2)"
      using True assms(1, 4) by (metis step4_ns_pair_eval_equiv step4_s12_eval_e_same)
    from this that show ?thesis
      using assms(1) precons
      by (metis Δ⇩0_Δ_no_spec ι_some_Instr assms(3) matching_delta nomispred nospec same_ι step4_s⇩1_s⇩2_no_mispred)
  qed
next
  have precons: "¬ halting⇩p⇩r⇩e⇩d s⇩1 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩2 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧
                 ¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧ ¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ∧
                 valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4') ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧
                 wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3 ∧ ι s⇩1 = ι s⇩3 ∧ ι s⇩2 = ι s⇩4"
    using assms apply auto
    apply (meson terminating_def wfs4_s⇩1_s⇩2_ns)+
    apply (meson wfs4_s⇩1_s⇩2_never_resolves)+
    apply (meson assms(1) step4_ns_no_resolve_s⇩3)
    apply (metis assms(1) step4_ns_no_resolve_s⇩4)
    using tpod_step⇩4_def apply blast+
    apply (simp_all add: wfs4_def)
    using assms(1) step2_ns_ι_same valid_step4_implies_valid_step2 by metis+

  have eval_pairs13: "⋀e. (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1)) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))" (* lemma *)
    using assms(2-4) apply (auto simp add: wfs4_def wfs2_def as_bool_def)
    by (metis (full_types) not_gr0 speculating_def)+
  have eval_pairs24: "⋀e. (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))" (* lemma *)
    using assms(2-4) apply (auto simp add: wfs4_def wfs2_def as_bool_def)
    by (metis (full_types) not_gr0 speculating_def)+
  have eval_pairs12: "⋀e. (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))" (* lemma *)
    using assms(2-4) apply (auto simp add: wfs4_def wfs2_def as_bool_def)
    by (metis (full_types) not_gr0 speculating_def)+
  case False
  from this that show ?thesis
    using assms precons by (metis ι_some_Instr eval_pairs12 eval_pairs13 step2_ns_s⇩s'_no_mispred 
                                  step4_s⇩1_s⇩2_no_mispred)
qed

text ‹No speculation to speculation›

lemma valid_step4_nospec_spec_cases [consumes 4]:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩3' ∧ speculating s⇩4'"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
obtains
  (NS_TMispred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"

  | (NS_NTMispred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"

  (* Suspicious states *)
  | (NS_Mispred_diff_ι) e c ea ca where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch ea ca"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch ea ca"
          "ι s⇩1 = ι s⇩3"
          "ι s⇩2 = ι s⇩4"
          "ι s⇩1 ≠ ι s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (NS_Mispred_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (NS_Mispred_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
proof -
  have precons: 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧
           ¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ∧ 
           ¬ halting⇩p⇩r⇩e⇩d s⇩1 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩2 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧
           ¬mispred (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬mispred (n s⇩2) (β s⇩2) (pc s⇩2) ∧ mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧
           valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4') ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧
           wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3"
    using assms apply auto
    using wfs4_s⇩1_s⇩2_never_resolves apply blast+
    using resolve_must_be_speculating wfs4_s⇩1_s⇩2_ns apply (simp_all add: wfs4_def)
    apply (meson assms(1) step4_ns_no_resolve_s⇩3)
    apply (meson assms(1) step4_ns_no_resolve_s⇩4)
    apply (meson assms(1) no_valid_transition_halt valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)
    apply (meson assms(1) no_valid_transition_halt valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)
    apply (metis assms(1) step4_not_resolve_s⇩3_spec_halting_not_specfence step4_ns_no_resolve_s⇩3)
    apply (metis assms(1) step4_not_resolve_s⇩4_spec_halting_not_specfence step4_ns_no_resolve_s⇩4)
    apply (metis assms(1) step4_s⇩1_s⇩2_no_mispred)
    apply (metis assms(1) step4_s⇩1_s⇩2_no_mispred)
    using assms(1) step4_s⇩4_ns_s_mispred step4_s⇩3_ns_s_mispred tpod_step⇩4_def by meson+
  have Instr_branch:  "∃e1 c1 e2 c2. ι s⇩3 = Branch e1 c1 ∧ ι s⇩4 = Branch e2 c2"
    using precons assms(1-3) by (meson step4_s⇩3_ns_s_branch step4_s⇩4_ns_s_branch)
  show ?thesis
    apply (subgoal_tac "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')")
    defer
    using assms(1) apply blast
    apply (drule valid_step4_implies_valid_step2)
    apply (elim conjE)
    apply (erule valid_step2_cases)
    apply (meson assms(4) wfs4_def)
    using assms(4) apply (simp only: wfs4_def)
    using assms(2-4) apply blast+
    apply (erule valid_step2_cases)
    using precons apply blast+
    using assms(2) apply blast+
    apply (cases "ι s⇩1 = ι s⇩2")
    apply (metis NS_TMispred Instr.inject(4) precons)
    using assms(1) step4_ι⇩0_same apply (smt (z3) NS_Mispred_diff_ι assms(4) halting⇩p⇩r⇩e⇩d_def pc⇩0_pc_no_spec precons wfs2_s⇩n⇩s_ns)
    apply (cases "ι s⇩1 = ι s⇩2")
    using precons apply (metis Instr.inject(4) NS_Mispred_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False assms(1) assms(4) step4_ns_pair_eval_equiv step4_s12_eval_e_same)    
    using assms(1) step4_no_spec_ι_same apply (smt (z3) NS_Mispred_diff_ι assms(4) halting⇩p⇩r⇩e⇩d_def pc⇩0_pc_no_spec precons wfs2_s⇩n⇩s_ns wfs4_ι⇩0_same)
    using assms(2, 3) apply blast+
    apply (erule valid_step2_cases)
    using precons apply blast+
    using assms(2) apply blast+
    apply (cases "ι s⇩1 = ι s⇩2")
    using assms(1) apply (metis Instr.inject(4) NS_Mispred_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True assms(4) precons step4_ns_pair_eval_equiv step4_s12_eval_e_same)
    using assms(1) step4_no_spec_ι_same apply (metis NS_Mispred_diff_ι assms(4) halting⇩p⇩r⇩e⇩d_def pc⇩0_pc_no_spec precons wfs2_s⇩n⇩s_ns)
    apply (cases "ι s⇩1 = ι s⇩2")
    apply (metis NS_NTMispred Instr.inject(4) precons)
    using assms(1) step4_no_spec_ι_same apply (metis NS_Mispred_diff_ι assms(4) halting⇩p⇩r⇩e⇩d_def pc⇩0_pc_no_spec precons wfs2_s⇩n⇩s_ns)
    using assms(2-3) by blast+
qed


lemma valid_step4_spec_nospec_cases [consumes 4]:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
obtains
  (S_SpecFence) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = SpecFence"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve_NS)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩3 = 1"
          "n s⇩4 = 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  (* suspicious states *)
  | (S_SpecFence_Resolve) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩3 = 1"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve_SpecFence)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "ι s⇩3 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩4 = 1"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
proof -
  have same_ns_s: "s⇩1' = s⇩1 ∧ s⇩2' = s⇩2"
    using assms(1) assms(3) by auto
  have common: "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)
              ∧ wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3
              ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧ valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')
              ∧ wfs2 s⇩1 s⇩3 ∧ wfs2 s⇩2 s⇩4
              ∧ ((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1' ∧ ((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2' ∧ ((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3' ∧ ((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4' 
              ∧ ((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1' ∧ ((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2' ∧ ((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3' ∧ ((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"    
    apply (intro conjI)
    using assms(4) wfs4_no_resolve_s⇩1_s⇩2 apply blast+
    using assms(1, 4) apply (simp_all only: valid_step4.simps wfs4_def same_ns_s)
    using assms(1, 3) step2_spec_retains_lo_hi same_ns_s valid_step4_implies_valid_step2 by blast+
  have resolve_s3: "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ n s⇩3 = 1"
    using assms(1-3) n_1_resolve_nospec by (metis step4_resolve_s⇩3_ns_n_eq_1)
  have resolve_s4: "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ n s⇩4 = 1"
    using assms(1-3) n_1_resolve_nospec by (metis step4_resolve_s⇩4_ns_n_eq_1)
  show ?thesis
    apply (subgoal_tac "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')")
    defer using assms(1) apply blast
    apply (drule valid_step4_implies_valid_step2)
    using common
    apply (elim conjE)
    apply (erule valid_step2_cases)
    using assms(3) apply blast+
    apply (erule valid_step2_cases)
    using assms(3) apply blast+
    using S_SpecFence common apply blast
    using assms(2) apply blast+
    using S_Resolve_SpecFence common apply fastforce
    using assms(2) apply blast+
    apply (erule valid_step2_cases)
    using assms(3) apply blast+
    using S_SpecFence_Resolve common apply fastforce
    using assms(2) apply blast+    
    using S_Resolve_NS common apply blast
    using assms(2) by blast+
qed


lemma valid_step4_spec_spec_cases [consumes 4]:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "speculating s⇩3' ∧ speculating s⇩4'"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
obtains
  (S_RegisterUpdate) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = RegisterUpdate r e"
          "ι s⇩4 = RegisterUpdate r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Load) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Load r e"
          "ι s⇩4 = Load r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Store) e⇩1 e⇩2 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Store e⇩1 e⇩2"
          "ι s⇩4 = Store e⇩1 e⇩2"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_TPred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_TMispred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NTPred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NTMispred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Goto) c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Goto c"
          "ι s⇩4 = Goto c"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NoOp)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = NoOp"
          "ι s⇩4 = NoOp"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "n s⇩3 > 1"
          "n s⇩4 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Other)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 ≠ ι s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_BranchEqPredNeqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(n s⇩3 = n s⇩4 ∧ Δ s⇩3 = Δ s⇩4)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_BranchNeqPredEqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4))"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_BranchNeqPredNeqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(n s⇩3 = n s⇩4 ∧ Δ s⇩3 = Δ s⇩4)"
          "¬(mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4))"
          "¬(n s⇩3 = n s⇩4 \<and> β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve3)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "n s⇩3 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve4)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "n s⇩4 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"

proof -
  have same_ns_s: "s⇩1' = s⇩1 ∧ s⇩2' = s⇩2"
    using assms(1) assms(3) by auto
  have common: "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)
              ∧ wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3
              ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧ valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')
              ∧ wfs2 s⇩1 s⇩3 ∧ wfs2 s⇩2 s⇩4 
              ∧ speculating s⇩3 ∧ speculating s⇩4
              ∧ ((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1' ∧ ((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2' ∧ ((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3' ∧ ((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4' 
              ∧ ((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1' ∧ ((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2' ∧ ((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3' ∧ ((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
    using assms(1) assms(3) assms(4) step4_spec_retains_hi_pairs step4_spec_retains_lo valid_step4.simps wfs4_def wfs4_no_resolve_s⇩1_s⇩2
    by (metis (no_types, lifting) same_ns_s valid_step4_implies_valid_step2)
  have resolve_s3: "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ n s⇩3 > 1"
    using assms(1-3) by (meson step4_resolve_s⇩3_spec_n_gt_1)
  have resolve_s4: "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ n s⇩4 > 1"
    using assms(1-3) by (meson step4_resolve_s⇩4_spec_n_gt_1)
  show ?thesis
    apply (subgoal_tac "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')")
    defer using assms(1) apply blast
    apply (drule valid_step4_implies_valid_step2)
    using common always_resolve_on_halt apply (elim conjE)
    apply (cases "ι s⇩3 = ι s⇩4")
    defer
    apply (metis S_Other S_Resolve S_Resolve3 S_Resolve4 resolve_s3 resolve_s4)    
    apply (cases "resolve (n s⇩4) (β s⇩4) (pc s⇩4)")
    apply (metis S_Resolve S_Resolve4 resolve_s3 resolve_s4)
    apply (erule valid_step2_cases)
    apply blast+
    using assms(2) apply blast+
    using valid_step2_s⇩s_always_step S_RegisterUpdate apply metis
    using valid_step2_s⇩s_always_step S_Load           apply metis
    using valid_step2_s⇩s_always_step S_Store          apply metis
    using valid_step2_s⇩s_always_step S_BranchEqPredNeqEval S_BranchNeqPredEqEval S_BranchNeqPredNeqEval S_TPred apply metis
    using valid_step2_s⇩s_always_step S_BranchEqPredNeqEval S_BranchNeqPredEqEval S_BranchNeqPredNeqEval S_TMispred apply metis
    using valid_step2_s⇩s_always_step S_BranchEqPredNeqEval S_BranchNeqPredEqEval S_BranchNeqPredNeqEval S_NTPred apply metis
    using valid_step2_s⇩s_always_step S_BranchEqPredNeqEval S_BranchNeqPredEqEval S_BranchNeqPredNeqEval S_NTMispred apply metis 
    using valid_step2_s⇩s_always_step S_Goto apply metis
    using valid_step2_s⇩s_always_step S_NoOp apply metis
    using assms(2) apply blast
    by (metis S_Resolve3)
qed


lemma valid_step4_cases [consumes 2]:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
obtains
  (* 7 non speculative cases *)
  (NS_SpecFence) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = SpecFence"
          "ι s⇩2 = SpecFence"
          "ι s⇩3 = SpecFence"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_TPred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_NTPred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_Goto) c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Goto c"
          "ι s⇩2 = Goto c"
          "ι s⇩3 = Goto c"
          "ι s⇩4 = Goto c"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_RegisterUpdate) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = RegisterUpdate r e"
          "ι s⇩2 = RegisterUpdate r e"
          "ι s⇩3 = RegisterUpdate r e"
          "ι s⇩4 = RegisterUpdate r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_Load) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Load r e"
          "ι s⇩2 = Load r e"
          "ι s⇩3 = Load r e"
          "ι s⇩4 = Load r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_Store) e⇩1 e⇩2 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Store e⇩1 e⇩2"
          "ι s⇩2 = Store e⇩1 e⇩2"
          "ι s⇩3 = Store e⇩1 e⇩2"
          "ι s⇩4 = Store e⇩1 e⇩2"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_NoOp)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = NoOp"
          "ι s⇩2 = NoOp"
          "ι s⇩3 = NoOp"
          "ι s⇩4 = NoOp"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  (* 2 additional cases leading to violations of secure spec *)
  | (NS_High_Trans)
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = ι s⇩3"
          "ι s⇩2 = ι s⇩4"
          "ι s⇩1 ≠ ι s⇩2"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
  | (NS_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False) e c where
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
  | (NS_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True) e c where
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬ mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
  | (NS_TMispred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "¬speculating s⇩3 ∧ ¬speculating s⇩4"
  | (NS_NTMispred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "¬speculating s⇩3 ∧ ¬speculating s⇩4"
  (* Suspicious states *)
  | (NS_Mispred_diff_ι) e c ea ca where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch ea ca"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch ea ca"
          "ι s⇩1 = ι s⇩3"
          "ι s⇩2 = ι s⇩4"
          "ι s⇩1 ≠ ι s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "¬speculating s⇩3 ∧ ¬speculating s⇩4"

  | (NS_Mispred_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩1"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (NS_Mispred_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "¬(\<lbrakk>e\<rbrakk>(Δ\<^sub>n s⇩1))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩2"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "ι s⇩1 = Branch e c"
          "ι s⇩2 = Branch e c"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "Δ⇩0 s⇩1 ≠ Δ⇩0 s⇩2"
          "¬mispred (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬mispred (n s⇩2) (β s⇩2) (pc s⇩2)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"
  | (S_SpecFence) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = SpecFence"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve_NS) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩3 = 1"
          "n s⇩4 = 1"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"    
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_RegisterUpdate) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = RegisterUpdate r e"
          "ι s⇩4 = RegisterUpdate r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Load) r e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Load r e"
          "ι s⇩4 = Load r e"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Store) e⇩1 e⇩2 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Store e⇩1 e⇩2"
          "ι s⇩4 = Store e⇩1 e⇩2"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NoOp)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = NoOp"
          "ι s⇩4 = NoOp"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_TPred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_TMispred) e c where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩3"
          "\<lbrakk>e\<rbrakk>Δ\<^sub>n s⇩4"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NTPred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "¬ mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_NTMispred) e c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))"
          "¬(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4))"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Goto) c where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩3 = Goto c"
          "ι s⇩4 = Goto c"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "n s⇩3 > 1"
          "n s⇩4 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  (* suspicious states *)
  | (S_SpecFence_Resolve) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩4 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩3 = 1"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)
  | (S_Resolve_SpecFence)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "ι s⇩3 = SpecFence"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "n s⇩4 = 1"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)
  | (S_Other)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "ι s⇩3 ≠ ι s⇩4"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)
  | (S_BranchEqPredNeqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(n s⇩3 = n s⇩4 ∧ Δ s⇩3 = Δ s⇩4)"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_BranchNeqPredEqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4))"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_BranchNeqPredNeqEval) c e where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3"
          "speculating s⇩4"
          "ι s⇩3 = Branch e c"
          "ι s⇩4 = Branch e c"
          "¬((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)) \<longleftrightarrow> (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
          "¬(n s⇩3 = n s⇩4 ∧ Δ s⇩3 = Δ s⇩4)"
          "¬(mispred (n s⇩3) (β s⇩3) (pc s⇩3) \<longleftrightarrow> mispred (n s⇩4) (β s⇩4) (pc s⇩4))"
          "¬(n s⇩3 = n s⇩4 \<and> β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
  | (S_Resolve3)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "n s⇩3 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)
  | (S_Resolve4)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3 ∧ speculating s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "n s⇩4 > 1"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)


  | (NNSN) e⇩3 c⇩3 e⇩4 c⇩4 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          " speculating s⇩3' ∧ ¬speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
          "ι s⇩2 = \<iota> s⇩4"
          "\<iota> s⇩4 = Branch e⇩4 c⇩4 \<longrightarrow> ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)
                                               ∧ ¬(β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "\<iota> s⇩4 \<noteq> Branch e⇩4 c⇩4 \<longrightarrow> pc s⇩1 \<noteq> pc s⇩2 
                                                      ∧ pc s⇩3 \<noteq> pc s⇩4"
          "ι s⇩1 = Branch e⇩3 c⇩3"
          "ι s⇩3 = Branch e⇩3 c⇩3"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"

  | (NNNS) e⇩3 c⇩3 e⇩4 c⇩4 where 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬speculating s⇩3' ∧  speculating s⇩4'"
          "¬speculating s⇩3  ∧ ¬speculating s⇩4"
          "ι s⇩1 = \<iota> s⇩3"
          "\<iota> s⇩3 = Branch e⇩3 c⇩3 \<longrightarrow> ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3)
                                               ∧ ¬(β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "\<iota> s⇩3\<noteq> Branch e⇩3 c⇩3 \<longrightarrow> pc s⇩1 \<noteq> pc s⇩2 
                                                     ∧ pc s⇩3 \<noteq> pc s⇩4"
          "ι s⇩2 = Branch e⇩4 c⇩4"
          "ι s⇩4 = Branch e⇩4 c⇩4"
          "mispred (n s⇩4) (β s⇩4) (pc s⇩4)"

  | (NSSS) e⇩3 c⇩3 where
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "¬ speculating s⇩3  ∧ speculating s⇩4"
          "\<iota> s⇩3 = Branch e⇩3 c⇩3"
          "mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ n s⇩4 > 1"          
          "op⇩ℒ s⇩1 = \<bottom>"


  | (NSSN)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ ¬ speculating s⇩4'"
          "¬ speculating s⇩3  ∧ speculating s⇩4"
          "op⇩ℒ s⇩1 = \<bottom>"


  | (NSNS)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ n s⇩4 > 1"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬ speculating s⇩3' ∧ speculating s⇩4'"
          "¬ speculating s⇩3  ∧ speculating s⇩4"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "ι s⇩1 = ι s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (SNNN) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ n s⇩3 = 1"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬ speculating s⇩3' ∧ ¬ speculating s⇩4'"
          "speculating s⇩3  ∧ ¬ speculating s⇩4"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩2 = ι s⇩4"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (SNSS)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3  ∧ ¬ speculating s⇩4"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (SNSN)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ n s⇩3 > 1"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬ speculating s⇩4' ∧ speculating s⇩3'"
          "¬ speculating s⇩4  ∧ speculating s⇩3"
          "¬ halting⇩p⇩r⇩e⇩d s⇩2"
          "¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "ι s⇩2 = ι s⇩4"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (SNNS)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬ speculating s⇩3' ∧ speculating s⇩4'"
          "speculating s⇩3  ∧ ¬ speculating s⇩4"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (NSNN) 
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3)"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ n s⇩4 = 1"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "¬ speculating s⇩3' ∧ ¬ speculating s⇩4'"
          "¬ speculating s⇩3  ∧ speculating s⇩4"
          "¬ halting⇩p⇩r⇩e⇩d s⇩1"
          "¬ halting⇩p⇩r⇩e⇩d s⇩3"
          "ι s⇩1 = ι s⇩3"
          "op⇩ℒ s⇩1 = \<bottom>"

  | (SSSN)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 \<noteq> SpecFence"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ 1 < n s⇩3"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 = SpecFence"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 = n s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3 ∧ speculating s⇩4"
          "speculating s⇩3' ∧ ¬ speculating s⇩4'"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)

  | (SSNS)
          "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1)"
          "¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
          "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 = SpecFence"
          "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ 1 = n s⇩3"
          "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 \<noteq> SpecFence"
          "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 < n s⇩4"
          "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)"
          "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4')"
          "valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3')"
          "wfs2 s⇩2 s⇩4"
          "wfs2 s⇩1 s⇩3"
          "speculating s⇩3 ∧ speculating s⇩4"
          "¬speculating s⇩3' ∧ speculating s⇩4'"
          "((op⇩ℒ s⇩1)::low_operation) = op⇩ℒ s⇩1'"
          "((op⇩ℒ s⇩2)::low_operation) = op⇩ℒ s⇩2'" 
          "((op⇩ℒ s⇩3)::low_operation) = op⇩ℒ s⇩3'"
          "((op⇩ℒ s⇩4)::low_operation) = op⇩ℒ s⇩4'" 
          "((op⇩ℋ s⇩1)::high_operation) = op⇩ℋ s⇩1'"
          "((op⇩ℋ s⇩2)::high_operation) = op⇩ℋ s⇩2'"
          "((op⇩ℋ s⇩3)::high_operation) = op⇩ℋ s⇩3'"
          "((op⇩ℋ s⇩4)::high_operation) = op⇩ℋ s⇩4'"
          (*Either all in low state or all in high state *)

proof -
  have invs: "valid_step2 (s⇩2, s⇩4) (s⇩2', s⇩4') ∧ valid_step2 (s⇩1, s⇩3) (s⇩1', s⇩3') ∧
        wfs2 s⇩2 s⇩4 ∧ wfs2 s⇩1 s⇩3 ∧ ¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)"
    using assms(1-2) valid_step4_implies_valid_step2 wfs4_no_resolve_s⇩1_s⇩2 wfs4_def by blast
  have ss: "speculating s⇩3' ∧ speculating s⇩4' ⟹ speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    using assms(1) assms(2)
    apply (induct_tac rule: valid_step4_spec_spec_cases)
    apply blast+
    using S_RegisterUpdate apply blast
    using S_Load apply blast
    using S_Store apply blast
    using S_TPred apply blast
    using S_TMispred apply blast
    using S_NTPred apply blast
    using S_NTMispred apply blast
    using S_Goto apply blast
    using S_NoOp apply blast
    using S_Resolve apply blast
    using S_Other  apply linarith
    using S_BranchEqPredNeqEval apply blast
    using S_BranchNeqPredEqEval apply blast
    using S_BranchNeqPredNeqEval apply blast
    using S_Resolve3 apply blast
    using S_Resolve4 by blast  
  have sn: "¬speculating s⇩3' ∧ ¬speculating s⇩4' ⟹ speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    using assms(1) assms(2)
    apply (induct_tac rule: valid_step4_spec_nospec_cases)
    apply blast+
    using S_SpecFence apply blast
    using S_Resolve_NS apply blast
    using S_SpecFence_Resolve apply blast
    using S_Resolve_SpecFence by blast
  have nn: "¬speculating s⇩3' ∧ ¬speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ ¬speculating s⇩4 ⟹ ?thesis"
    using assms(1) assms(2)
    apply (induct_tac rule: valid_step4_nospec_cases)
    apply blast+
    using NS_SpecFence apply blast
    using NS_TPred apply blast
    using NS_NTPred apply blast
    using NS_Goto apply blast
    using NS_RegisterUpdate apply blast
    using NS_Load apply blast
    using NS_Store apply blast
    using NS_NoOp apply blast
    using NS_High_Trans apply blast
    using NS_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False apply blast
    using NS_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True by blast
  have ns: "speculating s⇩3' ∧ speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ ¬speculating s⇩4 ⟹ ?thesis"
    using assms(1) assms(2)
    apply (induct_tac rule: valid_step4_nospec_spec_cases)
    apply blast+
    using NS_TMispred apply blast
    using NS_NTMispred apply blast
    using NS_Mispred_diff_ι apply presburger
    using NS_Mispred_Branch_s⇩1s⇩3_True_s⇩2s⇩4_False apply blast
    using NS_Mispred_Branch_s⇩1s⇩3_False_s⇩2s⇩4_True by blast

  have nnns: "¬speculating s⇩3' ∧ speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ ¬speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "mispred (n s⇩4) (β s⇩4) (pc s⇩4)") defer
     apply (meson assms(1) step4_s⇩4_ns_s_mispred)
    apply (subgoal_tac "\<And>e c. \<iota> s⇩3 = Branch e c \<longrightarrow> ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬(β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)") defer    
    using assms(1) step4_s⇩3_ns_ns_no_mispred apply (metis gr0I speculating_def)
    apply (subgoal_tac "\<exists>ea ca. \<iota> s⇩4 = Branch ea ca") defer
    using assms(1) step4_s⇩4_ns_s_branch apply blast
    apply (subgoal_tac "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)") defer
    using assms(1,2) step4_ns_no_resolve 
     apply blast
    by (smt (verit, ccfv_SIG) NNNS assms(1) assms(2) invs pc⇩0_pc_no_spec step2_ns_ι_same step4_ns_ι_s12_same step2_ns_ρ_equal wfs2_s⇩n⇩s_ns)


  have nnsn: "speculating s⇩3' ∧ ¬speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ ¬speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "mispred (n s⇩3) (β s⇩3) (pc s⇩3)") defer
    apply (meson assms(1) step4_s⇩3_ns_s_mispred)
    apply (subgoal_tac "\<exists>ea ca. \<iota> s⇩3 = Branch ea ca") defer
    using assms(1) step4_s⇩3_ns_s_branch apply blast
    apply (subgoal_tac "\<And>e c. \<iota> s⇩4 = Branch e c \<longrightarrow> ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ ¬(β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)") defer
    using assms(1) step4_s⇩4_ns_ns_no_mispred apply (metis gr0I speculating_def)
    apply (subgoal_tac "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4)") defer
    using assms(1,2) step4_ns_no_resolve apply blast
    by (smt (verit, ccfv_SIG) NNSN assms(1) assms(2) invs pc⇩0_pc_no_spec step2_ns_ι_same step4_ns_ι_s12_same step2_ns_ρ_equal wfs2_s⇩n⇩s_ns)


  have nsnn: "¬ speculating s⇩3' ∧ ¬ speculating s⇩4' ∧ ¬ speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) NSNN 
    apply (subgoal_tac "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')")
    defer using assms(1) apply blast
    apply (drule valid_step4_implies_valid_step2)
    using invs apply (elim conjE)
    apply (erule valid_step2_cases)
    apply blast+
    using step4_not_resolve_s⇩4_ns_halting_specfence step4_resolve_s⇩4_ns_n_eq_1 by metis+

  have nsns: "¬speculating s⇩3' ∧ speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>") defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    apply (subgoal_tac "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 < n s⇩4") defer
    apply auto[1]
    using assms(1) apply (simp add: speculating_def valid_step_def trans_resolve_def)
    using assms(1-2) One_nat_def speculating_def step4_resolve_s⇩4_spec_n_gt_1 apply presburger
    by (metis NSNS always_resolve_on_halt invs no_valid_transition_halt resolve_must_be_speculating 
              step2_ns_ι_same valid_step2_step)

  have nsnn: "¬ speculating s⇩3' ∧ ¬ speculating s⇩4' ∧ ¬ speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) NSNN
    apply (subgoal_tac "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')")
    defer using assms(1) apply blast
    apply (drule valid_step4_implies_valid_step2)
    using invs nsnn apply (elim conjE)
    apply (erule valid_step2_cases)
    by blast+

  have nssn: "speculating s⇩3' ∧ ¬speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) NSSN invs by blast

  have nsss: "speculating s⇩3' ∧ speculating s⇩4' ⟹ ¬speculating s⇩3 ∧ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>") defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    apply (subgoal_tac "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 < n s⇩4") defer
    apply auto[1]
    using assms(1) apply (simp add: speculating_def valid_step_def trans_resolve_def)
    using One_nat_def assms(1) speculating_def step4_resolve_s⇩4_spec_n_gt_1 apply presburger
    using NSSS invs speculating_branch speculating_mispred valid_step2_s⇩s_always_step by blast

  have snns: "speculating s⇩3 ∧ ¬speculating s⇩4 ∧ ¬speculating s⇩3' ∧ speculating s⇩4' ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) SNNS invs by blast

  have snsn: "speculating s⇩3 ∧ ¬speculating s⇩4 ∧ speculating s⇩3' ∧ ¬speculating s⇩4' ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) SNSN
    by (metis always_resolve_on_halt invs no_valid_transition_halt step2_ns_ι_same step4_ns_no_resolve_s⇩4 step4_resolve_s⇩3_spec_n_gt_1 valid_step2_s⇩n⇩s_ns_step valid_step2_s⇩s_always_step)
    
  have snss: "speculating s⇩3 ∧ ¬speculating s⇩4 ∧ speculating s⇩3' ∧ speculating s⇩4' ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) SNSS invs by blast

  have snnn: "¬ speculating s⇩3' ∧ ¬ speculating s⇩4' ∧ speculating s⇩3 ∧ ¬ speculating s⇩4 ⟹ ?thesis"
    apply (subgoal_tac "op⇩ℒ s⇩1 = \<bottom>")
      defer
    apply (metis assms(1) assms(2) step4_op⇩ℒ_spec_diff_invalid step4_spec_retains_lo)
    using assms(1) assms(2) SNNN by (metis always_resolve_on_halt invs n_1_resolve_nospec 
                                           no_valid_transition_halt step2_ns_ι_same 
                                           step4_ns_no_resolve_s⇩4 valid_step2_s⇩n⇩s_ns_step 
                                           valid_step2_s⇩s_always_step)

  have sssn: "speculating s⇩3 ∧ speculating s⇩4 ∧ speculating s⇩3' ∧ ¬ speculating s⇩4' ⟹ ?thesis"
    apply (subgoal_tac "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ 1 < n s⇩3") defer
    using assms(1) step4_resolve_s⇩3_spec_n_gt_1 apply blast
    apply (subgoal_tac "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 = n s⇩4") defer
    using assms(1) step4_resolve_s⇩4_ns_n_eq_1 apply blast
    apply (subgoal_tac "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 = SpecFence") defer
    using assms(1) step4_not_resolve_s⇩4_ns_halting_specfence apply blast
    apply (subgoal_tac "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 \<noteq> SpecFence") defer
    using assms(1) step4_not_resolve_s⇩3_spec_halting_not_specfence apply blast
    
    apply (subgoal_tac "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)") defer
    apply (rule notI)
    apply (elim conjE)
    using assms(2) apply (simp add:  wfs4_def wfs2_def)
    apply auto[1]

    using assms(1) assms(2) SSSN invs step2_spec_retains_lo_hi by blast


  have ssns: "speculating s⇩3 ∧ speculating s⇩4 ∧ ¬speculating s⇩3' ∧ speculating s⇩4' ⟹ ?thesis"
    apply (subgoal_tac "resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ 1 < n s⇩4") defer
    using assms(1) step4_resolve_s⇩4_spec_n_gt_1 apply blast
    apply (subgoal_tac "resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ 1 = n s⇩3") defer
    using assms(1) step4_resolve_s⇩3_ns_n_eq_1 apply blast
    apply (subgoal_tac "¬ resolve (n s⇩3) (β s⇩3) (pc s⇩3) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩3 ∧ ι s⇩3 = SpecFence") defer
    using assms(1) step4_not_resolve_s⇩3_ns_halting_specfence apply blast
    apply (subgoal_tac "¬ resolve (n s⇩4) (β s⇩4) (pc s⇩4) ⟶ ¬ halting⇩p⇩r⇩e⇩d s⇩4 ∧ ι s⇩4 \<noteq> SpecFence") defer
    using assms(1) step4_not_resolve_s⇩4_spec_halting_not_specfence apply blast
    
    apply (subgoal_tac "¬(n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4)") defer
    apply (rule notI)
    apply (elim conjE)
    using assms(2) apply (simp add: wfs4_def wfs2_def)
    apply auto[1]

    using assms(1) assms(2) SSNS invs step2_spec_retains_lo_hi by blast

  show ?thesis
    by (metis nn nnns nnsn ns nsnn nsns nssn nsss sn snnn snns snsn snss ss ssns sssn)
qed


lemma const_pc⇩0_Π_equal_ι⇩0:
  assumes "Π s⇩1 = Π s⇩2 ∧ Π s⇩1 = Π s⇩3 ∧ Π s⇩1 = Π s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2 ∧ pc⇩0 s⇩1 = pc⇩0 s⇩3 ∧ pc⇩0 s⇩1 = pc⇩0 s⇩4" 
    shows "ι⇩0 s⇩1 = ι⇩0 s⇩3 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩2 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩4"
  using assms by auto

lemma const_pc⇩0_Π_equal_ns_ι:
  assumes "Π s⇩1 = Π s⇩2 ∧ Π s⇩1 = Π s⇩3 ∧ Π s⇩1 = Π s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2 ∧ pc⇩0 s⇩1 = pc⇩0 s⇩3 ∧ pc⇩0 s⇩1 = pc⇩0 s⇩4" 
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "ι s⇩1 = ι s⇩3 ∧ ι s⇩1 = ι s⇩2 ∧ ι s⇩1 = ι s⇩4"
proof -
  have "ι⇩0 s⇩1 = ι⇩0 s⇩3 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩2 ∧ ι⇩0 s⇩1 = ι⇩0 s⇩4"
    using assms(1-2) const_pc⇩0_Π_equal_ι⇩0 by blast
  then show ?thesis
    using assms(3-5) apply (simp add: wfs4_def wfs2_def)
    by (metis ι⇩0_ι_no_spec assms(3) wfs4_s⇩1_s⇩2_ns)
qed














section ‹Transitions›


subsubsection ‹Program memory (Π)›

lemma step4_Π_const4:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')⟧ ⟹ Π s⇩1 = Π s⇩1' ∧ Π s⇩2 = Π s⇩2' ∧ Π s⇩3 = Π s⇩3' ∧ Π s⇩4 = Π s⇩4'"
  using valid_step4_implies_valid_step2 step2_Π_const by blast

lemma step4_Π_equal4:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); Π s⇩1 = Π s⇩2; Π s⇩1 = Π s⇩3; Π s⇩1 = Π s⇩4⟧ 
    ⟹ Π s⇩1' = Π s⇩1 ∧ Π s⇩2' = Π s⇩1 ∧ Π s⇩3' = Π s⇩1 ∧ Π s⇩4' = Π s⇩1"
  by (metis step4_Π_const4)


subsubsection ‹Trusted program memory (𝒯ρ)›

lemma step4_𝒯⇩ρ_const4:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩1' ∧ 𝒯⇩ρ s⇩2 = 𝒯⇩ρ s⇩2' ∧ 𝒯⇩ρ s⇩3 = 𝒯⇩ρ s⇩3' ∧ 𝒯⇩ρ s⇩4 = 𝒯⇩ρ s⇩4'"
  using assms valid_step4_implies_valid_step2 step2_𝒯⇩ρ_const by blast

lemma step4_𝒯⇩ρ_equal4:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩2" and "𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩3" and "𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩4" 
    shows "𝒯⇩ρ s⇩1' = 𝒯⇩ρ s⇩1 ∧ 𝒯⇩ρ s⇩2' = 𝒯⇩ρ s⇩1 ∧ 𝒯⇩ρ s⇩3' = 𝒯⇩ρ s⇩1 ∧ 𝒯⇩ρ s⇩4' = 𝒯⇩ρ s⇩1"
  using assms by (metis step4_𝒯⇩ρ_const4)


subsubsection ‹Entrypoint (ℰ𝒫)›

lemma step4_ℰ𝒫_const:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "ℰ𝒫 s⇩1 = ℰ𝒫 s⇩1' ∧ ℰ𝒫 s⇩2 = ℰ𝒫 s⇩2' ∧ ℰ𝒫 s⇩3 = ℰ𝒫 s⇩3' ∧ ℰ𝒫 s⇩4 = ℰ𝒫 s⇩4'"
  using assms valid_step4_implies_valid_step2 step2_ℰ𝒫_const by blast

lemma step4_ℰ𝒫_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "ℰ𝒫 s⇩1' = ℰ𝒫 s⇩1 ∧ ℰ𝒫 s⇩2' = ℰ𝒫 s⇩1 ∧ ℰ𝒫 s⇩3' = ℰ𝒫 s⇩1 ∧ ℰ𝒫 s⇩4' = ℰ𝒫 s⇩1"
  using assms valid_step4_implies_valid_step2
  apply (simp only: wfs4_def wfs2_def)
  by (metis step2_ℰ𝒫_const)


subsubsection ‹Secret set of memory (𝒮𝒯)›

lemma step4_𝒮⇩𝒯_const:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "𝒮⇩𝒯 s⇩1 = 𝒮⇩𝒯 s⇩1' ∧ 𝒮⇩𝒯 s⇩2 = 𝒮⇩𝒯 s⇩2' ∧ 𝒮⇩𝒯 s⇩3 = 𝒮⇩𝒯 s⇩3' ∧ 𝒮⇩𝒯 s⇩4 = 𝒮⇩𝒯 s⇩4'"
  using assms valid_step4_implies_valid_step2
  using step2_𝒮⇩𝒯_const by blast

lemma step4_𝒮⇩𝒯_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "𝒮⇩𝒯 s⇩1' = 𝒮⇩𝒯 s⇩1 ∧ 𝒮⇩𝒯 s⇩2' = 𝒮⇩𝒯 s⇩1 ∧ 𝒮⇩𝒯 s⇩3' = 𝒮⇩𝒯 s⇩1 ∧ 𝒮⇩𝒯 s⇩4' = 𝒮⇩𝒯 s⇩1"
  using assms valid_step4_implies_valid_step2
  apply (simp only: wfs4_def wfs2_def)
  by (metis step2_𝒮⇩𝒯_const)


subsubsection ‹Adversary readable memory (𝒰rd)›

lemma step4_𝒰rd_const:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "𝒰rd s⇩1 = 𝒰rd s⇩1' ∧ 𝒰rd s⇩2 = 𝒰rd s⇩2' ∧ 𝒰rd s⇩3 = 𝒰rd s⇩3' ∧ 𝒰rd s⇩4 = 𝒰rd s⇩4'"
  using assms valid_step4_implies_valid_step2
  using step2_𝒰rd_const by blast

lemma step4_𝒰rd_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "𝒰rd s⇩1' = 𝒰rd s⇩1 ∧ 𝒰rd s⇩2' = 𝒰rd s⇩1 ∧ 𝒰rd s⇩3' = 𝒰rd s⇩1 ∧ 𝒰rd s⇩4' = 𝒰rd s⇩1"
  using assms valid_step4_implies_valid_step2
  apply (simp only: wfs4_def wfs2_def)
  by (metis step2_𝒰rd_const)


subsubsection ‹Adversary writable memory (𝒰wr)›

lemma step4_𝒰wr_const:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
    shows "𝒰wr s⇩1 = 𝒰wr s⇩1' ∧ 𝒰wr s⇩2 = 𝒰wr s⇩2' ∧ 𝒰wr s⇩3 = 𝒰wr s⇩3' ∧ 𝒰wr s⇩4 = 𝒰wr s⇩4'"
  using assms valid_step4_implies_valid_step2
  using step2_𝒰wr_const by blast

lemma step4_𝒰wr_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
    shows "𝒰wr s⇩1' = 𝒰wr s⇩1 ∧ 𝒰wr s⇩2' = 𝒰wr s⇩1 ∧ 𝒰wr s⇩3' = 𝒰wr s⇩1 ∧ 𝒰wr s⇩4' = 𝒰wr s⇩1"
  using assms valid_step4_implies_valid_step2
  apply (simp only: wfs4_def wfs2_def)
  by (metis step2_𝒰wr_const)



subsubsection ‹Adversary observable memory (μ𝒜)›

lemma step4_const_μ⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "(∀e⇩1 e⇩2. ι s⇩1 ≠ Store e⇩1 e⇩2)"
    shows "μ⇩0 s⇩1' = μ⇩0 s⇩1 ∧ μ⇩0 s⇩2' = μ⇩0 s⇩2 ∧ μ⇩0 s⇩3' = μ⇩0 s⇩3 ∧ μ⇩0 s⇩4' = μ⇩0 s⇩4"
proof -
  have const_μ1: "μ⇩0 s⇩1' = μ⇩0 s⇩1"
    using assms(1) assms(4) assms(7) step2_const_μ⇩0 valid_step4_implies_valid_step2 wfs4_def by blast
  have const_μ2: "μ⇩0 s⇩2' = μ⇩0 s⇩2"
    by (metis (no_types, lifting) assms(1) assms(4) assms(5) assms(7) pc⇩0_pc_no_spec step2_const_μ⇩0 valid_step4_implies_valid_step2 wfs4_def wfs4_s⇩1_s⇩2_ns)
  show ?thesis
    by (metis (no_types, lifting) const_μ1 const_μ2 assms(1) assms(4) step2_μ⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs2_def wfs4_def)
qed

lemma step4_store_μ⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Store e⇩1 e⇩2"
    shows "μ⇩0 s⇩1' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩1)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩1)) (μ⇩0 s⇩1)
         ∧ μ⇩0 s⇩2' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩2)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩2)) (μ⇩0 s⇩2)
         ∧ μ⇩0 s⇩3' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩3)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩3)) (μ⇩0 s⇩3)
         ∧ μ⇩0 s⇩4' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩4)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩4)) (μ⇩0 s⇩4)"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  using assms(3) assms(7) step2_store_μ⇩0 apply presburger
  apply (metis Instr.distinct(29) assms(7))
  apply (smt (z3) assms(3) assms(4) const_pc⇩0_Π_equal_ns_ι wfs2_def wfs4_def)
  using assms(6) by blast+


lemma step4_store_μ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Store e⇩1 e⇩2"
      and "μ⇩0 s⇩1 = μ⇩0 s⇩2"
    shows "μ⇩0 s⇩1' = μ⇩0 s⇩2' ∧ μ⇩0 s⇩1' = μ⇩0 s⇩3' ∧ μ⇩0 s⇩1' = μ⇩0 s⇩4'"
proof - 
  have same: "μ⇩0 s⇩1' = μ⇩0 s⇩3' ∧ μ⇩0 s⇩2' = μ⇩0 s⇩4'"
    using assms(1,4-6,8) by (meson step2_μ⇩0_equal valid_step4_implies_valid_step2 wfs4_def)
  have step2: "μ⇩0 s⇩1' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩1)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩1)) (μ⇩0 s⇩1)
             ∧ μ⇩0 s⇩2' = μ⇩n⇩w⇩r⇩i⇩t⇩e (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩2)) (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩2)) (μ⇩0 s⇩2)"
    using assms(1-8) step4_store_μ⇩0 by blast
  have eval_e1: "(\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩1)) = (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩2))" (* lemma *)
    by (meson assms(4) assms(6) eval_exp_same_Δ⇩0_equiv wfs2_def wfs4_def wfs4_s⇩1_s⇩2_ns)
  have eval_e2: "(\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩1)) = (\<lbrakk>e⇩2\<rbrakk>(Δ⇩n s⇩2))"
    by (meson assms(4) assms(6) eval_exp_same_Δ⇩0_equiv wfs2_def wfs4_def wfs4_s⇩1_s⇩2_ns)  
  show ?thesis
    using assms(8) eval_e1 eval_e2 same step2 by metis
qed


lemma step4_μ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "μ⇩0 s⇩1 = μ⇩0 s⇩2"
    shows "μ⇩0 s⇩1' = μ⇩0 s⇩2' ∧ μ⇩0 s⇩1' = μ⇩0 s⇩3' ∧ μ⇩0 s⇩1' = μ⇩0 s⇩4'"
  using assms(4) apply (simp add: wfs4_def wfs2_def)
  using assms by (metis step4_const_μ⇩0 step4_store_μ⇩0_equal)


subsubsection ‹Adversary observable registers (Δ𝒜)›

lemma step4_const_Δ⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "(∀e r. ι s⇩1 ≠ Load r e ∧ ι s⇩1 ≠ RegisterUpdate r e)"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩1 ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩2 ∧ Δ⇩0 s⇩3' = Δ⇩0 s⇩3 ∧ Δ⇩0 s⇩4' = Δ⇩0 s⇩4"
proof -
  have const_Δ⇩01: "Δ⇩0 s⇩1' = Δ⇩0 s⇩1"
    using step2_const_Δ⇩0 wfs4_def assms(1,3,4,7) by (meson valid_step4_implies_valid_step2)
  have const_Δ⇩02: "Δ⇩0 s⇩2' = Δ⇩0 s⇩2"
    using step2_const_Δ⇩0 wfs4_def assms(1,3,4,5,7) by (metis (no_types, lifting) step4_ns_ι_s12_same valid_step4_implies_valid_step2)
  show ?thesis
    by (metis (no_types, lifting) const_Δ⇩01 const_Δ⇩02 assms(1) assms(4) step2_Δ⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs2_def wfs4_def)
qed


lemma step4_load_Δ⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Load r e"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩1 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩1 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))) 
         ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩2 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩2 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)))
         ∧ Δ⇩0 s⇩3' = Δ⇩0 s⇩3 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩3 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3)))
         ∧ Δ⇩0 s⇩4' = Δ⇩0 s⇩4 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩4 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (smt assms(3) assms(4) assms(7) step2_load_Δ⇩0 wfs4_def)
  using step2_load_Δ⇩0 assms(3) assms(7) apply auto[1]
  apply (intro conjI)
  using assms(3) assms(7) step2_load_Δ⇩0 apply blast
  using assms(7) apply fastforce+
  using assms(1,3-4) step4_ns_ι_s12_same apply blast
  using assms(6) by blast+


lemma step4_load_Δ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "μ⇩0 s⇩1 = μ⇩0 s⇩2"
      and "ι s⇩1 = Load r e"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩2' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩3' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩4'"
proof -
  have same: "Δ⇩0 s⇩1' = Δ⇩0 s⇩3' ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩4'"
    using assms by (meson step2_Δ⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs4_def)
  have step4: "Δ⇩0 s⇩1' = Δ⇩0 s⇩1 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩1 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))) ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩2 (r ↦ μ⇩0⇩r⇩e⇩a⇩d s⇩2 (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)))"
    using assms step4_load_Δ⇩0 by blast
  show ?thesis
    using assms(6) assms(7) same step4 by (metis assms(1) assms(4) step4_s12_eval_e_same)
qed


lemma step4_reg_update_Δ⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = RegisterUpdate r e"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩1 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))) ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩2 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))) ∧
           Δ⇩0 s⇩3' = Δ⇩0 s⇩3 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩3))) ∧ Δ⇩0 s⇩4' = Δ⇩0 s⇩4 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩4)))"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  using step2_reg_update_Δ⇩0 assms(3, 4, 7) wfs4_def apply auto[1]
  apply (simp add: assms(7))
  apply (simp add: assms(7))
  using assms(7) apply force
  using assms(1) assms(3) assms(4) step4_ns_ι_s12_same apply blast
  using assms(6) by auto


lemma step4_reg_update_Δ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = RegisterUpdate r e"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩2' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩3' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩4'"
proof - 
  have same: "Δ⇩0 s⇩1' = Δ⇩0 s⇩3' ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩4'"
    using assms by (meson step2_Δ⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs4_def)
  have step4: "Δ⇩0 s⇩1' = Δ⇩0 s⇩1 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))) ∧ Δ⇩0 s⇩2' = Δ⇩0 s⇩2 (r ↦ (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)))"
    using assms step4_reg_update_Δ⇩0 by blast
  show ?thesis
    using assms(6) same step4 by (metis assms(1) assms(4) step4_s12_eval_e_same)
qed


lemma step4_Δ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "μ⇩0 s⇩1 = μ⇩0 s⇩2"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩2' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩3' ∧ Δ⇩0 s⇩1' = Δ⇩0 s⇩4'"
  using assms(4) apply (simp add: wfs4_def wfs2_def)  
  using assms by (metis step4_const_Δ⇩0 step4_load_Δ⇩0_equal step4_reg_update_Δ⇩0_equal)


subsubsection ‹Adversary observable program counter (pc𝒜)›


lemma step4_const_pc⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "(∀r e⇩1 e⇩2 c. ι s⇩1 ≠ Store e⇩1 e⇩2 ∧ ι s⇩1 ≠ Load r e⇩1 ∧ ι s⇩1 ≠ RegisterUpdate r e⇩1 ∧ ι s⇩1 ≠ Branch e⇩1 c ∧ ι s⇩1 ≠ Goto c ∧ \<iota> s⇩1 ≠ SpecFence ∧ \<iota> s⇩1 ≠ NoOp)"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩1 ∧ pc⇩0 s⇩2' = pc⇩0 s⇩2 ∧ pc⇩0 s⇩3' = pc⇩0 s⇩3 ∧ pc⇩0 s⇩4' = pc⇩0 s⇩4"
proof -
  have const_pc⇩01: "pc⇩0 s⇩1' = pc⇩0 s⇩1"
    using assms(1,4,7) ι_some_Instr by blast
  have const_pc⇩02: "pc⇩0 s⇩2' = pc⇩0 s⇩2"
    using assms(1,4,5,7) ι_some_Instr by blast
  show ?thesis
    by (metis (no_types, lifting) const_pc⇩01 const_pc⇩02 assms(1) assms(4) assms(5) step2_pc⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs2_def wfs4_def)
qed

lemma step4_const_pc⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "(∀r e⇩1 e⇩2 c. ι s⇩1 ≠ Store e⇩1 e⇩2 ∧ ι s⇩1 ≠ Load r e⇩1 ∧ ι s⇩1 ≠ RegisterUpdate r e⇩1 ∧ ι s⇩1 ≠ Branch e⇩1 c ∧ ι s⇩1 ≠ Goto c ∧ \<iota> s⇩1 ≠ SpecFence ∧ \<iota> s⇩1 ≠ NoOp)"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩2' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩4'"
proof -
  have const_pc⇩0: "pc⇩0 s⇩1' = pc⇩0 s⇩1 ∧ pc⇩0 s⇩2' = pc⇩0 s⇩2 ∧ pc⇩0 s⇩3' = pc⇩0 s⇩3 ∧ pc⇩0 s⇩4' = pc⇩0 s⇩4"
    using assms step4_const_pc⇩0 by presburger
  show ?thesis
    by (metis (no_types, lifting) assms(1) assms(4) assms(5) local.const_pc⇩0 step2_pc⇩0_equal valid_step4_implies_valid_step2 wfs4_def)
qed
  

lemma step4_next_pc⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Store e⇩1 e⇩2 ∨ ι s⇩1 = Load r e⇩1 ∨ ι s⇩1 = RegisterUpdate r e⇩1 ∨ ι s⇩1 = Branch e⇩1 c \<or> \<iota> s⇩1 = NoOp"
      and "ι s⇩1 = Branch e⇩1 c ⟶ (nt_pred s⇩1 e⇩1 ∨ nt_mispred s⇩1 e⇩1)"
    shows "pc⇩0 s⇩1' = next_ρ (ρ s⇩1) ∧ pc⇩0 s⇩2' = next_ρ (ρ s⇩2) ∧ pc⇩0 s⇩3' = next_ρ (ρ s⇩3) ∧ pc⇩0 s⇩4' = next_ρ (ρ s⇩4)"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  using assms(7) apply auto[1]
  defer defer
  using assms(7) apply auto[1]
  apply simp
  using assms(3) wfs2_s⇩n⇩s_ns next_pc⇩0 tpod_step⇩2_def valid_step_def apply force
  apply simp
  using assms(3) wfs2_s⇩n⇩s_ns next_pc⇩0 tpod_step⇩2_def valid_step_def apply force
  apply simp
  using assms(3) wfs2_s⇩n⇩s_ns next_pc⇩0 tpod_step⇩2_def valid_step_def apply force
  apply (metis Instr.distinct(35) assms(3) step2_next_pc⇩0)  
  using assms(1) assms(3) assms(4) step4_ns_ι_s12_same apply blast
  using assms(6) apply blast+
  apply (metis Instr.distinct(15) Instr.distinct(23) Instr.distinct(35) Instr.distinct(5) Instr.inject(4) assms(7) assms(8))
  using assms(3) step2_next_pc⇩0 by blast


lemma step4_next_pc⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Store e⇩1 e⇩2 ∨ ι s⇩1 = Load r e⇩1 ∨ ι s⇩1 = RegisterUpdate r e⇩1 ∨ ι s⇩1 = Branch e⇩1 c \<or> \<iota> s⇩1 = NoOp"
      and "ι s⇩1 = Branch e⇩1 c ⟶ (nt_pred s⇩1 e⇩1 ∨ nt_mispred s⇩1 e⇩1)"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩2' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩4'"
proof - 
  have same: "pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩2' = pc⇩0 s⇩4'"
    using assms by (meson step2_pc⇩0_equal wfs4_def valid_step4_implies_valid_step2 wfs4_def)
  have step4: "pc⇩0 s⇩1' = next_ρ (ρ s⇩1) ∧ pc⇩0 s⇩2' = next_ρ (ρ s⇩2)"
    using assms step4_next_pc⇩0 by blast
  show ?thesis
    using assms by (metis next_ρ_equal same step4 wfs4_s⇩1_s⇩2_ns)
qed


lemma step4_jmp_pc⇩0:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Branch e c ∨ ι s⇩1 = Goto c"
      and "ι s⇩1 = Branch e c ⟶ (t_pred s⇩1 e ∨ t_mispred s⇩1 e)"
    shows "pc⇩0 s⇩1' = c ∧ pc⇩0 s⇩2' = c ∧ pc⇩0 s⇩3' = c ∧ pc⇩0 s⇩4' = c"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  using assms(7) apply auto[1]
  using assms(3, 7) apply (metis Instr.inject(4) step2_jmp_pc⇩0)
  using assms(7, 8) apply (metis Instr.distinct(31) Instr.inject(4))
  using assms(3, 7, 8) step2_jmp_pc⇩0 apply auto[1]
  using assms(3, 7, 8) step2_jmp_pc⇩0 apply auto[1]
  using assms(7) apply auto[1]
  apply (metis Instr.distinct(23) Instr.distinct(25) assms(7))
  apply (metis Instr.distinct(35) Instr.distinct(39) assms(7))
  using assms(1, 3, 4) step4_ns_ι_s12_same apply blast
  using assms(6) by blast+


lemma step4_jmp_pc⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Branch e c ∨ ι s⇩1 = Goto c"
      and "ι s⇩1 = Branch e c ⟶ (t_pred s⇩1 e ∨ t_mispred s⇩1 e)"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩2' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩4'"
proof - 
  have same: "pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩2' = pc⇩0 s⇩4'"
    using assms(1,2,3,4,5,6,7,8) step4_jmp_pc⇩0 by presburger
  have jmp: "pc⇩0 s⇩1' = c ∧ pc⇩0 s⇩2' = c"
    using assms step4_jmp_pc⇩0 by blast
  show ?thesis
    using jmp same by auto
qed


lemma step4_pc⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
    shows "pc⇩0 s⇩1' = pc⇩0 s⇩2' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩3' ∧ pc⇩0 s⇩1' = pc⇩0 s⇩4'"
proof -
  have ι_equal: "ι s⇩1 = ι s⇩2 ∧ ι s⇩1 = ι s⇩3 ∧ ι s⇩1 = ι s⇩4"
    using assms(4) apply (simp add: wfs4_def wfs2_def)
    by (metis assms(3) assms(4) assms(5) pc⇩0_pc_no_spec wfs4_s⇩1_s⇩2_ns)
  show ?thesis
    using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
    using assms step4_const_pc⇩0_equal apply (smt (z3) Instr.distinct(33) architecural_pc⇩0 step2_next_pc⇩0 step2_ns_ρ_equal wfs2_s⇩n⇩s_ns)
    using assms step4_jmp_pc⇩0_equal apply blast
    using assms step4_next_pc⇩0_equal apply blast
    using assms step4_jmp_pc⇩0_equal apply (metis Instr.distinct(31))
    using assms step4_next_pc⇩0_equal apply (metis Instr.distinct(5))
    using assms step4_next_pc⇩0_equal apply (metis Instr.distinct(15))
    using assms step4_next_pc⇩0_equal apply (metis Instr.distinct(23))
    apply (metis Instr.distinct(35) assms(3) pc⇩0_pc_no_spec step2_next_pc⇩0 step2_pc⇩0_equal wfs2_s⇩n⇩s_ns)
    using ι_equal apply blast
    using assms(6) by auto
qed


lemma step4_Π:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); Π s⇩1 = Π s⇩2; Π s⇩1 = Π s⇩3; Π s⇩1 = Π s⇩4⟧ 
    ⟹ Π s⇩1' = Π s⇩2' ∧ Π s⇩1' = Π s⇩3' ∧ Π s⇩1' = Π s⇩4'"
  using valid_step4_implies_valid_step2
  by (metis step2_Π_const)


lemma step4_𝒯⇩ρ:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); 𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩2; 𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩3; 𝒯⇩ρ s⇩1 = 𝒯⇩ρ s⇩4⟧ 
    ⟹ 𝒯⇩ρ s⇩1' = 𝒯⇩ρ s⇩2' ∧ 𝒯⇩ρ s⇩1' = 𝒯⇩ρ s⇩3' ∧ 𝒯⇩ρ s⇩1' = 𝒯⇩ρ s⇩4'"
  using valid_step4_implies_valid_step2
  by (metis step2_𝒯⇩ρ_const)

lemma step4_wf:
  "⟦(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4'); wfs s⇩1; wfs s⇩2; wfs s⇩3; wfs s⇩4⟧ 
      ⟹ wfs s⇩1' ∧ wfs s⇩2' ∧ wfs s⇩3' ∧ wfs s⇩4'"
  using valid_step4_implies_valid_step2
  using step2_wf by blast



subsubsection ‹Trace of Instructions and memory accessed (ω)›

lemma step4_det_ω_equal_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "inst⇩𝒯 s⇩1' = \<bottom>"
      and "ω s⇩1 = ω s⇩2"
    shows "ω s⇩1' = ω s⇩2'"
  using assms(1)   using valid_step4_implies_valid_step2 
apply (simp only: valid_step4.simps)
  apply (elim conjE impE)
  using assms(3) apply blast+
  using assms(1) assms(4) assms(5) step4_s⇩1_s⇩2_never_speculates low_equiv_s_simp
  by (meson tpod_step⇩4_def)

text ‹The length of the ω pairs are always equal when not speculating ›

lemma step4_det_ω_len_equal_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "length (ω s⇩1)  = length (ω s⇩2) ∧ length (ω s⇩3) = length (ω s⇩4)"
    shows "length (ω s⇩1') = length (ω s⇩2') ∧ length (ω s⇩3') = length (ω s⇩4')"  
  apply (subgoal_tac "¬resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧
                      ¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬resolve (n s⇩4) (β s⇩4) (pc s⇩4)")
  defer
  using assms(1-3) step4_ns_no_resolve apply blast
  using assms valid_step4_implies_valid_step2 Step2Semantics.valid_step2_step
  by (metis (mono_tags, lifting) eval_instr_length_ω_preserved no_valid_transition_halt step_def valid_step_def wfs4_s⇩1_s⇩2_ns)

text ‹If the ω pairs are ever not equal (in the high state) they remain not equal and can be
            detected on return to the low state.›

lemma step4_nospec_\<omega>\<^sub>1_noteq_\<omega>\<^sub>2:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "\<omega> s⇩1 \<noteq> \<omega> s⇩2"
    shows "\<omega> s⇩1' \<noteq> \<omega> s⇩2'"
  apply (subgoal_tac "¬resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧
                      ¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬resolve (n s⇩4) (β s⇩4) (pc s⇩4)")
  defer
  using assms(1-4) step4_ns_no_resolve apply blast
  using assms valid_step4_implies_valid_step2 Step2Semantics.valid_step2_step
  using eval_instr_ω_ne_preserved step_def valid_step_def
  by (metis (no_types, hide_lams) always_resolve_on_halt no_valid_transition_halt)

text ‹Instructions which do not access memory only add their program address to the trace.›

lemma step4_ω_ρ_None:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "ι s⇩1 = RegisterUpdate r e ∨ ι s⇩1 = Branch e c ∨ ι s⇩1 = Goto c \<or> ι s⇩1 = SpecFence \<or> ι s⇩1 = NoOp"
    shows "ω s⇩1' = ω s⇩1 @ [(ρ s⇩1, None)] ∧ ω s⇩2' = ω s⇩2 @ [(ρ s⇩2, None)]
         ∧ ω s⇩3' = ω s⇩3 @ [(ρ s⇩3, None)] ∧ ω s⇩4' = ω s⇩4 @ [(ρ s⇩4, None)]"
  using assms(1-4) apply (induction rule: valid_step4_nospec_cases)
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  using assms(6) apply auto[1]
  using assms(6) apply auto[1]
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def apply blast
  apply (metis assms(1) assms(3) assms(4) assms(5) step4_ns_ι_s12_same)
  using assms(2) assms(3) assms(4) step2_ω_ρ_None wfs4_def by blast+

text ‹Instructions that access memory add this to the trace along with their program address.›

lemma step4_ω_ρ_a:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Load r e ∨ ι s⇩1 = Store e e⇩2"
      and "a = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))"
    shows "ω s⇩1' = ω s⇩1 @ [(ρ s⇩1, Some a)] ∧ ω s⇩3' = ω s⇩3 @ [(ρ s⇩3, Some a)]
         ∧ ω s⇩2' = ω s⇩2 @ [(ρ s⇩2, Some a)] ∧ ω s⇩4' = ω s⇩4 @ [(ρ s⇩4, Some a)]"
proof - 
  have eval_e: "(\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1)) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))" (* lemma *)
    by (meson assms(1) assms(4) assms(6) step4_s12_eval_e_same)
  show ?thesis
    apply (subgoal_tac "Δ⇩n s⇩1 = Δ⇩n s⇩2")
    defer 
    using Δ⇩0_Δ_no_spec assms(4) assms(6) wfs4_s⇩1_s⇩2_ns apply auto[1]
    using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
    using assms(7) apply auto[1]
    using assms(7) apply auto[1]
    using assms(7) apply auto[1]
    using assms(7) apply auto[1]
    using assms(7) apply auto[1]
    using step2_ω_ρ_a assms(2-4, 7, 8) eval_e wfs4_def apply auto[1]
    using step2_ω_ρ_a assms(2-4, 7, 8) eval_e wfs4_def apply auto[1]
    using assms(4) apply (simp only: wfs4_def wfs2_def)
    apply (metis Instr.distinct(21) Instr.distinct(29) assms(7))
    apply (metis assms(1) assms(3) assms(4) step4_ns_ι_s12_same)
    using assms(7) by auto
qed

lemma step4_ω_ρ_a_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Load r e ∨ ι s⇩1 = Store e e⇩2"
      and "(a::Const) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))"
      and "ω s⇩1 = ω s⇩2 ∧ ω s⇩3 = ω s⇩4"
    shows "ω s⇩1' = ω s⇩2' ∧ ω s⇩3' = ω s⇩4'"
proof -
  have "ω s⇩1' = ω s⇩1 @ [(ρ s⇩1, Some a)] ∧ ω s⇩3' = ω s⇩3 @ [(ρ s⇩3, Some a)]
      ∧ ω s⇩2' = ω s⇩2 @ [(ρ s⇩2, Some a)] ∧ ω s⇩4' = ω s⇩4 @ [(ρ s⇩4, Some a)]"
    using assms(1-8) by (rule step4_ω_ρ_a)
  moreover have "ρ s⇩1 = ρ s⇩2"
    using assms(4) assms(5) pc⇩0_pc_no_spec wfs4_s⇩1_s⇩2_ns by auto
  moreover have "ρ s⇩3 = ρ s⇩4"
    using assms(4) apply (simp add: wfs4_def wfs2_def)
    by (metis assms(3) assms(5) pc⇩0_pc_no_spec)
  ultimately show ?thesis
    using assms(8, 9) by simp
qed

text ‹Trace pairs are equal when not speculating.›

lemma step4_ω12_ns_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "\<omega> s⇩1' = \<omega> s⇩2'"
  using assms by (metis (no_types, lifting) violation\<^sub>\<L>_def step4_det_ω_len_equal_ns valid_step4.simps wfs4_triv_def)  



section ‹Program Counter (pc)›

lemma step4_ns_pc_ρ_eq:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ρ s⇩1 = ρ s⇩2"
    shows "pc s⇩1 = pc s⇩2 ∧ pc s⇩1 = pc s⇩3 ∧ pc s⇩1 = pc s⇩4"
proof -
  have not_spec: "¬speculating s⇩1 ∧ ¬speculating s⇩2"
    using assms(2) wfs4_s⇩1_s⇩2_ns by auto

  have halt: "¬halting⇩p⇩r⇩e⇩d s⇩1 ∧ ¬halting⇩p⇩r⇩e⇩d s⇩2"
    using assms by (meson no_valid_transition_halt not_spec valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)

  show ?thesis
    using assms(2-4) ns_pc_ρ_eq wfs4_def wfs2_def by (metis (no_types, hide_lams) halt not_spec ns_pc⇩0_pc_equal)
qed

lemma step4_\<omega>_eq_pc_eq:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "pc s⇩1 = pc s⇩2 ∧ pc s⇩1 = pc s⇩3 ∧ pc s⇩3 = pc s⇩4"
  apply (subgoal_tac "ω s⇩1' = ω s⇩2'")
  apply (subgoal_tac "\<exists>a b. ω s⇩1' = ω s⇩1 @ [(\<rho> s⇩1, a)] ∧ ω s⇩2' = ω s⇩2 @ [(\<rho> s⇩2, b)]")
  apply auto[1]
  using assms(1,2,4) step4_ns_pc_ρ_eq apply metis
  apply (metis assms(1) assms(2) assms(4) step4_ns_pc_ρ_eq)
  using assms(1,2,4) apply (metis step4_ns_pc_ρ_eq)
  using assms(1,2,4) apply (meson ω_always_appends_ρ valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2 step4_ns_no_resolve)
  using assms(1-4) step4_ω12_ns_equal by blast

lemma step4_\<omega>_eq_addr_eval_eq:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = Load r e ∨ ι s⇩1 = Store e e⇩2"
    shows "((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))::Const) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))"
  apply (subgoal_tac "ι s⇩1 = ι s⇩2") defer
  using assms apply (metis step4_ω_eq_pc_eq step4_ns_ι_s12_same)
  apply (subgoal_tac "ω s⇩1' = ω s⇩2'") defer
  using assms step4_ω12_ns_equal apply blast
  apply (subgoal_tac "ω s⇩1' = ω s⇩1 @ [(\<rho> s⇩1, Some (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1)))] ∧ ω s⇩2' = ω s⇩2 @ [(\<rho> s⇩2, Some (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2)))]")
  apply auto[1]
  apply (subgoal_tac "¬ resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬ resolve (n s⇩2) (β s⇩2) (pc s⇩2)") defer
  using assms(1, 2, 4) step4_ns_no_resolve apply blast
  apply (subgoal_tac "s⇩1 ↝ s⇩1' \<and> s⇩2 ↝ s⇩2'")
  using ω_ρ_a  assms(5) apply auto[1]                            
  using assms(1) valid_step4_implies_valid_step2 Step2Semantics.valid_step2_step assms(4) by blast


lemma step4_ω34_ns_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "\<omega> s⇩3' = \<omega> s⇩4'"
  apply (subgoal_tac "pc s⇩1 = pc s⇩2")
  apply (subgoal_tac "¬speculating s⇩1 ∧ ¬speculating s⇩2")
  apply (subgoal_tac "ρ s⇩3 = ρ s⇩4")
  using assms(3) apply (simp add: wfs4_triv_def)
  using assms(1-2) apply (induction rule: valid_step4_cases)
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None apply force 
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None apply force
  apply (subgoal_tac "((\<lbrakk>e\<rbrakk>(Δ⇩n s⇩1))::Const) = (\<lbrakk>e\<rbrakk>(Δ⇩n s⇩2))")
  using step2_ω_ρ_a apply force
  using assms(1) assms(2) assms(3) step4_ω_eq_addr_eval_eq apply blast
  apply (subgoal_tac "((\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩1))::Const) = (\<lbrakk>e⇩1\<rbrakk>(Δ⇩n s⇩2))")
  using step2_ω_ρ_a apply force
  using assms(1) assms(2) assms(3) step4_ω_eq_addr_eval_eq apply blast
  using step2_ω_ρ_None apply force
  apply (metis assms(1) assms(2) step4_ns_ι_s12_same)
  using step2_ω_ρ_None assms(4) apply force
  using step2_ω_ρ_None assms(4) apply force
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None apply force
  using step2_ω_ρ_None assms(4) apply force
  using step2_ω_ρ_None assms(4) apply force
  using assms(4) apply blast+
  apply (metis (no_types, hide_lams) ω_ρ_None valid_step2_s⇩s_always_step)
  apply (metis (no_types, hide_lams) ω_ρ_None valid_step2_s⇩s_always_step)
  apply (metis (no_types, hide_lams) ω_ρ_None valid_step2_s⇩s_always_step)
  apply (metis (no_types, hide_lams) ω_ρ_None valid_step2_s⇩s_always_step)
  using assms(4) apply blast+
  apply (metis assms(1) assms(2) assms(4) pc⇩0_pc_no_spec step4_ns_pc_ρ_eq)
  using assms(1) assms(2) assms(3) step4_ω_eq_pc_eq apply blast
  using assms(4) apply blast+
  apply (metis assms(1) assms(2) assms(3) assms(4) pc⇩0_pc_no_spec step4_ω_eq_pc_eq)
  using assms(2) wfs4_s⇩1_s⇩2_ns apply blast
  by (meson assms(1) assms(2) assms(3) assms(4) step4_ω_eq_pc_eq)

lemma step4_ω_ns_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "\<omega> s⇩1' = \<omega> s⇩2' ∧ \<omega> s⇩3' = \<omega> s⇩4'"
  using assms step4_ω12_ns_equal step4_ω34_ns_equal by blast


subsubsection ‹Branch predictor (β)›

text ‹Branch predictor pairs are equal when not speculating.›

lemma step4_β_equal_low_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_ns s⇩1 s⇩2 s⇩3 s⇩4"
      and "inst⇩𝒯 s⇩1' = \<bottom>"
      and "β s⇩1  = β s⇩2"
    shows "β s⇩1' = β s⇩2' ∧ β s⇩1' = β s⇩3' ∧ β s⇩1' = β s⇩4'"
proof -
  have "β s⇩1' = β s⇩2'"
    using assms apply (induction rule: valid_step4_nospec_cases, simp_all)    
    using const_β tpod_step⇩2_def valid_step_def apply force
    using tpod_step⇩2_def tpod_step⇩4_def low_equiv_s.simps low_equiv_sub_def
    using assms(1) step4_s⇩1_s⇩2_never_speculates by blast+

  then show ?thesis
    using assms by (metis (mono_tags, lifting) step2_β⇩ω_equal_ns valid_step4_implies_valid_step2 
                    wfs4_def wfs4_ns_def)
qed

text ‹Non branching Instructions do not change the branch predictor.›

lemma step4_const_β_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = RegisterUpdate r e ∨ ι s⇩1 = Load r e ∨ ι s⇩1 = Store e⇩1 e⇩2 ∨ ι s⇩1 = SpecFence \<or> ι s⇩1 = NoOp"
    shows "β s⇩1' = β s⇩1 ∧ β s⇩2' = β s⇩2 ∧ β s⇩3' = β s⇩3 ∧ β s⇩4' = β s⇩4"
  using assms(1-5) apply (induction rule: valid_step4_nospec_cases)
  using assms(2,3) step2_const_β_ns apply blast
  using assms(7) apply (metis Instr.distinct(15) Instr.distinct(23) Instr.distinct(33) Instr.distinct(35) Instr.distinct(5))
  using assms(7) apply (metis Instr.distinct(15) Instr.distinct(23) Instr.distinct(33) Instr.distinct(35) Instr.distinct(5))
  using assms(7) apply (metis Instr.distinct(17) Instr.distinct(25) Instr.distinct(37) Instr.distinct(39) Instr.distinct(7))
  using assms(2,3) step2_const_β_ns apply blast
  using assms(2,3) step2_const_β_ns apply blast
  using assms(2,3) step2_const_β_ns apply blast
  using assms(2,3) step2_const_β_ns apply blast
  using assms(1,3,4) step4_ns_ι_s12_same apply blast
  using assms(6) by blast+


lemma step4_const_β_ns_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = RegisterUpdate r e ∨ ι s⇩1 = Load r e ∨ ι s⇩1 = Store e⇩1 e⇩2 ∨ ι s⇩1 = SpecFence \<or> ι s⇩1 = NoOp"
      and "β s⇩1 = β s⇩2 ∧ β s⇩3 = β s⇩4"
    shows "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'"
  using assms step4_const_β_ns by (smt (z3))


text ‹Branching Instructions add accesses to the branch predictor.›


lemma step4_update_β_False_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Branch e c ∨ ι s⇩1 = Goto c"
      and "ι s⇩1 = Branch e c ⟶ t_pred s⇩1 e ∨ nt_pred s⇩1 e"
    shows "β s⇩1' = update (n s⇩1) (ρ s⇩1) False (β s⇩1) ∧ β s⇩2' = update (n s⇩2) (ρ s⇩2) False (β s⇩2)
        ∧ β s⇩3' = update (n s⇩3) (ρ s⇩3) False (β s⇩3) ∧ β s⇩4' = update (n s⇩4) (ρ s⇩4) False (β s⇩4)"
  using assms(1-4) apply (induction rule: valid_step4_nospec_cases)  
  using assms(7) apply (metis Instr.distinct(33) Instr.distinct(37))
  using assms(2) assms(3) assms(4) wfs4_def step2_update_β_False_ns apply blast
  using assms(2) assms(3) assms(4) wfs4_def step2_update_β_False_ns apply blast
  using assms(2) assms(3) assms(4) wfs4_def step2_update_β_False_ns step2_s⇩n⇩s_no_mispred apply blast
  using assms(7) apply auto[1]
  using assms(7) apply auto[1]
  using assms(7) apply auto[1]
  apply (metis Instr.distinct(35) Instr.distinct(39) assms(7))
  using assms(1, 3-5) step4_ns_ι_s12_same apply blast
  using assms(6) by blast+


lemma step4_update_β_False_ns_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "pc⇩0 s⇩1 = pc⇩0 s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "ι s⇩1 = Branch e c ∨ ι s⇩1 = Goto c"
      and "ι s⇩1 = Branch e c ⟶ t_pred s⇩1 e ∨ nt_pred s⇩1 e"
      and "β s⇩1 = β s⇩2 ∧ β s⇩3 = β s⇩4"
    shows "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'"
proof - 
  have "β s⇩1' = update (n s⇩1) (ρ s⇩1) False (β s⇩1) ∧
        β s⇩2' = update (n s⇩2) (ρ s⇩2) False (β s⇩2) ∧
        β s⇩3' = update (n s⇩3) (ρ s⇩3) False (β s⇩3) ∧
        β s⇩4' = update (n s⇩4) (ρ s⇩4) False (β s⇩4)"
    using assms step4_update_β_False_ns by blast
  then show ?thesis
    using assms(9) apply auto 
    apply (metis assms(4) assms(5) not_gr_zero speculating_def wfs4_s⇩1_s⇩2_ns)
    by (metis assms(1) assms(3) assms(4) assms(5) not_gr_zero speculating_def step4_ns_pc_ρ_eq wfs4_s⇩1_s⇩2_ns)
qed


text ‹The length of the β pairs are always equal when not speculating ›

lemma step4_det_β_len_equal_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "length (β s⇩1)  = length (β s⇩2)"
    shows "length (β s⇩1') = length (β s⇩2')"
  apply (subgoal_tac "¬resolve (n s⇩1) (β s⇩1) (pc s⇩1) ∧ ¬resolve (n s⇩2) (β s⇩2) (pc s⇩2) ∧ 
                      ¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬resolve (n s⇩4) (β s⇩4) (pc s⇩4)")
  apply (subgoal_tac "pc s⇩1 = pc s⇩2")
  using assms apply (simp add: step_def tpod_step⇩4_def)
  using eval_instr_length_β_preserved_same_i 
  apply (metis assms(1) eval_instr_length_ω_greater less_not_refl step4_ω12_ns_equal step4_ns_ι_s12_same wfs4_triv_def)
  using assms(1) assms(2) assms(3) assms(4) step4_ω_eq_pc_eq wfs4_ns_def apply blast
  using assms(1, 2, 4) step4_ns_no_resolve by blast


lemma step4_β12_equal_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "β s⇩1' = β s⇩2'"
  by (metis (no_types, lifting) assms(1) assms(2) assms(3) assms(4) step4_det_β_len_equal_ns valid_step4_equiv_valid_step2 violation⇩ℒ_def wfs4_triv_def)

lemma step4_β_equal_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
    shows "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'"
proof -
  have pc_equal: "pc s⇩1 = pc s⇩2 ∧ pc s⇩1 = pc s⇩3 ∧ pc s⇩1 = pc s⇩4"
    by (metis assms(1) assms(2) assms(3) assms(4) step4_ω_eq_pc_eq)

  have "β s⇩3' = β s⇩4'"
    apply (subgoal_tac "β s⇩3 = β s⇩4")
    apply (subgoal_tac "n s⇩3 = n s⇩4")
    using assms(1,2) pc_equal apply (induction rule: valid_step4_cases) 
    using const_β valid_step2_step apply metis
    using step2_update_β_False_ns apply (smt (z3))
    apply (smt (z3) step2_update_β_False_ns)
    using step2_update_β_False_ns apply (smt (z3) Instr.distinct(31))
    using step2_const_β_ns apply metis
    using step2_const_β_ns apply metis
    using step2_const_β_ns apply metis
    using step2_const_β_ns apply metis
    apply (metis assms(1) assms(2) step4_ns_ι_s12_same)
    defer defer
    apply (metis update_β_True valid_step2_step)
    apply (metis update_β_True valid_step2_step)
    apply (metis assms(1) assms(2) step4_ns_ι_s12_same)
    apply (metis update_β_True valid_step2_step)
    defer
    using assms(4) apply blast+
    apply (metis update_β_True valid_step2_s⇩s_always_step)
    using assms(4) apply blast+
    apply (metis update_β_True valid_step2_s⇩s_always_step)
    using assms(4) apply blast+
    apply force+
    using assms(4) apply blast+
    using assms(4) apply (simp add: speculating_def)
    using assms(3) wfs4_triv_def apply blast
    apply (smt (z3) update_β_False valid_step2_step)
    apply (smt (z3) assms(4) step2_no_mispred_ns step2_update_β_False_ns)
    by (metis update_β_True valid_step2_s⇩s_always_step)

  then show ?thesis
    by (meson assms(1) assms(2) assms(3) assms(4) step4_β12_equal_ns)
qed

section ‹Invariants for step 4›

lemma step4_wfs4:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4" 
    shows "wfs4 s⇩1' s⇩2' s⇩3' s⇩4'"
  using assms apply (auto simp only: wfs4_def)
  using step2_wfs2 wfs4_def valid_step4_implies_valid_step2 apply blast
  apply (meson step2_wfs2 valid_step4_implies_valid_step2)
  by (simp_all only: step4_Π_const4 step4_𝒯⇩ρ_const4 step4_ℰ𝒫_const step4_𝒮⇩𝒯_const step4_𝒰rd_const
                     step4_𝒰wr_const)



text ‹Non speculative triviality›


lemma wfs4_ns_not_or:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "¬speculating s⇩3' \<or> ¬speculating s⇩4'"
    shows "¬speculating s⇩3' ∧ ¬speculating s⇩4'"  
  apply (subgoal_tac "\<omega> s⇩1' = \<omega> s⇩2'") defer
  using assms(1-4) step4_ω12_ns_equal wfs4_ns_def apply blast

  using assms apply (induction rule: valid_step4_cases)
  apply blast+
  apply (meson assms(1) step4_s⇩3_ns_s_mispred step4_s⇩4_ns_s_mispred)  
  apply (meson assms(1) step4_s⇩3_ns_s_mispred step4_s⇩4_ns_s_mispred)
  apply blast+
  using step2_ns_s⇩s'_no_mispred apply blast
  using step2_ns_s⇩s'_no_mispred apply blast
  apply blast+
  using step2_ns_s⇩s'_no_mispred apply blast+
  using assms(1) assms(2) step4_ω_eq_pc_eq wfs4_triv_def apply blast
  using assms(1) assms(2) step4_ω_eq_pc_eq wfs4_triv_def apply blast
  by blast+


lemma wfs4_ns_triv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
    shows "wfs4_triv s⇩1' s⇩2' s⇩3' s⇩4'"
  using assms apply (simp only: wfs4_triv_def)
  apply (intro conjI)
  using assms(5) step4_β12_equal_ns wfs4_ns_def apply blast
  using step4_ω12_ns_equal wfs4_triv_def apply blast
  using assms(5) step4_β_equal_ns apply blast
  using assms(5) step4_ω34_ns_equal by blast

lemma wfs4_ns_triv_or:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3' \<or> ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
    shows "wfs4_triv s⇩1' s⇩2' s⇩3' s⇩4'"
  using assms wfs4_ns_not_or wfs4_ns_triv by blast



lemma wfs4_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "wfs4_ns s⇩1 s⇩2 s⇩3 s⇩4"
    shows "wfs4_ns s⇩1' s⇩2' s⇩3' s⇩4'"
  using assms apply (simp only: wfs4_ns_def wfs4_triv_def)
  by (meson assms(5) step2_wfs2_ns step4_β_equal_ns step4_ω12_ns_equal step4_ω34_ns_equal 
            valid_step4_implies_valid_step2  wfs4_def wfs4_ns_def)

lemma step4_ns_triv_and_triv_low_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "s⇩3 ≈⇩ℒ s⇩4"
    shows "s⇩3' ≈⇩ℒ s⇩4'"
proof - 
  have wfs4': "wfs4 s⇩1' s⇩2' s⇩3' s⇩4'"
    using assms by (meson step4_wfs4)

  have "inst⇩𝒯 s⇩3' ≠ ⊥ \<Longrightarrow> inst⇩𝒯 s⇩4' ≠ ⊥ \<Longrightarrow> s⇩3' ≈⇩ℒ s⇩4'"    
    using low_equiv_s_op_high by auto

  have "inst⇩𝒯 s⇩3' = \<bottom> \<Longrightarrow> inst⇩𝒯 s⇩1' = ⊥ ∨ halting (pc⇩0 s⇩1') (Π s⇩1')"
    using inst⇩𝒯_NotHaltTrusted_NotNone op⇩ℒ_NotHaltNotTrusted_NotNone op⇩ℒ_not_inst⇩𝒯_s wfs2_def
    by (metis (no_types, lifting) wfs4' wfs4_def)

  moreover have "inst⇩𝒯 s⇩3' = \<bottom> \<Longrightarrow> s⇩3' ≈⇩ℒ s⇩4'"
    using wfs4' wfs4_def apply auto[1]
    apply (subgoal_tac "Δ⇩0 s⇩1' = Δ⇩0 s⇩2'
        ∧ (∀a . a ∈ 𝒰rd s⇩1' ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1' a = μ⇩0⇩r⇩e⇩a⇩d s⇩2' a)
        ∧ ω s⇩1' = ω s⇩2'
        ∧ β s⇩1' = β s⇩2'")
    apply (auto simp add: low_equiv_sub_def wfs2_def)[1]
    using assms(1-4) apply (rule step4_ω34_ns_equal)
    apply (subgoal_tac "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'")
    apply blast
    using assms(1-4) apply (rule step4_β_equal_ns)
    using assms(1-4) apply (rule step4_ω34_ns_equal)
    apply (subgoal_tac "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'")
    apply blast
    using assms(1-4) apply (rule step4_β_equal_ns)
    apply (subgoal_tac "s⇩1' ≈⇩ℒ s⇩2'") defer
    using assms(1) apply (simp only: valid_step4.simps tpod_step⇩4_def)
    apply (simp add: low_equiv_sub_def)
    apply (subgoal_tac "speculating s⇩1' ⟶ ¬ speculating s⇩2'")
    apply (subgoal_tac "inst⇩𝒯 s⇩1' = ⊥")
    apply blast defer
    using wfs2_s⇩n⇩s_ns apply force
    using calculation op⇩ℋ⇩ℒ_Halt_None by presburger

  moreover have "inst⇩𝒯 s⇩4' = \<bottom> \<Longrightarrow> s⇩3' ≈⇩ℒ s⇩4'"
    using wfs4' wfs4_def apply auto[1]
    apply (subgoal_tac "Δ⇩0 s⇩1' = Δ⇩0 s⇩2'
        ∧ (∀a . a ∈ 𝒰rd s⇩1' ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1' a = μ⇩0⇩r⇩e⇩a⇩d s⇩2' a)
        ∧ ω s⇩1' = ω s⇩2'
        ∧ β s⇩1' = β s⇩2'")
    apply (auto simp add: low_equiv_sub_def wfs2_def)[1]
    using assms(1-4) apply (rule step4_ω34_ns_equal)
    apply (subgoal_tac "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'")
    apply blast
    using assms(1-4) apply (rule step4_β_equal_ns)
    using assms(1-4) apply (rule step4_ω34_ns_equal)
    apply (subgoal_tac "β s⇩1' = β s⇩2' ∧ β s⇩3' = β s⇩4'")
    apply blast
    using assms(1-4) apply (rule step4_β_equal_ns)
    apply (subgoal_tac "s⇩1' ≈⇩ℒ s⇩2'") defer
    using assms(1) apply (simp only: valid_step4.simps tpod_step⇩4_def)
    apply (simp add: low_equiv_sub_def)
    apply (subgoal_tac "speculating s⇩2' ⟶ ¬ speculating s⇩1'")
    apply (subgoal_tac "inst⇩𝒯 s⇩2' = ⊥")
    apply blast defer
    using wfs2_s⇩n⇩s_ns apply force
    using calculation op⇩ℋ⇩ℒ_Halt_None
  proof -
    assume a1: "inst⇩𝒯 s⇩4' = ⊥"
    assume a2: "wfs2 s⇩2' s⇩4'"
    assume "Π s⇩1' = Π s⇩2'"
    assume "𝒯⇩ρ s⇩1' = 𝒯⇩ρ s⇩2'"
    show "inst⇩𝒯 s⇩2' = ⊥"
      using a2 a1 by (smt (z3) inst⇩𝒯_NotHaltTrusted_NotNone op⇩ℋ⇩ℒ_Halt_None op⇩ℒ_NotHaltNotTrusted_NotNone op⇩ℒ_not_inst⇩𝒯_s wfs2_def)
  qed

  then show ?thesis
    using ‹inst⇩𝒯 s⇩3' ≠ ⊥ ⟹ inst⇩𝒯 s⇩4' ≠ ⊥ ⟹ s⇩3' ≈⇩ℒ s⇩4'› using calculation(2) by blast
qed

lemma step4_ns_and_triv_low_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_ns s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "¬speculating s⇩3' ∧ ¬speculating s⇩4'"
      and "s⇩3 ≈⇩ℒ s⇩4"
    shows "s⇩3' ≈⇩ℒ s⇩4'"
  using assms step4_ns_triv_and_triv_low_equiv wfs4_ns_def by blast


lemma step4_ns_triv_low_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_triv s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "¬speculating s⇩3' \<or> ¬speculating s⇩4'"
      and "s⇩3 ≈⇩ℒ s⇩4"
    shows "s⇩3' ≈⇩ℒ s⇩4'"
  using assms(1-6) by (metis step4_ns_triv_and_triv_low_equiv wfs4_ns_not_or)


lemma step4_ns_low_equiv:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "wfs4_ns s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "¬speculating s⇩3' \<or> ¬speculating s⇩4'"
      and "s⇩3 ≈⇩ℒ s⇩4"
    shows "s⇩3' ≈⇩ℒ s⇩4'"
  using assms(1-6) step4_ns_and_triv_low_equiv wfs4_ns_def wfs4_ns_not_or by blast

text ‹Low equivalent triviality›

lemma step4_op⇩ℒ_s12_Δ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "s⇩3 ≈⇩ℒ s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "Δ⇩0 s⇩1' = Δ⇩0 s⇩2'"
  apply (subgoal_tac "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'")
  using step4_wfs4 wfs4_ns_opL_Δ⇩0_equal assms apply blast
  using assms(1) by auto

lemma step4_op⇩ℒ_Δ⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "s⇩3 ≈⇩ℒ s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "Δ⇩0 s⇩3' = Δ⇩0 s⇩4'"
  apply (subgoal_tac "Δ⇩0 s⇩1' = Δ⇩0 s⇩2'")
  defer
  using assms(1) assms(2) assms(3) assms(4) step4_op⇩ℒ_s12_Δ⇩0_equal apply blast
  using assms(1-3) apply (induction rule: valid_step4_cases)
  by (metis step2_Δ⇩0_equal)+

lemma step4_op⇩ℒ_s12_𝒰rd_\<mu>⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "∀a. a ∈ 𝒰rd s⇩1' ⟶
             μ⇩0⇩r⇩e⇩a⇩d s⇩1' a =
             μ⇩0⇩r⇩e⇩a⇩d s⇩2' a"
  apply (subgoal_tac "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'")
  using step4_wfs4 wfs4_ns_opL_μ⇩0_equal assms apply blast
  using assms(1) by auto

lemma step4_op⇩ℒ_𝒰rd_\<mu>⇩0_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "s⇩3 ≈⇩ℒ s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "∀a. a ∈ 𝒰rd s⇩3' ⟶
             μ⇩0⇩r⇩e⇩a⇩d s⇩3' a =
             μ⇩0⇩r⇩e⇩a⇩d s⇩4' a"
  apply (subgoal_tac "∀a. a ∈ 𝒰rd s⇩1' ⟶ μ⇩0⇩r⇩e⇩a⇩d s⇩1' a = μ⇩0⇩r⇩e⇩a⇩d s⇩2' a") defer
  using assms(1,2,4) step4_op⇩ℒ_s12_𝒰rd_\<mu>⇩0_equal apply blast
  apply (subgoal_tac "𝒰rd s⇩3' = 𝒰rd s⇩1'") defer
  using assms(1,2) step4_𝒰rd_equal apply blast
  apply (subgoal_tac "μ⇩0 s⇩1' = μ⇩0 s⇩3'") defer
  using assms(1,2) wfs4_def step2_μ⇩0_equal apply (meson valid_step4_implies_valid_step2)
  apply (subgoal_tac "μ⇩0 s⇩2' = μ⇩0 s⇩4'") defer
  using assms(1,2) wfs4_def step2_μ⇩0_equal apply (meson valid_step4_implies_valid_step2)
  using assms(1-3) by (induction rule: valid_step4_cases, simp_all)

lemma step4_op⇩ℒ_s12_\<beta>_equal:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "op⇩ℒ s⇩1 ≠ \<bottom> ∧ op⇩ℒ s⇩1' ≠ \<bottom>"
    shows "\<beta> s⇩1' = \<beta> s⇩2'"
  apply (subgoal_tac "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'")
  using step4_wfs4 wfs4_ns_opL_β_equal assms apply blast
  using assms(1) by auto


subsection ‹Instructions ι›


lemma step4_RegisterUpdate_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = RegisterUpdate r e"
      and "ι s⇩1 = ι s⇩2"
    shows "s⇩1' = trans_register_update (ρ s⇩1) r e s⇩1 ∧
           s⇩2' = trans_register_update (ρ s⇩2) r e s⇩2 ∧
           s⇩3' = trans_register_update (ρ s⇩3) r e s⇩3 ∧
           s⇩4' = trans_register_update (ρ s⇩4) r e s⇩4"
proof - 
  have "ι s⇩2 = RegisterUpdate r e"
    using assms(4-5) by auto
  then show ?thesis
    using step2_RegisterUpdate_ns assms wfs4_def valid_step4_implies_valid_step2 by meson
qed

lemma RegisterUpdate_ns_step4:
  assumes "s⇩1' = trans_register_update (ρ s⇩1) r e s⇩1 ∧
           s⇩2' = trans_register_update (ρ s⇩2) r e s⇩2 ∧
           s⇩3' = trans_register_update (ρ s⇩3) r e s⇩3 ∧
           s⇩4' = trans_register_update (ρ s⇩4) r e s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = RegisterUpdate r e"
      and "ι s⇩1 = ι s⇩2"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms valid_step4_equiv_valid_step2 apply simp 
  using assms RegisterUpdate_ns_step2 apply (simp only: tpod_step\<^sub>4_def wfs4_def)
  by presburger


lemma step4_Load_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = Load r e"
      and "ι s⇩1 = ι s⇩2"
    shows "s⇩1' = trans_load (ρ s⇩1) r e s⇩1 ∧
           s⇩2' = trans_load (ρ s⇩2) r e s⇩2 ∧
           s⇩3' = trans_load (ρ s⇩3) r e s⇩3 ∧
           s⇩4' = trans_load (ρ s⇩4) r e s⇩4"
proof -
  have "ι s⇩2 = r := mem[e, el]:u64"
    using assms(4) assms(5) by force
  then show ?thesis
    using step2_Load_ns assms wfs4_def valid_step4_implies_valid_step2 by meson
qed

lemma Load_ns_step4:
  assumes "s⇩1' = trans_load (ρ s⇩1) r e s⇩1 ∧
           s⇩2' = trans_load (ρ s⇩2) r e s⇩2 ∧
           s⇩3' = trans_load (ρ s⇩3) r e s⇩3 ∧
           s⇩4' = trans_load (ρ s⇩4) r e s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = Load r e"
      and "ι s⇩1 = ι s⇩2"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms valid_step4_equiv_valid_step2 apply simp 
  using assms Load_ns_step2 apply (simp only: tpod_step\<^sub>4_def wfs4_def)
  by presburger

lemma Goto_ns_step4:
  assumes "s⇩1' = trans_goto (ρ s⇩1) c s⇩1 ∧
           s⇩2' = trans_goto (ρ s⇩2) c s⇩2 ∧
           s⇩3' = trans_goto (ρ s⇩3) c s⇩3 ∧
           s⇩4' = trans_goto (ρ s⇩4) c s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3  ∧ ¬speculating s⇩4"
      and "ι s⇩1 = Goto c"
      and "ι s⇩1 = ι s⇩2"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms valid_step4_equiv_valid_step2 apply simp 
  using assms Goto_ns_step2 apply (simp only: tpod_step\<^sub>4_def wfs4_def)
  by presburger

lemma step4_BranchMispred_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
      and "ι s⇩1 = Branch e c"
      and "ι s⇩1 = ι s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
    shows "(s⇩1' = trans_tpred (ρ s⇩1) c s⇩1 ∧ 
            s⇩2' = trans_tpred (ρ s⇩2) c s⇩2 ∧ 
            s⇩3' = trans_tmispred (ρ s⇩3) c s⇩3 ∧ 
            s⇩4' = trans_tmispred (ρ s⇩4) c s⇩4) ∨
           (s⇩1' = trans_ntpred (ρ s⇩1) s⇩1 ∧ 
            s⇩2' = trans_ntpred (ρ s⇩2) s⇩2 ∧
            s⇩3' = trans_ntmispred (ρ s⇩3) c s⇩3 ∧ 
            s⇩4' = trans_ntmispred (ρ s⇩4) c s⇩4)"
  using assms apply (induct rule: valid_step4_cases)
  apply simp_all
  using step_BranchTrueMispred step_BranchTruePred eval_exp_bool.simps valid_step4_implies_valid_step2 valid_step2_step
  using eval_instr.simps(4) step_def trans_branch.simps(1) trans_branch.simps(2) apply presburger
  using step_BranchTrueMispred step_BranchTruePred eval_exp_bool.simps valid_step4_implies_valid_step2 valid_step2_step
  using eval_instr.simps(4) step_def trans_branch.simps(3) trans_branch.simps(4) apply presburger
  apply (smt (z3) Instr.inject(4) assms(1) eval_exp_bool.simps eval_instr.simps(4) step2_ns_ι_same step2_ns_eval_equiv_F step_BranchTruePred step_def tpod_step⇩2_def trans_branch.simps(1) valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)
  apply (smt (z3) Instr.inject(4) assms(1) eval_exp_bool.simps eval_instr.simps(4) step2_ns_ι_same step2_ns_eval_equiv_F step_BranchFalsePred step_def tpod_step⇩2_def trans_branch.simps(3) valid_step2_s⇩n⇩s_ns_step valid_step4_implies_valid_step2)
  apply (metis assms(1) step4_s⇩4_ns_ns_no_mispred)
  by (metis assms(1) step4_s⇩3_ns_ns_no_mispred)


lemma step4_Branch_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "¬speculating s⇩3 ∧ ¬speculating s⇩4"
      and "ι s⇩1 = Branch e c"
      and "pc s⇩1 = pc s⇩2"
      and "Δ⇩0 s⇩1 = Δ⇩0 s⇩2"
      and "β s⇩3 = β s⇩4"
    shows "(s⇩1' = trans_tpred (ρ s⇩1) c s⇩1 ∧
            s⇩2' = trans_tpred (ρ s⇩2) c s⇩2 ∧ 
            s⇩3' = trans_tmispred (ρ s⇩3) c s⇩3 ∧ 
            s⇩4' = trans_tmispred (ρ s⇩4) c s⇩4) ∨
           (s⇩1' = trans_ntpred (ρ s⇩1) s⇩1 ∧ 
            s⇩2' = trans_ntpred (ρ s⇩2) s⇩2 ∧
            s⇩3' = trans_ntmispred (ρ s⇩3) c s⇩3 ∧ 
            s⇩4' = trans_ntmispred (ρ s⇩4) c s⇩4) \<or>
           (s⇩1' = trans_tpred (ρ s⇩1) c s⇩1 ∧ 
            s⇩2' = trans_tpred (ρ s⇩2) c s⇩2 ∧
            s⇩3' = trans_tpred (ρ s⇩3) c s⇩3 ∧ 
            s⇩4' = trans_tpred (ρ s⇩4) c s⇩4) \<or>
           (s⇩1' = trans_ntpred (ρ s⇩1) s⇩1 ∧ 
            s⇩2' = trans_ntpred (ρ s⇩2) s⇩2 ∧
            s⇩3' = trans_ntpred (ρ s⇩3) s⇩3 ∧ 
            s⇩4' = trans_ntpred (ρ s⇩4) s⇩4)"
  apply (subgoal_tac "ι s⇩1 = ι s⇩2")
  using assms apply (induct rule: valid_step4_cases, simp_all)
  using eval_exp_bool.simps eval_instr.simps(4) step_def trans_branch.simps(2) apply presburger
  using eval_exp_bool.simps eval_instr.simps(4) step_def trans_branch.simps(4) apply presburger
  using eval_exp_bool.simps eval_instr.simps(4) step_def trans_branch.simps(1) trans_branch.simps(2) apply presburger  
  using eval_exp_bool.simps eval_instr.simps(4) step_def trans_branch.simps(3) trans_branch.simps(4) apply presburger
  apply (smt (z3) assms(1) step4_BranchMispred_ns)
  apply (smt (z3) assms(1) step4_BranchMispred_ns)
  apply (metis assms(1) pc⇩0_pc_no_spec step4_ns_pc_ρ_eq wfs2_s⇩n⇩s_ns)
  apply (metis assms(1) pc⇩0_pc_no_spec step4_ns_pc_ρ_eq wfs2_s⇩n⇩s_ns)
  by (metis assms(1) step4_ns_ι_s12_same)

lemma Branch_ns_step4:
  assumes "(s⇩1' = trans_tpred (ρ s⇩1) c s⇩1 ∧ (⟦e⟧Δ⇩n s⇩1) ∧ s⇩2' = trans_tpred (ρ s⇩2) c s⇩2 ∧ (⟦e⟧Δ⇩n s⇩2)
        ∧ (⟦e⟧Δ⇩n s⇩3) ∧ (⟦e⟧Δ⇩n s⇩4) ∧ (
            (s⇩3' = trans_tpred (ρ s⇩3) c s⇩3    ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ s⇩4' = trans_tpred (ρ s⇩4) c s⇩4    ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)) \<or>
            (s⇩3' = trans_tmispred (ρ s⇩3) c s⇩3 ∧  mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ s⇩4' = trans_tmispred (ρ s⇩4) c s⇩4 ∧  mispred (n s⇩4) (β s⇩4) (pc s⇩4))
          )) \<or>
           (s⇩1' = trans_ntpred (ρ s⇩1) s⇩1 ∧ ¬(⟦e⟧Δ⇩n s⇩1) ∧ s⇩2' = trans_ntpred (ρ s⇩2) s⇩2 ∧ ¬(⟦e⟧Δ⇩n s⇩2)
        ∧ ¬(⟦e⟧Δ⇩n s⇩3) ∧ ¬(⟦e⟧Δ⇩n s⇩4) ∧ (
            (s⇩3' = trans_ntpred (ρ s⇩3) s⇩3      ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3)) \<and> (s⇩4' = trans_ntpred (ρ s⇩4) s⇩4      ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)) \<or>
            (s⇩3' = trans_ntmispred (ρ s⇩3) c s⇩3 ∧  mispred (n s⇩3) (β s⇩3) (pc s⇩3)) \<and> (s⇩4' = trans_ntmispred (ρ s⇩4) c s⇩4 ∧  mispred (n s⇩4) (β s⇩4) (pc s⇩4))
          ))"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩1 = Branch e c"
      and "pc s⇩1 = pc s⇩2"
      and "¬speculating s⇩3" and "¬speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(15-18) tpod_step⇩4_def valid_step4_equiv_valid_step2 apply blast
  using assms(1-4,5-8,10-16) Branch_ns_step2 apply (simp only: tpod_step\<^sub>4_def wfs4_def)
  by (metis (no_types, lifting) assms(9) pc⇩0_pc_no_spec wfs2_s⇩n⇩s_ns)

lemma step4_SpecFence_s:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "ι s⇩3 = SpecFence"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "ι s⇩3 = ι s⇩4"
    shows "(s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_specfence (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_specfence (ρ s⇩4) s⇩4) ∨ 
           (s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
  using assms valid_step4_implies_valid_step2 valid_step2_step
  by (metis step4_not_resolve_s⇩3_ns_halting_specfence step4_not_resolve_s⇩3_spec_halting_not_specfence step4_not_resolve_s⇩4_ns_halting_specfence step4_not_resolve_s⇩4_spec_halting_not_specfence step_Resolve step_SpecFence)

lemma SpecFence_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_specfence (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_specfence (ρ s⇩4) s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = SpecFence"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7,11,13,15,17-18) SpecFence_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) SpecFence_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma step4_Load_s:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "ι s⇩3 = Load r e"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "ι s⇩3 = ι s⇩4"
    shows "(s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_load (ρ s⇩3) r e s⇩3 ∧ s⇩4' = trans_load (ρ s⇩4) r e s⇩4) ∨ 
           (s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
  using assms valid_step4_implies_valid_step2 valid_step2_step
  using step4_not_resolve_s⇩3_ns_halting_specfence step4_not_resolve_s⇩3_spec_halting_not_specfence step4_not_resolve_s⇩4_ns_halting_specfence step4_not_resolve_s⇩4_spec_halting_not_specfence
  by (metis (full_types) step_Load step_Resolve)


lemma step4_RegisterUpdate_s:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "ι s⇩3 = RegisterUpdate r e"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "ι s⇩3 = ι s⇩4"
    shows "(s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_register_update (ρ s⇩3) r e s⇩3 ∧ s⇩4' = trans_register_update (ρ s⇩4) r e s⇩4) ∨ 
           (s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
  using assms valid_step4_implies_valid_step2 valid_step2_step
  using step4_not_resolve_s⇩3_ns_halting_specfence step4_not_resolve_s⇩3_spec_halting_not_specfence step4_not_resolve_s⇩4_ns_halting_specfence step4_not_resolve_s⇩4_spec_halting_not_specfence
  by (metis (full_types) step_RegisterUpdate step_Resolve)


lemma step4_Goto_s:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "ι s⇩3 = Goto c"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "ι s⇩3 = ι s⇩4"
    shows "(s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_goto (ρ s⇩3) c s⇩3 ∧ s⇩4' = trans_goto (ρ s⇩4) c s⇩4) ∨ 
           (s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
  using assms valid_step4_implies_valid_step2 valid_step2_step
  using step4_not_resolve_s⇩3_ns_halting_specfence step4_not_resolve_s⇩3_spec_halting_not_specfence step4_not_resolve_s⇩4_ns_halting_specfence step4_not_resolve_s⇩4_spec_halting_not_specfence
  by (metis (no_types, hide_lams) step_Goto step_Resolve)


lemma Goto_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_goto (ρ s⇩3) c s⇩3 ∧ s⇩4' = trans_goto (ρ s⇩4) c s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = Goto c"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) Goto_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) Goto_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma NoOp_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_noop (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_noop (ρ s⇩4) s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = NoOp"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) NoOp_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) NoOp_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma Halt_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "s⇩3' = trans_resolve (ρ s⇩3) s⇩3"
      and "s⇩4' = trans_resolve (ρ s⇩4) s⇩4"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "halting⇩p⇩r⇩e⇩d s⇩3" and "halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  using assms Halt_s_step2 wfs4_def tpod_step⇩4_def apply fastforce
  using assms wfs2_def wfs4_def by force

lemma RegisterUpdate_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_register_update (ρ s⇩3) r e s⇩3 ∧ s⇩4' = trans_register_update (ρ s⇩4) r e s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = RegisterUpdate r e"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) RegisterUpdate_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) RegisterUpdate_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma Load_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_load (ρ s⇩3) r e s⇩3 ∧ s⇩4' = trans_load (ρ s⇩4) r e s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = Load r e"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) Load_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) Load_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma Store_s_step4:
  assumes "s⇩1' = s⇩1"
      and "s⇩2' = s⇩2"
      and "(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_store (ρ s⇩3) e\<^sub>1 e\<^sub>2 s⇩3 ∧ s⇩4' = trans_store (ρ s⇩4) e\<^sub>1 e\<^sub>2 s⇩4) ∨ 
           (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = Store e\<^sub>1 e\<^sub>2"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) Store_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) Store_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force

lemma Branch_s_step4:
  assumes \<open>s⇩1' = s⇩1\<close>
      and \<open>s⇩2' = s⇩2\<close>
      and ‹(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (
            (s⇩3' = trans_tpred (ρ s⇩3) c s⇩3 ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (⟦e⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_ntpred (ρ s⇩3) s⇩3 ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬(⟦e⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_tmispred (ρ s⇩3) c s⇩3 ∧ mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (⟦e⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_ntmispred (ρ s⇩3) c s⇩3 ∧ mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬(⟦e⟧(Δ⇩n s⇩3)))) \<and> 
           (
            (s⇩4' = trans_tpred (ρ s⇩4) c s⇩4 ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ (⟦e⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_ntpred (ρ s⇩4) s⇩4 ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ ¬(⟦e⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_tmispred (ρ s⇩4) c s⇩4 ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ (⟦e⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_ntmispred (ρ s⇩4) c s⇩4 ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ ¬(⟦e⟧(Δ⇩n s⇩4))))
           ) ∨ (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)›
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = Branch e c"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(1-2) assms(17) assms(19-20) tpod_step⇩4_def apply fastforce
  apply (subgoal_tac \<open>ι s⇩3 = ι s⇩4\<close>)
  apply (rule conjI)
  using assms(1,3-7, 11, 13, 15, 17-18) Branch_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(2-8,12,14,16-18) Branch_s_step2 wfs4_def tpod_step⇩4_def apply (metis (no_types, hide_lams))
  using assms(4,6) wfs2_def wfs4_def by force


lemma Branch_s_diff_eval_step4:
  assumes \<open>s⇩1' = s⇩1\<close>
      and \<open>s⇩2' = s⇩2\<close>
      and ‹(¬resolve (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (
            (s⇩3' = trans_tpred (ρ s⇩3) c s⇩3 ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (⟦e⇩1⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_ntpred (ρ s⇩3) s⇩3 ∧ ¬mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬(⟦e⇩1⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_tmispred (ρ s⇩3) c s⇩3 ∧ mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ (⟦e⇩1⟧(Δ⇩n s⇩3))) ∨ 
            (s⇩3' = trans_ntmispred (ρ s⇩3) c s⇩3 ∧ mispred (n s⇩3) (β s⇩3) (pc s⇩3) ∧ ¬(⟦e⇩1⟧(Δ⇩n s⇩3)))) \<and> 
           (
            (s⇩4' = trans_tpred (ρ s⇩4) c s⇩4 ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ (⟦e⇩2⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_ntpred (ρ s⇩4) s⇩4 ∧ ¬mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ ¬(⟦e⇩2⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_tmispred (ρ s⇩4) c s⇩4 ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ (⟦e⇩2⟧(Δ⇩n s⇩4))) ∨ 
            (s⇩4' = trans_ntmispred (ρ s⇩4) c s⇩4 ∧ mispred (n s⇩4) (β s⇩4) (pc s⇩4) ∧ ¬(⟦e⇩2⟧(Δ⇩n s⇩4))))
           ) ∨ (resolve (n s⇩3) (β s⇩3) (pc s⇩3) \<and> s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)›
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι s⇩3 = Branch e⇩1 c"
      and "ι s⇩4 = Branch e⇩2 c"
      and "n s⇩3 = n s⇩4 ∧ β s⇩3 = β s⇩4 ∧ pc s⇩3 = pc s⇩4"
      and "speculating s⇩3" and "speculating s⇩4"
      and "¬halting⇩p⇩r⇩e⇩d s⇩1" and "¬halting⇩p⇩r⇩e⇩d s⇩2" and "¬halting⇩p⇩r⇩e⇩d s⇩3" and "¬halting⇩p⇩r⇩e⇩d s⇩4"
      and "conformantEntrypoint s⇩1 s⇩1'" and "conformantEntrypoint s⇩2 s⇩2'" and "conformantEntrypoint s⇩3 s⇩3'" and "conformantEntrypoint s⇩4 s⇩4'"
      and "tpod_step\<^sub>4 s⇩1 s⇩2 s⇩3 s⇩4" and "tpod_step\<^sub>4 s⇩1' s⇩2' s⇩3' s⇩4'"
      and \<open>violation⇩ℒ s⇩1 s⇩2 s⇩3 s⇩4\<close> and \<open>violation⇩ℒ s⇩1' s⇩2' s⇩3' s⇩4'\<close>
    shows "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
  apply (cases \<open>e⇩1 = e⇩2\<close>)
  using assms Branch_s_step4 apply blast
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3') \<and> (s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms apply (auto simp add: step_def tpod_step⇩4_def)
  using terminating_def by blast+

lemma step4_Halt_s:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "speculating s⇩3 ∧ speculating s⇩4"
      and "halting⇩p⇩r⇩e⇩d s⇩3 ∧ halting⇩p⇩r⇩e⇩d s⇩4"
    shows "(s⇩1' = s⇩1 ∧ s⇩2' = s⇩2 ∧ s⇩3' = trans_resolve (ρ s⇩3) s⇩3 ∧ s⇩4' = trans_resolve (ρ s⇩4) s⇩4)"
  using assms valid_step4_implies_valid_step2 valid_step2_step
  using step4_not_resolve_s⇩3_ns_halting_specfence step4_not_resolve_s⇩3_spec_halting_not_specfence step4_not_resolve_s⇩4_ns_halting_specfence step4_not_resolve_s⇩4_spec_halting_not_specfence
  by (metis step_Resolve)

lemma step4_forward_jumps: 
  assumes ‹(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')›
      and ‹forward_jumps s⇩1› and ‹forward_jumps s⇩2› and ‹forward_jumps s⇩3› and ‹forward_jumps s⇩4›
    shows ‹forward_jumps s⇩1' ∧ forward_jumps s⇩2' ∧ forward_jumps s⇩3' ∧ forward_jumps s⇩4'›
  using assms valid_step4_implies_valid_step2 by (metis step2_forward_jumps)

lemma step4_future⇩\<rho>_P: 
  assumes ‹(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')›
      and ‹forward_jumps s⇩1› and ‹forward_jumps s⇩2› and ‹forward_jumps s⇩3› and ‹forward_jumps s⇩4›
      and ‹future⇩\<rho>_P s⇩1 P› and ‹future⇩\<rho>_P s⇩2 P› and ‹future⇩\<rho>_P s⇩3 P› and ‹future⇩\<rho>_P s⇩4 P›
    shows ‹future⇩\<rho>_P s⇩1' P ∧ future⇩\<rho>_P s⇩2' P ∧ future⇩\<rho>_P s⇩3' P ∧ future⇩\<rho>_P s⇩4' P›
  using assms step2_future⇩\<rho>_P valid_step4_implies_valid_step2 by meson

lemma step4_future⇩ι_P: 
  assumes ‹(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')›
      and ‹forward_jumps s⇩1› and ‹forward_jumps s⇩2› and ‹forward_jumps s⇩3› and ‹forward_jumps s⇩4›
      and ‹future⇩ι_P s⇩1 P› and ‹future⇩ι_P s⇩2 P› and ‹future⇩ι_P s⇩3 P› and ‹future⇩ι_P s⇩4 P›
    shows ‹future⇩ι_P s⇩1' P ∧ future⇩ι_P s⇩2' P ∧ future⇩ι_P s⇩3' P ∧ future⇩ι_P s⇩4' P›
  using assms step2_future⇩ι_P valid_step4_implies_valid_step2 wfs4_def by meson

lemma step4_s\<^sub>3_no_mispred_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬mispred (n s⇩3) (β s⇩3) (pc s⇩3)"
      and "¬speculating s⇩3"
    shows "¬speculating s⇩3'"
  apply (subgoal_tac "(s⇩1, s⇩3) ↝⇩2 (s⇩1', s⇩3')")
  using assms(2-3) step2_no_mispred_ns apply blast
  using assms(1) valid_step4_implies_valid_step2 by blast

lemma step4_s\<^sub>4_no_mispred_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
      and "¬speculating s⇩4"
    shows "¬speculating s⇩4'"
  apply (subgoal_tac "(s⇩2, s⇩4) ↝⇩2 (s⇩2', s⇩4')")
  using assms(2-3) step2_no_mispred_ns apply blast
  using assms(1) valid_step4_implies_valid_step2 by blast

lemma step4_no_mispred_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "¬mispred (n s⇩3) (β s⇩3) (pc s⇩3)" and "¬mispred (n s⇩4) (β s⇩4) (pc s⇩4)"
      and "¬speculating s⇩3" and "¬speculating s⇩4"
    shows "¬speculating s⇩3' \<and> ¬speculating s⇩4'"
  using assms(1-5) by (metis step4_s⇩3_ns_s_mispred step4_s⇩4_ns_s_mispred)

lemma step4_no_branch_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "(\<forall> e c. \<iota> s⇩3 \<noteq> Branch e c)" 
      and "(\<forall> e c. \<iota> s⇩4 \<noteq> Branch e c)"
      and "¬speculating s⇩3" and "¬speculating s⇩4"
    shows "¬speculating s⇩3' \<and> ¬speculating s⇩4'"
  using assms apply (induction rule: valid_step4_cases)
  by blast+

lemma step4_ι_not_branch_ns:
  assumes "(s⇩1, s⇩2, s⇩3, s⇩4) ↝⇩4 (s⇩1', s⇩2', s⇩3', s⇩4')"
      and "wfs4 s⇩1 s⇩2 s⇩3 s⇩4"
      and "ι_not_branch (\<iota> s⇩3)"
      and "ι_not_branch (\<iota> s⇩4)"
      and "¬speculating s⇩3" and "¬speculating s⇩4"
    shows "¬speculating s⇩3' \<and> ¬speculating s⇩4'"
  using assms step4_no_branch_ns ι_not_branch_def by blast

end
